from main_dash           import app
from marker              import MarkerColor
from views.visualisation import layout_visualisation_tab, visualisation_modal, layout_timebar
from views.rms           import layout_rm_tab, initialiseRmsDataToColor
from views.sensors       import layout_sensors_tab
from views.configuration import layour_configuration_tab
from views.devices       import layout_devices_tab, checkForDuplicities, initialiseDevicesToColor
from views.view          import layout_blank_view_tab
from png_generator       import PngGenerator
from appEnums            import AppTabs, ConfigEnum, LimitsEnum, RequestEnums, RedisServerEnum, RedisEnum
from utils               import checkLimits, getDictData, getJammingIndex, getJammingIndexes, fillWithZeros

import globals

import dash_bootstrap_components as dbc
import json
import numpy                     as np
import os
import time
import redis

from dash                         import Output, Input, dcc
from multiprocessing              import Process, Manager
from werkzeug.middleware.profiler import ProfilerMiddleware


CONFIGURATION_FILE = '_app_config.json'
TIME_INTERVAL_MS   = 100


def filterJamDataByIndex(data, indexes):
    ''' Filter all Jamming data by index, so only jamming data with some device in it are transmitted to Dash app.'''
    result = {}
    for i in indexes:
        if i < len(data):
            result.update({i: data[i]})
    return result


def jamming_data_processing(queue_request, queue_data):
    ''' Main method processing data from Redis server to Dash app.
        Map Jamming Data from Redis server are used in PNG image Generator for Dash ImageOverlay layer.
    '''
    flip        = {}
    i           = 0
    thread_tick = 0
    indexes     = []
    png_dimensions         = []     # Number if tiles in PNG Image, [0] x, [1] y
    png_resolution         = 800    # Bigger resolution might cause interupts and lagging
    redis_config           = None
    redis_connect          = None
    previous_redis_config  = None
    previous_fusion_config = None
    png_generator          = None
    colormap_min           = None
    colormap_max           = None
    image_path             = None
    is_redis_connected     = False
    
    while True:

        retrieved_data        = None
        redis_connect_changed = False
        redis_updated         = False

        # Get request first
        if not queue_request.empty():
            requests = []
            while not queue_request.empty():
                requests.append(queue_request.get())

            request       = requests[-1]
            indexes       = request[RequestEnums.INDEXES   .value]
            flip          = request[RequestEnums.FLIP      .value]
            redis_config  = request[RequestEnums.REDIS     .value]
            fusion_config = request[RequestEnums.CONFIG    .value]
            dimensions    = request[RequestEnums.DIMENSIONS.value]  # [0] lat, [1] lon

            # Init PNG Generator or update it if steps for lat, lon changed
            if png_generator == None or png_dimensions != dimensions:
                png_dimensions = dimensions
                png_generator  = PngGenerator(png_resolution, png_resolution, dimensions[LimitsEnum.LONGITUDE.value], dimensions[LimitsEnum.LATITUDE.value])
                                                                                    #    ^^^ first is Longitude, as that represents 'x' axis
            # Connect to Redis if possile AND check if Redis config did not change
            if redis_connect == None or redis_config != previous_redis_config:
                redis_connect = redis.Redis(host=redis_config[RedisEnum.IP.string], port=redis_config[RedisEnum.PORT.string], db=0)
                previous_redis_config = redis_config

                redis_connect_changed = True

        # Send full config to Redis only if something changed
        try:
            if redis_connect != None:
                if fusion_config != previous_fusion_config or redis_connect_changed:
                    config_json = json.dumps(fusion_config)
                    redis_connect.set(RedisServerEnum.CONFIGURATION.value, config_json)
                    previous_fusion_config = fusion_config
                    app.logger.info(f'App/REDIS: Updating Configuration')
                    redis_updated = True
                else:
                    pass

                # Read data from Redis
                retrieved_data_string = redis_connect.get(RedisServerEnum.RESULTS.value)
                
                if retrieved_data_string != None:
                    retrieved_data = json.loads(retrieved_data_string)
                    is_redis_connected = True
        
        except redis.exceptions.ConnectionError as e:
            app.logger.warning(f'App/REDIS: {e}')
            is_redis_connected = False
    
        # Process data only if somehing received from Redis
        if retrieved_data != None:

            # REDIS Data - Map Visualisation data are **always** in ['jammer']['density'][0]
            data_to_process = retrieved_data['jammer']['density'][0]

            # Density index - 0 (as that is for map)
            # lat_i_min, lat_i_max, lon_i_min, lon_i_max = getJammingIndexes(fusion_config, LimitsEnum.LONGITUDE, LimitsEnum.LATITUDE, 0)

            lon_i_min, lon_i_max, _, _ = getJammingIndex(fusion_config, LimitsEnum.LONGITUDE, 0)
            lat_i_min, lat_i_max, _, _ = getJammingIndex(fusion_config, LimitsEnum.LATITUDE,  0)

            data_to_process = fillWithZeros(data_to_process,
                                            dimensions[LimitsEnum.LATITUDE.value],
                                            dimensions[LimitsEnum.LONGITUDE.value],
                                            lat_i_min,
                                            lat_i_max,
                                            lon_i_min,
                                            lon_i_max)


            # Update min max for colorbar if necessary
            data_min = np.array(data_to_process).min()
            data_max = np.array(data_to_process).max()

            # update_colorbar = False
            # if colormap_min == None or colormap_min > data_min:
            #     colormap_min = data_min
            #     update_colorbar = True
            # if colormap_max == None or colormap_max < data_max:
            #     colormap_max = data_max
            #     update_colorbar = True
            # if update_colorbar:
            #     png_generator.update_colorbar(colormap_min,colormap_max)

            png_generator.update_colorbar(data_min,data_max)


            # Generate PNG image
            try:
                picture    = png_generator.generate_picture(data_to_process, flip)
                image_path = os.path.join("assets", f"redis_jamming_{i}.png")

                picture.save(image_path)

            except ValueError as e:
                app.logger.warning(e)

            # Filter Map Jamming data - only those needed / with device on it
            filtered_jam_data = filterJamDataByIndex(data_to_process, indexes)

            # [REDIS]
            stream_data = {
                ConfigEnum.IMAGE.value:        image_path,
                ConfigEnum.DATA.value:         filtered_jam_data,
                ConfigEnum.TIME.value:         thread_tick,
                ConfigEnum.REDIS_ACTIVE.value: is_redis_connected,
                ConfigEnum.REDIS_UPDATED.value:redis_updated,
                ConfigEnum.ALL_DATA.value:     retrieved_data
            }

            # Send data to queue to be processed by Dash app
            queue_data.put(stream_data)


        time.sleep(1/TIME_INTERVAL_MS)
        i           += 1
        thread_tick += 1
        if i >= 100:
            i = 0


def createLayout(sensor_init_data, rms_init_data, limits_init_data, processing_init_data, devices_init_data, initial_app_data, redis_init_data, spec_view_init_data, queue_request, queue_data):

    # Prepare all data
    devices_init_data, _, _ = checkForDuplicities(devices_init_data)

    # Update Device and RMs Init data, as they can have color set to wrong value
    devices_init_data = initialiseDevicesToColor(devices_init_data, MarkerColor.GREEN)
    rms_init_data     = initialiseRmsDataToColor(rms_init_data,     MarkerColor.GREEN)

    limits_init_data = checkLimits(limits_init_data)

    gnssLimits,                  layout_configuration = layour_configuration_tab(limits_init_data, processing_init_data, redis_init_data)
    devices_data_with_tileIndex, layout_visualisation = layout_visualisation_tab(devices_init_data, initial_app_data, queue_request, queue_data, gnssLimits, limits_init_data)

    tab_style          = {'height': '3em', 'padding':'0.6em'}
    tab_style_selected = {'height': '3em', 'padding':'0.6em'}

    layout = dbc.Container(
        [   
            # Intervals - NOTE - must be here as top hierarchy - if lower, non-active Tab will disable the counter
            #                    and all timing will not propagate into the page.

            # Interval for all the logic, in [ms] - every 1 sec
            dcc.Interval(id="interal_logic", interval=TIME_INTERVAL_MS),

            # Interval for saving app config periodically [ms] - 5 secs should be ok
            dcc.Interval(id='interal_save_session', interval=5000),

            # Interval for switching to Config Tab, if no config file present (launching for first time)
            dcc.Interval(id="interval_initial_config", interval=1000, n_intervals=0),

            # Modal in Visualisation
            visualisation_modal(),

            dcc.Tabs(
                [
                    dcc.Tab(label="Visualisation",          value=AppTabs.MAIN   .value, children=[layout_visualisation],                                         style=tab_style, selected_style=tab_style_selected),
                    dcc.Tab(label="Specific Visualisation", value=AppTabs.SPEC   .value, children=[layout_blank_view_tab(spec_view_init_data, limits_init_data)], style=tab_style, selected_style=tab_style_selected),
                    dcc.Tab(label="RMs",                    value=AppTabs.RMS    .value, children=[layout_rm_tab(rms_init_data)],                                 style=tab_style, selected_style=tab_style_selected),
                    dcc.Tab(label="Devices",                value=AppTabs.DEVICES.value, children=[layout_devices_tab(devices_data_with_tileIndex)],              style=tab_style, selected_style=tab_style_selected),
                    dcc.Tab(label="Sensors",                value=AppTabs.SENSORS.value, children=[layout_sensors_tab(sensor_init_data)],                         style=tab_style, selected_style=tab_style_selected),
                    dcc.Tab(label="Configuration",          value=AppTabs.CONFIG .value, children=[layout_configuration],                                         style=tab_style, selected_style=tab_style_selected),
                ],
                id="tabs_main",
                value=AppTabs.MAIN.value,
            ),

            layout_timebar()
        ],
        style={'margin-top': '2.999em', 'margin-bottom': '4em'},
        fluid=True
    )
    return layout


''' switch_tabs & update_max_intervals ensures app will switch to 'Configuration' tab when there
        is not confgiruation file present in root folder. '''
@app.callback(
    Output('tabs_main',              'value'),
    Input('interval_initial_config', 'n_intervals')
)
def switch_tabs(n_intervals):
    # Check if the configuration file is present - if not, switch to config tab
    if os.path.isfile( os.path.join(os.path.dirname(__file__), CONFIGURATION_FILE) ):
        return AppTabs.MAIN.value
    else:
        if n_intervals >= 1:  # Switch to Config Tab after 1 second
            return AppTabs.MAIN.value
        else:
            return AppTabs.CONFIG.value

@app.callback(
    Output('interval_initial_config', 'max_intervals'),
    Input('tabs_main',                'value')
)
def update_max_intervals(selected_tab):
    ''' Set max_intervals to 'disable' interval for empty config file, causing switch from map to config tab. '''
    return 1 if selected_tab == AppTabs.MAIN.value else 1



def check_for_config():
    """ 
    
    Parameters
    ----------


    Returns
    -------
        sensor_data : pd.DataFrame
            Sensor config data
        
        limits_data : dict
            Limits data
    """
    config_file_path = os.path.join(os.path.dirname(__file__), CONFIGURATION_FILE)
    
    if os.path.exists(config_file_path):
        with open(config_file_path, 'r') as config_file:
            data = json.load(config_file)

            try:
                sensor_data         = getDictData(data[ConfigEnum.CONFIG.value], ConfigEnum.SENSORS   .value, True)
                limits_data         = getDictData(data[ConfigEnum.CONFIG.value], ConfigEnum.LIMITS    .value, False)
                processing_data     = getDictData(data[ConfigEnum.CONFIG.value], ConfigEnum.PROCESSING.value, False)
                redis_init_data     = getDictData(data[ConfigEnum.CONFIG.value], ConfigEnum.REDIS     .value, False)
                spec_view_init_data = getDictData(data[ConfigEnum.CONFIG.value][ConfigEnum.INTEGRATION.value], ConfigEnum.JAMMER.value, True)
            except KeyError as e:
                app.logger.warning(f"{__name__}: {e}")
                sensor_data     = None
                limits_data     = None
                processing_data = None
                redis_init_data = None
                spec_view_init_data = None

            rms_data         = getDictData(data, ConfigEnum.RMS     .value, True)
            devices_data     = getDictData(data, ConfigEnum.DEVICES .value, True)
            initial_app_data = getDictData(data, ConfigEnum.APP_DATA.value, False)

            return sensor_data, rms_data, limits_data, processing_data, devices_data, initial_app_data, redis_init_data, spec_view_init_data
    else:
        app.logger.warning(f"Config file {CONFIGURATION_FILE} not found - new session.")
        return (None,)*8



def server_layout():
    ''' Reload layout completely when reloading page, not only when app starts. '''
    sensor_data, rms_data, limits_data, processing_data, devices_data, initial_app_data, redis_init_data, spec_view_init_data = check_for_config()
    return createLayout(sensor_data, rms_data, limits_data, processing_data, devices_data, initial_app_data, redis_init_data, spec_view_init_data, queue_request, queue_data)


# -----------------------------------------------------------------------------
if __name__ == '__main__':

    globals.init()
    
    multiprocess_manager = Manager()
    queue_request = multiprocess_manager.Queue()
    queue_data    = multiprocess_manager.Queue()

    app.layout = server_layout

    # Profiler settings (if necessary)
    if os.getenv("PROFILER", None):
        app.server.config["PROFILE"] = True
        app.server.wsgi_app = ProfilerMiddleware(
            app.server.wsgi_app, 
            sort_by=["cumtime"], 
            restrictions=[50],
            stream=None,
            profile_dir='_profiling'
        )

    # Run Jamming Data processing
    jamming_data_process = Process(target=jamming_data_processing, args=(queue_request, queue_data,))
    jamming_data_process.start()

    #app.run_server(host='147.229.177.187', port='8050', debug=True, dev_tools_hot_reload=False)
    # app.run_server(host='127.0.0.1', port='8050', debug=True, dev_tools_hot_reload=False)
    # app.run_server(host='127.0.0.1', port='8050', debug=True, dev_tools_hot_reload=True)
    app.run_server(host='127.0.0.1', port='8050', debug=False)


