from enum import Enum
import os

from main_dash import app

class Marker:
    """
    tileIndex       - index of tile in GeoJson structure, to which this marker belongs to.
                        It's used to access the tile's sensitivity and check, if sensitivity in 
                        the marker's position is below/above marker's sensitivity threshold
                        (which is RMs records)
    timeJammingSince - internal time, sice when is Marker being jammed - interference
                        is bigger than threshold

    """
    def __init__(self, name, typ, color, lat, lon, hgt, sensitivity, incoming, backup, enabledInFusion=True, tileIndex=None, timeJammingSince=None, probability=None, rms_confirmed=None, mlat_system=None):
        self.name             = name
        self.typ              = typ
        self.color            = color
        self.lat              = lat
        self.lon              = lon
        self.hgt              = hgt
        self.sensitivity      = sensitivity
        self.incoming         = incoming
        self.backup           = backup
        self.enabledInFusion  = enabledInFusion
        self.tileIndex        = tileIndex
        self.timeJammingSince = timeJammingSince
        self.probability      = probability         # calculated probability based on incoming Jamming Data and set Sensitivity
        self.rms_confirmed    = rms_confirmed       # list of RMs for given device type and if individual RM was confirmed or not
        self.mlat_system      = mlat_system         # Id name of MLAT system, ONLY for device type of 'MLAT Antena'

    def __str__(self):
        return f"{MarkerProperty.NAME.value}: {self.name}, \
                {MarkerProperty.TYP.value}:{self.typ}, \
                {MarkerProperty.COLOR.value}:{self.color}, \
                {MarkerProperty.LAT.value}:{self.lat}, \
                {MarkerProperty.LON.value}:{self.lon}, \
                {MarkerProperty.HGT.value}:{self.hgt}, \
                {MarkerProperty.SENSITIVITY.value}:{self.sensitivity}, \
                {MarkerProperty.INCOMING.value}:{self.incoming}, \
                {MarkerProperty.BACKUP.value}:{self.backup}, \
                {MarkerProperty.ENABLED_FUSION.value}:{self.enabledInFusion}, \
                {MarkerProperty.TILE_INDEX.value}:{self.tileIndex}, \
                {MarkerProperty.TIME_JAM_SINCE.value}:{self.timeJammingSince}, \
                {MarkerProperty.PROBABILITY.value}:{self.probability}, \
                {MarkerProperty.RMS_CONFIRMED.value}:{self.rms_confirmed}, \
                {MarkerProperty.MLAT_SYSTEM.value}:{self.mlat_system}"

    def to_dict(self):
        return {
            MarkerProperty.NAME.value:           self.name,
            MarkerProperty.TYP.value:            self.typ,
            MarkerProperty.COLOR.value:          self.color,
            MarkerProperty.LAT.value:            self.lat,
            MarkerProperty.LON.value:            self.lon,
            MarkerProperty.HGT.value:            self.hgt,
            MarkerProperty.SENSITIVITY.value:    self.sensitivity,
            MarkerProperty.INCOMING.value:       self.incoming,
            MarkerProperty.BACKUP.value:         self.backup,
            MarkerProperty.ENABLED_FUSION.value: self.enabledInFusion,
            MarkerProperty.TILE_INDEX.value:     self.tileIndex,
            MarkerProperty.TIME_JAM_SINCE.value: self.timeJammingSince,
            MarkerProperty.PROBABILITY.value:    self.probability,
            MarkerProperty.RMS_CONFIRMED.value:  self.rms_confirmed,
            MarkerProperty.MLAT_SYSTEM.value:    self.mlat_system,
        }

    @staticmethod
    def dict_to_marker(dict_data):
        return Marker(
            dict_data[MarkerProperty.NAME.value],
            dict_data[MarkerProperty.TYP.value],
            dict_data[MarkerProperty.COLOR.value],
            dict_data[MarkerProperty.LAT.value],
            dict_data[MarkerProperty.LON.value],
            dict_data[MarkerProperty.HGT.value],
            dict_data[MarkerProperty.SENSITIVITY.value],
            dict_data[MarkerProperty.INCOMING.value],
            dict_data[MarkerProperty.BACKUP.value],
            dict_data[MarkerProperty.ENABLED_FUSION.value],
            dict_data[MarkerProperty.TILE_INDEX.value],
            dict_data[MarkerProperty.TIME_JAM_SINCE.value],
            dict_data[MarkerProperty.PROBABILITY.value],
            dict_data[MarkerProperty.RMS_CONFIRMED.value],
            dict_data[MarkerProperty.MLAT_SYSTEM.value]
        )

    @staticmethod
    def marker_to_dict(marker_data):
        result = []
        if marker_data != None:
            for m in marker_data:
                result.append(m)
        return result


class MarkerProperty(Enum):
    NAME           = ("name", "Name")
    TYP            = ("typ", "Type")
    COLOR          = ("color", "Color")
    LAT            = ("lat", "Latitude")
    LON            = ("lon", "Longitude")
    HGT            = ("hgt", "Height")
    SENSITIVITY    = ("sensitivity", "Sensitivity")
    INCOMING       = ("incoming", "Incoming")
    BACKUP         = ("backup", "Backup")
    ENABLED_FUSION = ("enabledInFusion", "Enabled in Fusion")
    TILE_INDEX     = ("tileIndex", "Tile Index")
    TIME_JAM_SINCE = ("timeJammingSince", "Jammed Since")
    IDENTIFIER     = ("identifier", "ID")
    PROBABILITY    = ("probability", "Probability")
    RMS_CONFIRMED  = ("rms_confirmed", "RMS Confirmed")
    MLAT_SYSTEM    = ('mlat_system', "MLAT System")

    def __new__(cls, value, string_value):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.string = string_value
        return obj

    @classmethod
    def get_enum_by_value(cls, value):
        for member in cls:
            if member._value_ == value:
                return member
        raise ValueError(f"No enum member with id value '{value}'")


class MarkerColor(Enum):
    GREEN =      "green"
    ORANGE =     "orange"
    RED =        "red"

    @classmethod
    def from_string(cls, string_value):
        for member in cls:
            if member.value == string_value:
                return member
        app.logger.warning(f"MarkerColor: No enum member with value '{string_value}'. Returning default - GREEN")
        # raise ValueError(f"No enum member with value '{string_value}'")
        return MarkerColor.GREEN


class MarkerType(Enum):
    AIRPLANE     = "Airplane"
    VEHICLE      = "Vehicle"
    RADAR        = "RADAR"
    GROUND       = "Ground COM System"
    ADSB         = "ADS-B System"
    MLAT         = "MLAT System"
    PROCESSING   = "Data Processing System"
    MLAT_ANTENNA = "MLAT Antenna"

    @classmethod
    def from_string(cls, string_value):
        for member in cls:
            if member.value == string_value:
                return member
        raise ValueError(f"No enum member with value '{string_value}'")



class MarkerIcon:
    ''' Dash Leaflet Marker Icon custom settings. Properties are tied to the .png icon itself.
        PNG Icon can be customised, there should be original .svg / .adobe / .affinity file
        in 'assets/icon' folder.
    '''
    ICON_PROPERTIES = {
        "iconSize":      [40,  40],
        "shadowSize":    [50,  64],
        "iconAnchor":    [20,  40],
        "shadowAnchor":  [ 4,  62],
        "popupAnchor":   [ 0, -50],
        "tooltipAnchor": [20, -20]
    }

    def __init__(self, icon_type, color):
        if color not in MarkerColor or icon_type not in MarkerType:
            raise ValueError("Invalid color or type")

        self.color = color
        self.type = icon_type

    def getProperties(self):
        icon_path = os.path.join("assets", "icons", f"marker_{self.color.value}_{self.type.value.lower().replace(' ', '_')}.png")
        json_structure = {
            "iconUrl": icon_path,
            **self.ICON_PROPERTIES
        }
        return json_structure

    @classmethod
    def getColor(cls, json_structure):
        for color in MarkerColor:
            if color.value in json_structure["iconUrl"]:
                break
        else:
            color = None

        return color

    @classmethod
    def getType(cls, json_structure):
        for markerType in MarkerType:
            if markerType.value in json_structure["iconUrl"]:
                break
        else:
            markerType = None

        return markerType

    @classmethod
    def getTypeAndColor(cls, json_structure):
        return MarkerIcon.getColor(json_structure), MarkerIcon.getType(json_structure)


