from appEnums import FlipEnum

from PIL import Image, ImageOps


class PngGenerator:
    ''' Generates a PNG image with given color scheme. '''
    def __init__(self, width, height, width_tiles, height_tiles, colormap_name='viridis'):
        self.width = width
        self.height = height
        self.width_tiles = width_tiles
        self.height_tiles = height_tiles
        self.tile_width = width // width_tiles
        self.tile_height = height // height_tiles
        self.canvas = Image.new("RGBA", (width, height))
        self.colormap = ColorIntensity()

    def generate_picture(self, intensities, flip):
        ''' '''
        if len(intensities) != self.width_tiles * self.height_tiles:
            raise ValueError("Number of intensities doesn't match the number of tiles")

        colors = [self.colormap.intensity_to_color(intensity) for intensity in intensities]

        for y in range(self.height_tiles):
            for x in range(self.width_tiles):
                color = colors[y * self.width_tiles + x]
                tile = Image.new("RGBA", (self.tile_width, self.tile_height), color)
                self.canvas.paste(tile, (x * self.tile_width, y * self.tile_height))

        if flip != None or not flip:
            try:
                # Apply vertical flip
                if flip[FlipEnum.VERTICAL.value]:
                    self.canvas = ImageOps.flip(self.canvas)

                # Apply horizontal flip
                if flip[FlipEnum.HORIZONTAL.value]:
                    self.canvas = ImageOps.mirror(self.canvas)
            except KeyError:
                pass

        return self.canvas


    def update_colorbar(self, colormap_min, colormap_max):
        ''' Updates range of the colormap. '''
        self.colormap = ColorIntensity(colormap_min, colormap_max)



''' Custom Color Intensity - clmap from matplotlib was way to computional heavy.'''
class ColorIntensity:

    def get_custom_color(self, input_value, num_steps):
        # Define the gradient colors from white to red
        gradient = [
            (1.0, 1.0, 1.0),  # White
            (1.0, 1.0, 0.0),  # Yellow
            (1.0, 0.5, 0.0),  # Orange
            (1.0, 0.0, 0.0)   # Red
        ]

        # Ensure the input value is within the range [0, 1]
        input_value = max(0.0, min(1.0, input_value))

        # Calculate the index and interpolation factor
        index = input_value * (len(gradient) - 1)
        lower_index = int(index)
        t = index - lower_index

        # Interpolate between the two gradient colors
        color_lower = gradient[lower_index]
        color_upper = gradient[min(lower_index + 1, len(gradient) - 1)]

        r = color_lower[0] * (1 - t) + color_upper[0] * t
        g = color_lower[1] * (1 - t) + color_upper[1] * t
        b = color_lower[2] * (1 - t) + color_upper[2] * t

        return (r, g, b)



    def __init__(self, minimum=0, maximum=1):
        self.colors  = []
        self.minimum = minimum
        self.maximum = maximum
        self.num_iterations = 20

        for i in range(self.num_iterations):
            color = self.get_custom_color(i/self.num_iterations, self.num_iterations)
            self.colors.append(color)

        # self.colors = [
        #     (1.0, 0.0, 0.0),  # Red
        #     (1.0, 0.5, 0.0),  # Orange
        #     (1.0, 1.0, 0.0),  # Yellow
        #     (0.0, 1.0, 0.0),  # Green
        #     (0.0, 1.0, 1.0),  # Cyan
        #     (0.0, 0.0, 1.0),  # Blue
        #     (0.5, 0.0, 1.0),  # Purple
        #     (1.0, 0.0, 1.0),  # Magenta
        #     (1.0, 0.0, 0.5),  # Pink
        #     (1.0, 0.0, 0.0),  # Red (again)
        # ]

    def intensity_to_color(self, intensity):
        # Scale the intensity to the range [0, 1] using a linear scale
        scaled_intensity = (intensity - self.minimum) / (self.maximum - self.minimum)
        scaled_intensity = max(0, min(1, scaled_intensity))  # Ensure it's within [0, 1]

        
        index = int(scaled_intensity * self.num_iterations)
        index = min(max(index, 0), self.num_iterations-1)  # Ensure index is within bounds
        r, g, b = self.colors[index]
        r = int(r * 255)
        g = int(g * 255)
        b = int(b * 255)
        a = 255 if intensity > 0 else 0

        return r,g,b,a

    def intensity_to_hex_color(self, intensity):
        r,g,b,_ = self.intensity_to_color(intensity)        
        hex_color = "#{:02X}{:02X}{:02X}".format(r, g, b)
        return hex_color
