import sys
sys.path.append('../Visualiser')

from appEnums  import FlipEnum, ConfigEnum, ConfigStrings, LimitsEnum
from main_dash import app
from marker    import MarkerProperty

import csv
import dash_leaflet as dl
import geojson
import json
import math
import numpy        as np

from dash             import html
from decimal          import localcontext, Decimal, ROUND_HALF_DOWN
from io               import StringIO
from shapely.geometry import Point, Polygon


"""
Parameters
----------

Returns
-------
    
"""

def check_range(missing_fields, limit_enum, data, range, unit, get_text_only=False):
        ''' GNSS in range < -180 ; 180 > '''
        # Do not check empty ones
        check_data = data[limit_enum.value] if type(limit_enum.value) == int else data[limit_enum.value.lower()]

        if check_data == None or check_data == "":
            return missing_fields if not get_text_only else None

        if float(check_data) < range[0] or \
           float(check_data) > range[1]:
            
            if not get_text_only:
                missing_fields.append(html.Div([html.B(f"{limit_enum.get_name()}: {limit_enum.string.title()}"), f" is not in range < {range[0]} , {range[1]} >{unit}.\n"]))
            else:
                return f"{limit_enum.string.title()} is not in range < {range[0]} , {range[1]} >{unit}.\n"

        return missing_fields if not get_text_only else None


def round_half_down(value):
    ''' Rounding numbers by 'round half down' as default python round() uses 'round half even'
        Note: simple int( value) does not work, for 64.50000000000003 returns 64 instead of 65 '''
    with localcontext() as ctx:
        ctx.rounding = ROUND_HALF_DOWN
        n = Decimal(value)
        return int( n.to_integral_value() )



def fillWithZeros(original_data, x_steps, y_steps, index_x_min, index_x_max, index_y_min, index_y_max):
    ''' Fill 2d graph with zeros, if not fully covered by original_data.
        Eg. 2d graph 100x100 - original_data should have 10000 values. If there is less data, which bounds are defined by index_x/y,
        the 'void' is filled with zeros.
        
        Original data          New data
         ----------           ----------
        |          |         |0000000000|
        |          |         |0000000000|
        |   xxxxxx |    ->   |000xxxxxx0|
        |   xxxxxx |         |000xxxxxx0|
        |   xxxxxx |         |000xxxxxx0|
        |          |         |0000000000|
         ----------           ----------
    '''
    grid = np.zeros((x_steps, y_steps))

    data_to_process = np.array(original_data)

    index_x_min = index_x_min if index_x_min >= 0 else 0
    index_y_min = index_y_min if index_y_min >= 0 else 0

    x_dim = index_x_max - index_x_min +1  # +1 as i want to have size
    y_dim = index_y_max - index_y_min +1
    
    # Fill the original data at the place defined by indexes
    try:
        grid[index_x_min:index_x_max+1, index_y_min:index_y_max+1] = data_to_process.reshape(x_dim, y_dim)
    except ValueError as e:
        app.logger.warning(f"Filling with zeros: {e}")
        app.logger.warning(f"Filling with zeros: {index_x_min}:{index_x_max+1}, {index_y_min}:{index_y_max+1} | xdim: {x_dim} | ydim: {y_dim}")

    # Flatten the grid into 1d array
    return grid.ravel()



def fillWithZeros3D(original_data, x_steps, y_steps, z_steps, index_x_min, index_x_max, index_y_min, index_y_max, index_z_min, index_z_max):
    ''' Same as fillWithZeros, but in 3d space. '''
    grid = np.zeros((x_steps, y_steps, z_steps))

    data_to_process = np.array(original_data)

    index_x_min = index_x_min if index_x_min >= 0 else 0
    index_y_min = index_y_min if index_y_min >= 0 else 0
    index_z_min = index_z_min if index_z_min >= 0 else 0

    x_dim = index_x_max - index_x_min + 1
    y_dim = index_y_max - index_y_min + 1
    z_dim = index_z_max - index_z_min + 1
    
    # Fill the original data at the place defined by indexes
    try:
        grid[index_x_min:index_x_max+1, index_y_min:index_y_max+1, index_z_min:index_z_max+1] = data_to_process.reshape(x_dim, y_dim, z_dim)
    except ValueError as e:
        app.logger.warning(f"Filling with zeros 3d: {e}")
        app.logger.warning(f"Filling with zeros 3d: {index_x_min}:{index_x_max+1}, {index_y_min}:{index_y_max+1} | xdim: {x_dim} | ydim: {y_dim}")

    # Flatten the grid into 1d array
    return grid.ravel()


def getJammingIndexes(fusion_config, first_limit, second_limit, density_index):
    ''' Returns indexes for (not only Map) Jamming - if no integration limits set, for 'resolution' 100x100 the indexes will be 0,99.
        However if integration limits present, indexes might be different, eg. lat 18,78, lon 15,99 for 'cutout' from the
        original data rectangle of 100x100.'''
    def get_step(minimal, maximal, n_steps):
        return (maximal - minimal) / n_steps

    def equation(value, cfg_min, step, is_checking_max):
        shouldFloor = False
        mid_div = 0.501 if is_checking_max else 0.5     # Important - if calculating min, we should round up, if then round down
        mid = (value - cfg_min - mid_div*step) / step

        # Exception - if min and max are the same, just opposite, floor the value instead of round, as the equation
        #   would return .5 and round up. But from the exeprience and testing, in this example we have to round down (floor)
        if value + cfg_min == 0:
            shouldFloor = True

        if shouldFloor:
            result = math.floor(mid)
        else:
            result = round(mid)

        if result < 0:
            result = 0

        return result

    first_min,  first_max,   first_n = prepareVars(first_limit .config.value, fusion_config[ConfigEnum.LIMITS.value], False, 'Config.Limit.One')
    second_min, second_max, second_n = prepareVars(second_limit.config.value, fusion_config[ConfigEnum.LIMITS.value], False, 'Config.Limit.Second')

    first_step  = get_step(first_min,   first_max,  first_n)
    second_step = get_step(second_min, second_max, second_n)

    density_data = fusion_config[ConfigEnum.INTEGRATION.value][ConfigEnum.JAMMER.value][ConfigEnum.DENSITY.value][density_index]

    first_local_min,   first_local_max, _ = prepareDensityVars(first_limit .config.value, density_data, False, 'Config.Limit.One')
    second_local_min, second_local_max, _ = prepareDensityVars(second_limit.config.value, density_data, False, 'Config.Limit.Two')

    first_local_min  = max( first_local_min,  first_min)
    first_local_max  = min( first_local_max,  first_max)
    second_local_min = max(second_local_min, second_min)
    second_local_max = min(second_local_max, second_max)

    first_i_min  = equation(first_local_min,   first_min, first_step,  False)
    first_i_max  = equation(first_local_max,   first_min, first_step,  True)
    second_i_min = equation(second_local_min, second_min, second_step, False)
    second_i_max = equation(second_local_max, second_min, second_step, True)

    return first_i_min, first_i_max, second_i_min, second_i_max


def getJammingIndex(fusion_config, limits, density_index):
    ''' Returns indexes for (not only Map) Jamming - if no integration limits set, for 'resolution' 100x100 the indexes will be 0,99.
        However if integration limits present, indexes might be different, eg. lat 18,78, lon 15,99 for 'cutout' from the
        original data rectangle of 100x100.'''
    def get_step(minimal, maximal, n_steps):
        return (maximal - minimal) / n_steps

    def equation(value, cfg_min, step, is_checking_max):
        mid = round_half_down((value - cfg_min - 0.5*step) / step)
        if mid < 0:
            mid = 0
        return mid

    minimum, maximum, steps = prepareVars(limits .config.value, fusion_config[ConfigEnum.LIMITS.value], False, 'Config.Limits.Unknown')

    step = get_step(minimum, maximum, steps)
    
    try:
        # vvvvvvvvvv might throw exception
        density_data = fusion_config[ConfigEnum.INTEGRATION.value][ConfigEnum.JAMMER.value][ConfigEnum.DENSITY.value][density_index]

        local_min, local_max, _ = prepareDensityVars(limits.config.value, density_data, False, 'Config.Limits.Unkwnow')
        
        local_min = max( local_min, minimum)
        local_max = min( local_max, maximum)
        
        index_min = equation(local_min, minimum, step, False)
        index_max = equation(local_max, minimum, step, True)

        # In case of GNSS which are originally in radians, convert them to degrees
        if limits == LimitsEnum.LATITUDE or limits == LimitsEnum.LONGITUDE:
            local_min = np.rad2deg(local_min)
            local_max = np.rad2deg(local_max)
        
        return index_min, index_max, local_min, local_max

    except IndexError as e:
        app.logger.warning(f"getJammingIndex(): Configration density has len {len(fusion_config[ConfigEnum.INTEGRATION.value][ConfigEnum.JAMMER.value][ConfigEnum.DENSITY.value])}, but should have one more. (density_index {density_index})")
        return None, None, None, None



def getMapJammingIndexes(fusion_config):
    ''' Returns indexes for Map Jamming - if no integration limits set, for 'resolution' 100x100 the indexes will be 0,99.
        However if integration limits present, indexes might be different, eg. lat 18,78, lon 15,99 for 'cutout' from the
        original rectangle of 100x100.'''
    def get_step(minimal, maximal, n_steps):
        return (maximal - minimal) / n_steps

    def equation(value, cfg_min, step):
        return round( (value - cfg_min - 0.5*step) / step)

    lat_min, lat_max, lat_n = prepareVars(ConfigStrings.LAT.value, fusion_config[ConfigEnum.LIMITS.value], False, 'Config.Limits.LAT')
    lon_min, lon_max, lon_n = prepareVars(ConfigStrings.LON.value, fusion_config[ConfigEnum.LIMITS.value], False, 'Config.Limits.LON')

    lat_step = get_step(lat_min, lat_max, lat_n)
    lon_step = get_step(lon_min, lon_max, lon_n)

    density_data = fusion_config[ConfigEnum.INTEGRATION.value][ConfigEnum.JAMMER.value][ConfigEnum.DENSITY.value][0]

    lat_local_min, lat_local_max, _ = prepareDensityVars(ConfigStrings.LAT.value, density_data, False, 'Config.Limits.LAT')
    lon_local_min, lon_local_max, _ = prepareDensityVars(ConfigStrings.LON.value, density_data, False, 'Config.Limits.LON')

    lat_i_min = equation(lat_local_min, lat_min, lat_step)
    lat_i_max = equation(lat_local_max, lat_min, lat_step)
    lon_i_min = equation(lon_local_min, lon_min, lon_step)
    lon_i_max = equation(lon_local_max, lon_min, lon_step)

    return lat_i_min, lat_i_max, lon_i_min, lon_i_max

def getValue(data, var, value, default, checkForZero, limit_name, limit_value, check_min=None, check_max=None):
    try:
        val = data[var][value]
    except KeyError:
        val = default
        app.logger.warning(f"KeyError in Field - {limit_name} : {limit_value} - replaced with 'default' value {default}")
    if val == None:
        val = default
        app.logger.warning(f"None value in Field - {limit_name} : {limit_value} - replaced with 'default' value {default}")
    if checkForZero and val == 0:
        val = default
        app.logger.warning(f"Detected zero in non-zero Field - {limit_name} : {limit_value} - replaced with 'default' value {default}")
    
    # If necesarry, check for min/max and if below/over, make it default
    if check_min != None and check_min:
        if val < default:
            val = default

    if check_max != None and check_max:
        if val > default:
            val = default
    return val


def prepareVars(var, data, init_empty, name):
    if init_empty:
        return -1e30, 1e30, 100
    else:
        var_min = getValue(data,var,ConfigStrings.MIN  .value, -1e30, False, name, 'MIN')
        var_max = getValue(data,var,ConfigStrings.MAX  .value,  1e30, False, name, 'MAX')
        var_n   = getValue(data,var,ConfigStrings.STEPS.value, 100,   True,  name, 'STEPS')
        return var_min, var_max, var_n

def prepareDensityVars(var, data, init_empty, name):
    if init_empty:
        return -1e30, 1e30, True        # Toggle default True, as is all False, we can get 20 000 000 values in no time - not really efficient for calculations
    else:
        var_min = getValue(data,var,ConfigStrings.MIN          .value, -1e30, False,  name, 'MIN')
        var_max = getValue(data,var,ConfigStrings.MAX          .value,  1e30, False,  name, 'MAX')
        var_tog = getValue(data,var,ConfigStrings.INTEGRATE_OUT.value, True,  False,  name, 'INTEGRATE')
        return var_min, var_max, var_tog

def prepareDensityVarsSpecViz(var, mins, maxs, data, init_empty, name):
    ''' Initial for Specific Vizualisation - if nothing set from previous app runtime, do not fill with 1e30, but with actual Limit min max. '''
    if init_empty:
        return mins, maxs, True        # Toggle default True, as is all False, we can get 20 000 000 values in no time - not really efficient for calculations
    else:
        var_min = getValue(data,var,ConfigStrings.MIN          .value, mins, False,  name, 'MIN',      check_min=True)
        var_max = getValue(data,var,ConfigStrings.MAX          .value, maxs, False,  name, 'MAX',      check_max=True)
        var_tog = getValue(data,var,ConfigStrings.INTEGRATE_OUT.value, True, False,  name, 'INTEGRATE')
        return var_min, var_max, var_tog

def prepareProbabilityVars(mightBePresent, var, data, init_empty):
    if init_empty:
        return None, None
    else:
        var_min = data[var][ConfigStrings.MINIMA.value]
        var_max = data[var][ConfigStrings.MAXIMA.value]
        return var_min, var_max

def updateVars(var, checkedLimit, data, init_empty, name):
    _, _, steps = prepareVars(var, data, init_empty, name)
    data[var][checkedLimit] = steps
    return data


def checkLimits(limits):
    ''' Checks if limits are within bounds - now only checks "steps" if they are bigger than 0. '''
    limits = updateVars(ConfigStrings.LAT.value,   ConfigStrings.STEPS.value, limits, False, 'checkLimits() Config.LAT')
    limits = updateVars(ConfigStrings.LON.value,   ConfigStrings.STEPS.value, limits, False, 'checkLimits() Config.LON')
    limits = updateVars(ConfigStrings.HGT.value,   ConfigStrings.STEPS.value, limits, False, 'checkLimits() Config.HGT')
    limits = updateVars(ConfigStrings.POWER.value, ConfigStrings.STEPS.value, limits, False, 'checkLimits() Config.PWR')

    return limits


def isMinMaxLimit(value) -> bool:
    ''' Checks if given value is at the limit or not. Value might be string or float. '''
    if value == None:
        return value

    limit = 1e30
    if type(value) == str:
        value = float(value)

    return value >= limit or value <= (-1*limit)


def convertValueDeg2Rad(shouldConvert, value):
    ''' Converts value to rad if necessary. '''
    if value == None:
        return value

    if shouldConvert:
        if type(value) == str:
            value = float(value)
        if not isMinMaxLimit(value):
            return np.deg2rad(value)
    return value



def getDictData(data, key, isList):
    ''' Reads list/dict based on key. '''
    try:
        result = data[key]
        return result
    except KeyError as e:
        app.logger.warning(f'Reading data[{key}] error: {e}')
        return None



def split_interval(min_value, max_value, num_steps):
    ''' Splits interval into num_steps values.'''
    step_size = (max_value - min_value) / (num_steps)
    return [min_value + i * step_size for i in range(num_steps+1)]


def prepare_geodata(lon_min, lon_max, lon_steps, lat_min, lat_max, lat_steps):
    ''' Prepares Geodata - takes the interval from min, max, splits it into 'steps' tiles
        and creates data_points representing rectangles with coords - left top corner
        and bottom right corner. '''
    lon_list = split_interval(lon_min, lon_max, lon_steps)
    lat_list = split_interval(lat_min, lat_max, lat_steps)
    data_points = create_lat_lon_tuples(lat_list, lon_list)
    return data_points


def create_lat_lon_tuples(latitudes, longitudes):
    ''' From 1d list of latitude, 1d list longitude
        creates 'tuples' for each lat&lon. '''
    tuples = []
    for lat_idx, lat in enumerate(latitudes):
        for lon_idx, lon in enumerate(longitudes):
            tuples.append((lat, lon))
    return tuples



def create_geojson_polygons(tuples, num_steps_lon, num_steps_lat):
    ''' From triplets array, creates GeoJSON polygon structure.'''
    polygons = []

    num_steps_lon += 1      # +1 as we have n_steps for tiles, but tile has start and end
    num_steps_lat += 1

    # For n_steps == 2 -> we have two tiles. But that means three coords in total.
    #
    #     0 --- 1 --- 2
    #     |     |     |
    #     |  A  |  B  |
    #     |     |     |
    #     3 --- 4 --- 5
    #     |     |     |
    #     |  C  |  D  |
    #     |     |     |
    #     6 --- 7 --- 9

    for i in range(num_steps_lat - 1):
        for j in range(num_steps_lon - 1):
            idx1 = i * num_steps_lon + j
            idx2 = idx1 + 1
            idx3 = (i + 1) * num_steps_lon + j
            idx4 = idx3 + 1
            
            lat1, lon1 = tuples[idx1]
            lat2, lon2 = tuples[idx2]
            lat3, lon3 = tuples[idx3]
            lat4, lon4 = tuples[idx4]

            coordinates = [
                (lon1, lat1),
                (lon2, lat2),
                (lon4, lat4),
                (lon3, lat3),
                (lon1, lat1)
            ]

            polygon = geojson.Polygon([coordinates])

            feature = geojson.Feature(geometry=polygon, id=f"{i}_{j}")
            polygons.append(feature)

    feature_collection = geojson.FeatureCollection(polygons)
    return feature_collection


def createDashRectangles(tuples, num_steps_lon, num_steps_lat):
    ''' Creates an array of Dash Leaflet Rectangles (dl.Rectangle), which will later change only color. '''
    rectangles = []
    for i in range(num_steps_lat - 1):
        for j in range(num_steps_lon - 1):
            idx1 = i * num_steps_lon + j
            # idx2 = idx1 + 1
            idx3 = (i + 1) * num_steps_lon + j
            idx4 = idx3 + 1
            
            lat1, lon1 = tuples[idx1]
            lat4, lon4 = tuples[idx4]

            rectangles.append(dl.Rectangle(bounds=[ [lat1,lon1], [lat4,lon4]], fillColor="#ff00ff", fillOpacity=0.8, stroke=False))

    return rectangles


def update_geojson_polygon_w_intensities(geodata, new_intensities):
    ''' Updating GeoJSON intensities. Original GeoData were created by iterationg
        over lat and over lon. This update keeps the order. '''
    for i,feature in enumerate(geodata['features']):
        feature['sensitivity'] = new_intensities[i]
    return geodata


def create_csv_from_triplets_with_id(triplets, num_steps_lon, num_steps_lat):
    ''' Creates CSV like style string with column ID and Intensity,
        so the GeoJSON data can be tied in folium.Choropleth() map. '''
    csv_string = StringIO()
    fieldnames = ['id', 'sensitivity']
    writer = csv.DictWriter(csv_string, fieldnames=fieldnames)
    writer.writeheader()

    for i in range(num_steps_lat - 1):
        for j in range(num_steps_lon - 1):
            idx1 = i * num_steps_lon + j

            lat1, lon1, sensitivity1 = triplets[idx1]
            writer.writerow({'id': f"{i}_{j}", 'sensitivity': sensitivity1})

    return csv_string.getvalue()


def getCallBackContextIndex(callback_context, index=0):
    """ Returns index of element if pattern-matching callback is used.

    Parameters
    ----------
        callback_context

    Returns
    -------
        int - index of element from which was callback initialised
    """
    input_id = callback_context.triggered[index]["prop_id"].split(".")[0]
    return json.loads(input_id)["index"]


def isPointInsideTile(tile_coordinates: list, point_coordinates: list) -> bool:
    """ Checks with Shapely if point is inside a polygon.
    Does not really matter what unit are the coords, but both parameters has to have the same unit.

    Parameters
    ----------
        tile_coordinates  - list of coordinates of the polygon - list of list of vertex coords
        point_coordinates - vertex coords

    Returns
    -------
        bool - if point is inside the polygon
    """
    def isEmpty(point_coord):
        return point_coord == None or point_coord == ''

    if isEmpty(point_coordinates[0]) or isEmpty(point_coordinates[1]):
        return None

    tile_polygon = Polygon(tile_coordinates[0])
    point = Point(point_coordinates)
    return tile_polygon.contains(point)


def getPolygonCoords(config_mins, config_maxs):
    ''' Returns list of coordinates to be used in Shapely.Polygon. '''

    lat_min = config_mins[LimitsEnum.LATITUDE .value]
    lon_min = config_mins[LimitsEnum.LONGITUDE.value]
    lat_max = config_maxs[LimitsEnum.LATITUDE .value]
    lon_max = config_maxs[LimitsEnum.LONGITUDE.value]

    # We need to coords in zig zak
    #
    #  4 --- 3
    #  |     |
    #  |     |   -> 1,2,4,3
    #  |     |
    #  1 --- 2

    coordinates = [
        (lon_min, lat_min),     # 1
        (lon_max, lat_min),     # 2
        (lon_max, lat_max),     # 3
        (lon_min, lat_max),     # 4
        (lon_min, lat_min)      # 1
    ]
    return coordinates


def getTileIndexOfPoint(geodata, point_coordinates: list) -> int:
    """ Returns index of tile in which the point_coordinates is located.
        If no tile is found, None is returned.

            geodata - geojson.feature.FeatureCollection

        Returns
        -------
        int - if point found, index of tile in geodata structure 
              if not found, None
    """
    res = None
    for i, g in enumerate(geodata.features):
        tile_coordinates = g.geometry.coordinates
        result = isPointInsideTile(tile_coordinates, point_coordinates)
        if result:
            # id of the tile:
            res = i
    return res


def updateMarkersWithTileIndex(geodata, marker_data) :
    """ Uodates all markers in @param marker_data with tileIndex if given marker
        is located inside some of the tiles.

    Parameters
    ----------
    
    Returns
    -------
        marker_data - updated inner markers list with indexes of tiles
        result      - list of indexes with device in it
    """
    if geodata != None and len(geodata) > 0:
        tiles_indexes = []
        for i, g in enumerate(geodata.features):
            tile_coordinates = g.geometry.coordinates

            for marker_id, marker in enumerate(marker_data):
                point_coordinates = [marker['lon'], marker['lat']]

                result = isPointInsideTile(tile_coordinates, point_coordinates)
                if result:
                    marker[MarkerProperty.TILE_INDEX.value] = i
                    tiles_indexes.append(i)
    else:
        marker_data, tiles_indexes = None, None

    return marker_data, tiles_indexes


def updateOneMarkerWithTileIndex(geodata, marker):
    """ 

    Parameters
    ----------
    
    Returns
    -------
        marker_data - updated inner markers list with indexes of tiles
        result      - list of indexes with device in it
    """
    marker[MarkerProperty.TILE_INDEX.value] = None  # by default when marker is not present in map jamminng data
    for i, g in enumerate(geodata.features):
        tile_coordinates = g.geometry.coordinates

        point_coordinates = [marker['lon'], marker['lat']]

        result = isPointInsideTile(tile_coordinates, point_coordinates)
        if result:
            marker[MarkerProperty.TILE_INDEX.value] = i
            return marker, i

    return marker, None


def getTileByIndex(geodata, search_index: int):
    """ Returns geodata feature with param serach_index.

    Parameters
    ----------
        geodata - geojson.feature.FeatureCollection
        search_index - int

    Returns
    -------
        geojson.feature.Feature
    """
    return geodata.features[search_index]


def getFlipImageVars(lon_min, lon_max, lat_min, lat_max):
    ''' Determine if generated image should be flip along some axis.
        As image is generated usually from top left corner, but GPS coords for
        central Europe usually starts from bottom left corner. Rule of thumb is, 
        positive latitude means flip the image vertically, negative longitude means
        flip the image horizontally.
    '''
    return {
        FlipEnum.VERTICAL.value:   lon_max >= 0.0,
        FlipEnum.HORIZONTAL.value: lat_max <  0.0
    }



import base64
import io
import pandas as pd
from dash import no_update
def loadCsvFile(contents, col_limit, message_wrong_file, newRowFunc):
    ''' Loads CSV file and returns new data together with list of active rows. '''
    content_type, content_string = contents.split(',')
    decoded_content = io.StringIO(base64.b64decode(content_string).decode('utf-8'))

    try:
        df = pd.read_csv(decoded_content, sep=';')

        if len(df.columns) != col_limit:
            app.logger.warning(f"{message_wrong_file} | CSV file has {df.columns}, but should have {col_limit} columns.")
            return None, None

        num_of_rows       = 0
        new_data          = []
        new_selected_rows = []

        for item in df.to_dict('records'):
            new_row = newRowFunc( *[item[df.columns[x]] for x in range(col_limit)] )
            new_data.append(new_row)

            if item[df.columns[0]]:
                new_selected_rows.append(num_of_rows)

            num_of_rows += 1

        return new_data, new_selected_rows

    except Exception as e:
        app.logger.warning(f"Error reading .csv file: {e}")
        return no_update, no_update



import dash_bootstrap_components as dbc
def getModal_wrongCsvFile(identifier, fileType):
    ''' Returns Modal with Warning abour file not correct as CSV. '''
    return dbc.Modal(
        id=identifier,
        children=[
            dbc.ModalHeader('Warning'),
            dbc.ModalBody(f'Selected file is not a correct {fileType} .CSV file.'),
            dbc.ModalFooter(
                dbc.Button('Close', id=f'close-{identifier}', n_clicks=0, color='primary')
            )
        ]
    )



import dash_bootstrap_components as dbc
def getModal_sensorNotValid(identifier):
    ''' Returns Modal with Warning Sensor not valid enry. '''
    return dbc.Modal(
        id=identifier,
        children=[
            dbc.ModalHeader('Warning'),
            dbc.ModalBody(
                [
                    f'Sensor Row(s) below have invalid values:',
                    html.Div(id=f'sensor-modal-{identifier}')
                ]),
            dbc.ModalFooter(
                dbc.Button('Close', id=f'close-{identifier}', n_clicks=0, color='primary')
            )
        ]
    )



import dash_bootstrap_components as dbc
def getModal_deviceNotValid(identifier):
    ''' Returns Modal with Warning Sensor not valid enry. '''
    return dbc.Modal(
        id=identifier,
        children=[
            dbc.ModalHeader('Warning'),
            dbc.ModalBody(
                [
                    f'Device Row(s) below have invalid values:',
                    html.Div(id=f'devices-modal-{identifier}')
                ]),
            dbc.ModalFooter(
                dbc.Button('Close', id=f'close-{identifier}', n_clicks=0, color='primary')
            )
        ]
    )



import dash_bootstrap_components as dbc
def getModal_rmsNotValid(identifier):
    ''' Returns Modal with Warning RMS not valid enry. '''
    return dbc.Modal(
        id=identifier,
        children=[
            dbc.ModalHeader('Warning'),
            dbc.ModalBody(
                [
                    f'RMS Row(s) below have invalid values:',
                    html.Div(id=f'rms-modal-{identifier}')
                ]),
            dbc.ModalFooter(
                dbc.Button('Close', id=f'close-{identifier}', n_clicks=0, color='primary')
            )
        ]
    )



def convertToNumeric(value_str):
    if isinstance(value_str, (int, float)):
        return value_str
    else:

        if value_str.isdigit():
            return int(value_str)
        else:
            try:
                return float(value_str)
            except ValueError:
                print("Invalid numeric format")
                return None



def getSliderStep_byRads(mins, maxs, steps):
    ''' Calculates step for slider for GNSS limits - the whole app is using Degress for GNSS limits,
        but when calculating step for sliders, ever-so-slightly-small error occurs when calculating step
        with values in degrees, which can cause unexpected results for the user - for example:
        - Configuration has Lat Lon steps set for 10
        - Map sliders are set so min and max are next to each other (not overlapping, the smallest gap possible)
        - In this setup, user is expecting to get resulting tiles 2x2 - but usually got 2x3 (3 in Latitude).

        BUT MAYBE, this is not the fix. But it looks like it works.
    '''
    r_mins         = np.deg2rad(mins)
    r_maxs         = np.deg2rad(maxs)
    r_slider_step  = (r_maxs - r_mins) / steps
    slider_step    = math.ceil(np.rad2deg(r_slider_step)*1e5) / 1e5
    return slider_step



def is_integer(mark_val):
    return abs(mark_val-round(mark_val)) < 1e-3



def attempt_create_int(mark_val):
    ''' Tries to create integer from input value - if not successfull, returns float. '''
    if is_integer(mark_val):
        return int(mark_val)
    return mark_val
