import sys
sys.path.append('../Visualiser')

from main_dash import app
from appEnums  import ConfigStrings, RedisEnum, LimitsEnum, ProcessingEnum
from utils     import convertValueDeg2Rad, check_range, getCallBackContextIndex, getSliderStep_byRads, prepareVars

from dash            import Output, Input, State, dcc, html, callback_context, ALL, MATCH, no_update
from dash.exceptions import PreventUpdate
from enum            import Enum

import dash_bootstrap_components as dbc
import dash_daq                  as daq
import numpy                     as np

# Globals for Configuration
global global_config_apply_revert_active   # Apply / Revert disabled button

class IntegrationMode(Enum):
    NONE         = (0, "none",         "None")
    EVERY_UPDATE = (1, "every_update", "Every update")
    PERIOD_TIME  = (2, "period_time",  "Period Time")
    PERIOD_MEAS  = (3, "period_meas",  "Period Meas")

    def __new__(cls, value, string_value, label_value):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.string  = string_value
        obj.label   = label_value
        return obj

    def __str__(self):
        return self.string



def getGnssLimits(config_mins, config_maxs, config_steps):
    ''' Returns jamming map limits (for Config) and bounds (for rendering ImageOverlay). '''
    lon_min = config_mins[LimitsEnum.LONGITUDE.value]
    lon_max = config_maxs[LimitsEnum.LONGITUDE.value]
    lat_min = config_mins[LimitsEnum.LATITUDE.value]
    lat_max = config_maxs[LimitsEnum.LATITUDE.value]

    bounds = [[lat_min, lon_min], [lat_max, lon_max]]
    limits = [  lon_min,
                lon_max,
                config_steps[LimitsEnum.LONGITUDE.value],
                lat_min,
                lat_max,
                config_steps[LimitsEnum.LATITUDE.value]]
    return limits, bounds


def layour_configuration_tab(limits_init_data, processing_init_data, redis_init_data):
    # Init global variables
    global global_config_apply_revert_active
    global_config_apply_revert_active = False

    def prepareRedis(data, key):
        return data[key] if data != None and data[key] != None else ""

    init_empty = limits_init_data == None
    lat_min, lat_max, lat_n = prepareVars(ConfigStrings.LAT  .value, limits_init_data, init_empty, 'Limit.Config.LAT')
    lon_min, lon_max, lon_n = prepareVars(ConfigStrings.LON  .value, limits_init_data, init_empty, 'Limit.Config.LON')
    h_min,   h_max,   h_n   = prepareVars(ConfigStrings.HGT  .value, limits_init_data, init_empty, 'Limit.Config.HGT')
    P_min,   P_max,   P_n   = prepareVars(ConfigStrings.POWER.value, limits_init_data, init_empty, 'Limit.Config.PWR')

    redis_ip   = prepareRedis(redis_init_data, RedisEnum.IP  .string)
    redis_port = prepareRedis(redis_init_data, RedisEnum.PORT.string)

    if processing_init_data != None:        
        r       = processing_init_data["record"]
        r_path  = processing_init_data["record_path"]
        r_start = processing_init_data["replay_start"]
        r_end   = processing_init_data["replay_stop"]
        r_speed = processing_init_data["replay_speed"]
        i_mode  = processing_init_data["integration_mode"]
        i_value = processing_init_data["integration_mode_value"]
        undul   = processing_init_data["undulation"]
        time_d  = processing_init_data["time_degradation"]
    else:
        r, r_path, r_start, r_end, r_speed, i_mode, i_value, undul, time_d = False, *([""])*8

    def createRow(name, v_min, v_max, v_n, index):
        return html.Tr([
            html.Td(name, style={'text-align': 'right'}),
            html.Td([dcc.Input(v_min, type="number", id={"type": "config_mins", "index": index}),         dcc.Store(data=v_min, id={"type":"hidden_config_mins", "index": index})]),
            html.Td([dcc.Input(v_max, type="number", id={"type": "config_maxs", "index": index}),         dcc.Store(data=v_max, id={"type":"hidden_config_maxs", "index": index})]),
            html.Td([dcc.Input(v_n,   type="number", id={"type": "config_step", "index": index}, step=1), dcc.Store(data=v_n,   id={"type":"hidden_config_step", "index": index})])
        ])

    def createProcRow(name, value, index, input_type, step):
        return html.Tr([
                html.Td(name),
                html.Td([dcc.Input(value, id={"type": "processings", "index": index}, type=input_type, step=step), dcc.Store(data=value, id={"type":"hidden_processings", "index": index})])
            ])

    gnssLimits = [lon_min, lon_max, lon_n, lat_min, lat_max, lat_n]

    return gnssLimits, html.Div([
        html.Br(),
        html.H2("Limits", style={'text-align': 'center'}),
        html.Table([
            html.Tr([
                html.Th(" ",    style={'text-align': 'right'}),
                html.Th("Min",  style={'text-align': 'center'}),
                html.Th("Max",  style={'text-align': 'center'}),
                html.Th("Number of Steps", style={'text-align': 'center'})
            ]),
            createRow( "Latitude [deg]", lat_min, lat_max, lat_n, LimitsEnum.LATITUDE .value),
            createRow("Longitude [deg]", lon_min, lon_max, lon_n, LimitsEnum.LONGITUDE.value),
            createRow(     "Height [m]", h_min,   h_max,   h_n,   LimitsEnum.HEIGHT   .value),
            createRow(    "Power [dBm]", P_min,   P_max,   P_n,   LimitsEnum.POWER    .value),
        ], style={'margin': 'auto'}),

        html.Br(),
        html.H2("Processing", style={'text-align': 'center'}),
        html.Table([
            html.Tr([
                html.Td(ProcessingEnum.RECORD.label),
                html.Td([daq.ToggleSwitch(id={"type": "processings", "index": ProcessingEnum.RECORD.value}, value=r, color="green"), dcc.Store(data=r, id={"type":"hidden_processings", "index": ProcessingEnum.RECORD.value})])
            ]),
            createProcRow(ProcessingEnum.PATH.label,  r_path,  ProcessingEnum.PATH.value,  "text",   None),
            html.Tr([
                html.Td(ProcessingEnum.REPLAY.label),
                html.Td([daq.ToggleSwitch(id={"type": "processings", "index": ProcessingEnum.REPLAY.value}, value=r, color="green"), dcc.Store(data=r, id={"type":"hidden_processings", "index": ProcessingEnum.REPLAY.value})])
            ]),
            createProcRow(ProcessingEnum.START.label, r_start, ProcessingEnum.START.value, "number", 1),
            createProcRow(ProcessingEnum.END.label,   r_end,   ProcessingEnum.END.value,   "number", 1),
            createProcRow(ProcessingEnum.SPEED.label, r_speed, ProcessingEnum.SPEED.value, "number", 1),
            html.Tr([
                html.Td(ProcessingEnum.MODE.label),
                html.Td([
                    dcc.Dropdown(
                        id={"type": "processings", "index": ProcessingEnum.MODE.value},
                        options=[{'label': mode.label, 'value': mode.string} for mode in IntegrationMode],
                        value=i_mode,
                        clearable=False
                    ), dcc.Store(data=i_mode, id={"type":"hidden_processings", "index": ProcessingEnum.MODE.value})
                    ]
                )
            ]),
            createProcRow(ProcessingEnum.M_VALUE.label, i_value, ProcessingEnum.M_VALUE.value, "number", 1),
            createProcRow(ProcessingEnum.UNDUL.label,   undul,   ProcessingEnum.UNDUL.value,   "number", 1),
            createProcRow(ProcessingEnum.TIME_D.label,  time_d,  ProcessingEnum.TIME_D.value,  "number", 'any')
        ], style={'margin': 'auto'}),
        
        html.Br(),
        html.H2("Redis", style={'text-align': 'center'}),
        html.Table([
            html.Tr([
                html.Td("IP"),
                html.Td("Port")
            ]),
            html.Tr([
                html.Td([dcc.Input(redis_ip,   type="text",   id={"type": "redis", "index": RedisEnum.IP.value}          ), dcc.Store(data=redis_ip,   id={"type":"hidden_redis", "index": RedisEnum.IP.value})]),
                html.Td([dcc.Input(redis_port, type="number", id={"type": "redis", "index": RedisEnum.PORT.value}, step=1), dcc.Store(data=redis_port, id={"type":"hidden_redis", "index": RedisEnum.PORT.value})]),
                html.Td("...", id='redis_connected', style={'display':'none'})
            ])
        ], style={'margin': 'auto'}),
        
        # Buttons
        html.Br(),
        html.Div([
            html.Button('Apply / Restart', id='configuration_save_config_file',   style={'margin':'0.5em'}),                 # Save is always enabled
            html.Button('Revert Changes',  id='configuration_revert_config_file', style={'margin':'0.5em'}, disabled=True),   # Revert enabled only if changes present
        ], style={'margin': 'auto', 'text-align': 'center'}),

        # Info screen about file was saved
        html.Div(id='configuration_save_config_file_output', style={'text-align': 'center', 'margin': '1em'}),
        dcc.Interval(id='configuration_clear_output_interval', interval=5000, n_intervals=0),  # 5000ms = 5 seconds

        # Modal for warning of some of the fields are empty
        dbc.Modal([
            dbc.ModalHeader("Alert"),
            dbc.ModalBody([
                    "Please fill in all input fields correctly:",
                    html.Div(id=f'modal_configuration_div')
                ]),
            dbc.ModalFooter(
                dbc.Button("Close", id="close-modal", className="ml-auto")
            ),
        ], id='modal_configuration'),

        # Timer for 'restarting' sw fusion
        dcc.Interval(id="configuration_timer", interval=1000, n_intervals=0, disabled=True)
    ])


def isEverythingFilled(config_mins, config_maxs, config_steps, processings, redis):
    ''' None of the fields in configuration can be empty - if some is missing, Modal warning is raised. '''

    missing_fields = []

    # Some values have limited range
    missing_fields = check_range(missing_fields, LimitsEnum.LATITUDE,  config_mins, [-180,180], unit=" [deg]")
    missing_fields = check_range(missing_fields, LimitsEnum.LATITUDE,  config_maxs, [-180,180], unit=" [deg]")
    missing_fields = check_range(missing_fields, LimitsEnum.LONGITUDE, config_mins, [-180,180], unit=" [deg]")
    missing_fields = check_range(missing_fields, LimitsEnum.LONGITUDE, config_maxs, [-180,180], unit=" [deg]")

    for i,item in enumerate(LimitsEnum):
        if item.config != None:     # LimitsEnum.GRAPH_PROBAB_DENSITY is not part of the 'config_step', without this IndexError
            missing_fields = check_range(missing_fields, item, config_steps, [1,1000], unit=" steps")

    missing_fields = check_range(missing_fields, RedisEnum.PORT, redis, [1,65535], unit="")

    # Check min is really smaller then max
    for i in range(len(LimitsEnum)-1):
        if config_mins[i] == None or config_maxs[i] == None:
            continue
        if config_mins[i] > config_maxs[i]:
            missing_fields.append(html.Div([html.B(f"{LimitsEnum.get_name()}: {LimitsEnum.get_enum_by_id(i).string}"), f" - Min is bigger than Max.\n"]))
        if config_mins[i] == config_maxs[i]:
            missing_fields.append(html.Div([html.B(f"{LimitsEnum.get_name()}: {LimitsEnum.get_enum_by_id(i).string}"), f" - Min and Max cannot be same value.\n"]))

    # Check emptiness
    for i,item in enumerate(config_mins):
        if item == None or item == "":
            missing_fields.append(html.Div([html.B(f"{LimitsEnum.get_name()}: {LimitsEnum.get_enum_by_id(i).string}"), f" - Min is empty.\n"]))
    for i,item in enumerate(config_maxs):
        if item == None or item == "":
            missing_fields.append(html.Div([html.B(f"{LimitsEnum.get_name()}: {LimitsEnum.get_enum_by_id(i).string}"), f" - Max is empty.\n"]))
    for i,item in enumerate(config_steps):
        if item == None or item == "":
            missing_fields.append(html.Div([html.B(f"{LimitsEnum.get_name()}: {LimitsEnum.get_enum_by_id(i).string}"), f" - Step is empty or not natural number.\n"]))
    for i,item in enumerate(processings):
        if item == None or item == "":
            missing_fields.append(html.Div([html.B(f"{ProcessingEnum.get_name()}: {ProcessingEnum.get_enum_by_id(i).label}"), f" is empty.\n"]))

    return missing_fields


def getRedisConfig(redis):
    return {
        RedisEnum.IP.string:   redis[RedisEnum.IP.value],
        RedisEnum.PORT.string: redis[RedisEnum.PORT.value],
    }



def getProcessingConfig(processings):
    return {
        "record":                        processings[ProcessingEnum.RECORD .value],
        "record_path":                   processings[ProcessingEnum.PATH   .value],
        "replay_start":              int(processings[ProcessingEnum.START  .value]),
        "replay_stop":               int(processings[ProcessingEnum.END    .value]),
        "replay_speed":            float(processings[ProcessingEnum.SPEED  .value]),
        "integration_mode":              processings[ProcessingEnum.MODE   .value],
        "integration_mode_value":  float(processings[ProcessingEnum.M_VALUE.value]),
        "undulation":              float(processings[ProcessingEnum.UNDUL  .value]),
        "time_degradation":        float(processings[ProcessingEnum.TIME_D .value])
    }

def getLimitsConfig(config_mins, config_maxs, config_steps, convertUnits: bool) -> dict:
    ''' Returns JSON structure for limits.
    
    Parameters
    ----------
        shouldConvertUnits - lat & lot are in degrees, but Fusion needs them in rad
                             h & Power stays the same        
    '''
    return {
        "lon": {
            "min":     convertValueDeg2Rad(convertUnits, config_mins [LimitsEnum.LONGITUDE.value]),
            "max":     convertValueDeg2Rad(convertUnits, config_maxs [LimitsEnum.LONGITUDE.value]),
            "n_steps":                                   config_steps[LimitsEnum.LONGITUDE.value],
        },
        "lat": {
            "min":     convertValueDeg2Rad(convertUnits, config_mins [LimitsEnum.LATITUDE.value]),
            "max":     convertValueDeg2Rad(convertUnits, config_maxs [LimitsEnum.LATITUDE.value]),
            "n_steps":                                   config_steps[LimitsEnum.LATITUDE.value],
        },
        "h": {
            "min":     config_mins [LimitsEnum.HEIGHT.value],
            "max":     config_maxs [LimitsEnum.HEIGHT.value],
            "n_steps": config_steps[LimitsEnum.HEIGHT.value],
        },
        "P": {
            "min":     config_mins [LimitsEnum.POWER.value],
            "max":     config_maxs [LimitsEnum.POWER.value],
            "n_steps": config_steps[LimitsEnum.POWER.value],
        }
    }


''' Updates min,max limits of range slider for integration Limits, based on value in Input(s) '''
@app.callback(
    Output({'type': 'integrate_view_slider', 'index': MATCH}, 'min',      allow_duplicate=True),
    Output({'type': 'integrate_view_slider', 'index': MATCH}, 'max',      allow_duplicate=True),
    Output({"type": "integrate_view_slider", "index": MATCH}, "value",    allow_duplicate=True),
    Output({'type': 'integrate_view_slider', 'index': MATCH}, 'marks',    allow_duplicate=True),
    Output({'type': 'integrate_view_slider', 'index': MATCH}, 'step',     allow_duplicate=True),
    Output({'type': 'integrate_view_slider', 'index': MATCH}, 'pushable', allow_duplicate=True),

    Output({'type': 'spec_vis_slider',       'index': MATCH}, 'min',      allow_duplicate=True),
    Output({'type': 'spec_vis_slider',       'index': MATCH}, 'max',      allow_duplicate=True),
    Output({"type": "spec_vis_slider",       "index": MATCH}, "value",    allow_duplicate=True),
    Output({'type': 'spec_vis_slider',       'index': MATCH}, 'marks',    allow_duplicate=True),
    Output({'type': 'spec_vis_slider',       'index': MATCH}, 'step',     allow_duplicate=True),
    Output({'type': 'spec_vis_slider',       'index': MATCH}, 'pushable', allow_duplicate=True),

    Output({'type': 'config_mins',           'index': MATCH}, 'value'),
    Output({'type': 'config_maxs',           'index': MATCH}, 'value'),
    Output({'type': 'config_step',           'index': MATCH}, 'value'),

    Input ({'type': 'hidden_config_mins',    'index': MATCH}, 'data'),
    Input ({'type': 'hidden_config_maxs',    'index': MATCH}, 'data'),
    Input ({'type': 'hidden_config_step',    'index': MATCH}, 'data'),
)
def updateSlidersMinMax(mins, maxs, steps):
    # Due to circular imports defined here
    from views.visualisation import createSlider_marks

    if steps <= 0:
        return [no_update]*5

    index_number = getCallBackContextIndex(callback_context)
    
    # In case of GNSS Limits, step has to be calculated from original (and SW Fusion used) Radians, not from Degrees.
    if index_number == LimitsEnum.LATITUDE.value or index_number == LimitsEnum.LONGITUDE.value:
        slider_step = getSliderStep_byRads(mins, maxs, steps)
    else:
        slider_step  = round( (maxs - mins) / steps, 5)

    # Update mins and max in Specific Visualisation as well - to avoid triple check of values already present there, if
    #   it was same as limit, smaller etc. we will chenge it to limit itself.
    can_be_int = LimitsEnum.get_enum_by_id(index_number).can_be_int
    return mins, maxs, [mins,maxs], createSlider_marks(mins, maxs, slider_step, steps, can_be_int), slider_step, slider_step/1.5, \
           mins, maxs, [mins,maxs], createSlider_marks(mins, maxs, slider_step, steps, can_be_int), slider_step, slider_step/1.5, \
           mins, maxs, steps
            # mins, mins, maxs, maxs, mins, maxs



@app.callback(
    Output('configuration_revert_config_file',    'disabled', allow_duplicate=True),
    Input ({'type': 'config_mins', 'index': ALL}, 'value'),
    Input ({'type': 'config_maxs', 'index': ALL}, 'value'),
    Input ({'type': 'config_step', 'index': ALL}, 'value'),
    Input ({'type': 'processings', 'index': ALL}, 'value'),
    Input ({'type': 'redis',       'index': ALL}, 'value'))
def updateRevertButton(mins, maxs, steps, processings, redis):
    ''' If any input value changed, enable 'Revert' button. '''
    global global_config_apply_revert_active

    if global_config_apply_revert_active:
        global_config_apply_revert_active = False
        return no_update

    return False



@app.callback(
    Output('configuration_revert_config_file',           'disabled', allow_duplicate=True),
    Output({'type': 'config_mins',        'index': ALL}, 'value',    allow_duplicate=True),
    Output({'type': 'config_maxs',        'index': ALL}, 'value',    allow_duplicate=True),
    Output({'type': 'config_step',        'index': ALL}, 'value',    allow_duplicate=True),
    Output({'type': 'processings',        'index': ALL}, 'value',    allow_duplicate=True),
    Output({'type': 'redis',              'index': ALL}, 'value',    allow_duplicate=True),
    Input ('configuration_revert_config_file',           'n_clicks'),
    State ({'type': 'hidden_config_mins', 'index': ALL}, 'data'),
    State ({'type': 'hidden_config_maxs', 'index': ALL}, 'data'),
    State ({'type': 'hidden_config_step', 'index': ALL}, 'data'),
    State ({'type': 'hidden_processings', 'index': ALL}, 'data'),
    State ({'type': 'hidden_redis',       'index': ALL}, 'data'))
def configurationRevertChanges(revert, hidden_mins, hidden_maxs, hidden_steps, hidden_processings, hidden_redis):
    global global_config_apply_revert_active
    global_config_apply_revert_active = True    # Global 'lock' so we can disable the Revert button
                                                # as Output of this method triggers Input of 'updateRevertButton()'
    app.logger.warning(f"Configuration: Reverting changes")

    return True, hidden_mins, hidden_maxs, hidden_steps, hidden_processings, hidden_redis



@app.callback(
    Output('modal_configuration', "is_open", allow_duplicate=True),
    Input("close-modal", "n_clicks"))
def close_modal(n):
    if n:
        return False
    return True


@app.callback(
    [
        Output('configuration_save_config_file_output',                  'children'),
        Output('modal_configuration',                                    'is_open',  allow_duplicate=True),
        Output('modal_configuration_div',                                'children', allow_duplicate=True),
        Output('configuration_timer',                                    'disabled'),
        Output('configuration_revert_config_file',                       'disabled', allow_duplicate=True),
        Output({'type': 'hidden_config_mins',        'index': ALL}, 'data'),
        Output({'type': 'hidden_config_maxs',        'index': ALL}, 'data'),
        Output({'type': 'hidden_config_step',        'index': ALL}, 'data'),
        Output({'type': 'hidden_processings',        'index': ALL}, 'data'),
        Output({'type': 'hidden_redis',              'index': ALL}, 'data'),
    ],
    Input('configuration_save_config_file',                          'n_clicks'),
    Input('configuration_timer',                                     'n_intervals'),
    State({'type': 'config_mins', 'index': ALL},                     'value'),
    State({'type': 'config_maxs', 'index': ALL},                     'value'),
    State({'type': 'config_step', 'index': ALL},                     'value'),
    State({'type': 'processings', 'index': ALL},                     'value'),
    State({'type': 'redis',       'index': ALL},                     'value'),
    State('configuration_timer',                                     'disabled'))
def save_config_to_file(n_clicks, n_intervals, config_mins, config_maxs, config_steps, processings, redis, timer_disabled):
    global global_config_apply_revert_active

    ctx = callback_context

    no_up_mins  = [no_update for i in range(len(config_mins))]
    no_up_maxs  = [no_update for i in range(len(config_maxs))]
    no_up_steps = [no_update for i in range(len(config_steps))]
    no_up_procs = [no_update for i in range(len(processings))]
    no_up_redis = [no_update for i in range(len(redis))]

    if not ctx.triggered:
        return "", False, "", timer_disabled, True, no_up_mins, no_up_maxs, no_up_steps, no_up_procs, no_up_redis

    trigger_id = ctx.triggered[0]['prop_id']

    # Interval-ed if
    if 'configuration_save_config_file' in trigger_id:

        # None of the fields can be empty
        missing_fields = isEverythingFilled(config_mins, config_maxs, config_steps, processings, redis)
        if len(missing_fields) > 0:
            return "", True, missing_fields, timer_disabled, False, no_up_mins, no_up_maxs, no_up_steps, no_up_procs, no_up_redis

        global_config_apply_revert_active = True

        # In case all ok and want to restart the Fusion, change one of the limits for a small amount of time
        return "Configuration saved and attempting to restart SW Fusion.", False, "", False, True, config_mins, config_maxs, config_steps, processings, redis

    # After 5 sec when button is pressed, clear text and return value back
    if n_intervals >= 10:
        global_config_apply_revert_active = True
        return "", False, "", True, True, config_mins, config_maxs, config_steps, processings, redis

    raise PreventUpdate


