import sys
sys.path.append('../Visualiser')

from main_dash           import app
from marker              import MarkerType, MarkerColor, MarkerProperty
from utils               import check_range, loadCsvFile, getModal_wrongCsvFile, getModal_deviceNotValid
# from views.visualisation import createDashLeafMarker, getMarkerIndex

from dash            import Output, Input, State, callback_context, dash_table, dcc, html, no_update
from dash.exceptions import PreventUpdate

import globals

import json

# Globals for Configuration
global global_devices_apply_revert_active   # Apply / Revert disabled button

DEVICES_COLS          = len(MarkerProperty) - 6     # TILE_INDEX, TIME_JAM_SINCE, COLOR, IDENTIFIER, PROBABILITY & RMS_CONFIRMED not stored in CSV
DEVICES_NAME          = 'Devices'
DEVICES_WARNING_MODAL = 'devices_modal_wrong_csv'
DEVICES_WARNING_VALID = 'devices_modal_not_valid'

# Dropdown options
type_options = [{"label": member.value, "value": member.value} for member in MarkerType]

backup_options = [
    {"label": "True", "value": "True"},
    {"label": "False", "value": "False"}
]

buttons_style       = {'margin-left':'0.5em'}
buttons_style_right = {'margin-left':'auto'}


def addRuntimePoperties(devices_init_data):
    ''' Extend Device data with identifier and other runtime properties, which is may not be
        stored in csv/json, but it is needed to identify deivce within the app. '''
    for i, d in enumerate(devices_init_data):

        d[MarkerProperty.IDENTIFIER    .value] = i
        d[MarkerProperty.COLOR         .value] = MarkerColor.GREEN.value
        
        d.setdefault(MarkerProperty.RMS_CONFIRMED .value, {})
        d.setdefault(MarkerProperty.TIME_JAM_SINCE.value, {})

        d.setdefault(MarkerProperty.PROBABILITY.value, float(0.0))

        d[MarkerProperty.RMS_CONFIRMED .value] = json.dumps(d[MarkerProperty.RMS_CONFIRMED .value])
        d[MarkerProperty.TIME_JAM_SINCE.value] = json.dumps(d[MarkerProperty.TIME_JAM_SINCE.value])

    return devices_init_data


def getListofMLATdevices(devices_data):
    ''' Get list of all MLAT Devices available, for MLAT System Dropdown (available only for MLAT Antena).'''
    result = []
    clear_result = []
    for d in devices_data:
        if d[MarkerProperty.TYP.value] == MarkerType.MLAT.value:
            result.append( {"label": d[MarkerProperty.NAME.value], "value": d[MarkerProperty.NAME.value]} )
            clear_result.append(d[MarkerProperty.NAME.value])
    return result, clear_result



def createStyleConditions():
    ''' Creates conditional styles for DataTable, so not valid cells for given type are highlighted. '''
    
    def getTypeStyles():
        ''' Iterator for all MarkerTypes. '''
        styles_wrong = []
        styles_ok    = []
        for device_type in MarkerType:
            
            # Cells which should be empty, but are not - 'wrong' paints cells red, 'ok' paints valid cells back to white
            not_filled_wrong = [ {'if': {'filter_query': '{{{}}} = "{}" && {{{}}} != ""'   .format(MarkerProperty.TYP.value.lower(), device_type.value, item), 'column_id': item}, 'backgroundColor': '#F70D1A', 'color': '#FBFBF9'} for item in getActiveColumns(device_type.value, get_negative=True) ]
            not_filled_ok    = [ {'if': {'filter_query': '{{{}}} = "{}" && {{{}}} is blank'.format(MarkerProperty.TYP.value.lower(), device_type.value, item), 'column_id': item}, 'backgroundColor': 'white',   'color': 'black'}   for item in getActiveColumns(device_type.value, get_negative=True) ]

            # Cells which should be filled, but are empty - paint them red
            not_empty_wrong  = [ {'if': {'filter_query': '{{{}}} = "{}" && {{{}}} is blank'.format(MarkerProperty.TYP.value.lower(), device_type.value, item), 'column_id': item}, 'backgroundColor': '#F70D1A', 'color': '#FBFBF9'} for item in getActiveColumns(device_type.value, get_negative=False) ]

            styles_wrong.append(not_filled_wrong)
            styles_ok   .append(not_filled_ok)
            styles_wrong.append(not_empty_wrong)

        return styles_wrong, styles_ok

    wrong, ok = getTypeStyles()

    # MarkerProperty.TYPE CANNOT have overflow hidden otherwise the dropdwon will NOT be visible!!!
    styles = (                                                               # NOTE - maxWidth has to be 0 as without it the overflow does not work
        [ {'if': {'column_id': MarkerProperty.NAME       .value.lower()}, 'overflow': 'hidden', 'textOverflow': 'ellipsis', 'maxWidth':0, 'width': '20%'} ] +
        [ {'if': {'column_id': MarkerProperty.TYP        .value.lower()},                                                                 'width': '15%', 'whiteSpace':'normal'} ] +
        [ {'if': {'column_id': MarkerProperty.LAT        .value.lower()}, 'overflow': 'hidden', 'textOverflow': 'ellipsis', 'maxWidth':0} ] +
        [ {'if': {'column_id': MarkerProperty.LON        .value.lower()}, 'overflow': 'hidden', 'textOverflow': 'ellipsis', 'maxWidth':0} ] +
        [ {'if': {'column_id': MarkerProperty.HGT        .value.lower()}, 'overflow': 'hidden', 'textOverflow': 'ellipsis', 'maxWidth':0} ] +
        [ {'if': {'column_id': MarkerProperty.SENSITIVITY.value.lower()}, 'overflow': 'hidden', 'textOverflow': 'ellipsis', 'maxWidth':0} ] +
        [ {'if': {'column_id': MarkerProperty.INCOMING   .value.lower()}, 'overflow': 'hidden', 'textOverflow': 'ellipsis', 'maxWidth':0} ] +
        [ {'if': {'column_id': MarkerProperty.BACKUP     .value.lower()}, 'overflow': 'hidden', 'textOverflow': 'ellipsis', 'maxWidth':0} ] +
        [ {'if': {'column_id': MarkerProperty.MLAT_SYSTEM.value.lower()}                                                                } ]
    )

    for i in wrong:
        styles.extend(i)
    
    for i in ok:
        styles.extend(i)

    return styles



def layout_devices_tab(devices_init_data):
    # Init global variables
    global global_devices_apply_revert_active
    global_devices_apply_revert_active = True

    devices_init_data = addRuntimePoperties(devices_init_data)

    dropdown_listof_mlat, _ = getListofMLATdevices(devices_init_data)

    return html.Div([
        # Store to have access to data and do not reload it eveyry tick
        #   Note: tick() in visualisation.py did return Dash.no_update to the DataTable "devices_table",
        #         however dispite that DataTable kept reloading, making editing of cells impossible - 
        #         original cell value kept returning back. Thus tick() has to send data to this Store,
        #         so the DataTable is not directly connected to the data.
        #
        #         It's technically just about Time Since Jamming and the color of the marker, nothing more.
        #           Color could be derived from DashMarkers, but cannot insert Time into DashMarker
        #           (maybe into children?)
    #    dcc.Store(id="devices_table_original", data=devices_init_data),

        # Modal Warning Window about wrong CSF file
        getModal_wrongCsvFile(DEVICES_WARNING_MODAL, DEVICES_NAME),

        # Modal Warning Window about invalid Device entry/row
        getModal_deviceNotValid(DEVICES_WARNING_VALID),

        # Load CVS file button
        html.Br(),
        html.Div([
            dcc.Upload(
                id='load_devices_file',
                children=html.Button('Load Devices .CSV File'),
                multiple=False,
                accept='text/csv'
            ),
            html.Button('Add Device',      id='devices_add_row_button',     style=buttons_style),
            html.Button('Clear Table',     id='devices_clear_table_button', style=buttons_style),
            html.Div([
                html.Button('Apply Changes', id='devices_apply_changes_button',   style=buttons_style, disabled=True),
                html.Button('Revert Changes', id='devices_revert_changes_button', style=buttons_style, disabled=True),
            ], style=buttons_style_right)
        ], style={'display': 'flex', 'flexDirection': 'row'}),
        html.Br(),
        
        # DataTable
        dash_table.DataTable(
            id='devices_table',
            columns=[
                {"name": "Name",             "type": "text",    "id": MarkerProperty.NAME       .value.lower()},
                {"name": "Type",                                "id": MarkerProperty.TYP        .value.lower(), "presentation": "dropdown"},
                {"name": "Latitude [deg]",   "type": "numeric", "id": MarkerProperty.LAT        .value.lower()},
                {"name": "Longitude [deg]",  "type": "numeric", "id": MarkerProperty.LON        .value.lower()},
                {"name": "Height [m]",       "type": "numeric", "id": MarkerProperty.HGT        .value.lower()},
                {"name": "Sensitivity [dBm]","type": "numeric", "id": MarkerProperty.SENSITIVITY.value.lower()},
                {"name": "Incoming",         "type": "text",    "id": MarkerProperty.INCOMING   .value.lower()},
                {"name": "Backup",           "type": "text",    "id": MarkerProperty.BACKUP     .value.lower()},
                {"name": "MLAT",                                "id": MarkerProperty.MLAT_SYSTEM.value.lower(), "presentation": "dropdown"},
                {"name": "identifier",       "type": "numeric", "id": MarkerProperty.IDENTIFIER .value.lower()},
                {"name": "Time Since",       "type": "numeric", "id": MarkerProperty.TIME_JAM_SINCE.value},
                {"name": "Tile Index",       "type": "numeric", "id": MarkerProperty.TILE_INDEX    .value},
                {"name": "RMS Confirm",      "type": "any",     "id": MarkerProperty.RMS_CONFIRMED .value},
            ],
            hidden_columns=[
                MarkerProperty.IDENTIFIER    .value.lower(),
                MarkerProperty.TIME_JAM_SINCE.value,
                MarkerProperty.TILE_INDEX    .value,
                MarkerProperty.RMS_CONFIRMED .value
            ],
            data=devices_init_data,
            dropdown={
                MarkerProperty.TYP        .value.lower(): {
                    'options': type_options,
                    'clearable': False,
                },
                MarkerProperty.MLAT_SYSTEM.value.lower(): {
                    'options': dropdown_listof_mlat,
                    'clearable': True
                }

            },
            editable=True,
            row_deletable=True,
            row_selectable="multi",
            sort_action="native",
            selected_rows=[i for i, item in enumerate(devices_init_data) if item[MarkerProperty.ENABLED_FUSION.value]],
            style_cell={'textAlign': 'center', 'width': '12%'},
            style_data_conditional=createStyleConditions(),
            # page_size=15,
            # style_table={'overflowY': 'auto'},
            css=[{"selector": ".show-hide", "rule": "display: none"}]   # hide 'Toggle Columns' button, enabled be attribute 'hidden_columns'
        ),

        html.Div(id="devices_json_out", style={'display':'none'}),

        # Dash Store to store data: 1) visible editable 2) hidden original
        dcc.Store(id="devices_table_hidden", data=devices_init_data)
    ])


def checkForDuplicities(devices_init_data):
    ''' Checks for duplicity names and if any found, later name is modified with id
        and disabled from SW Fusion and Visualisation. '''
    unique_names = []
    items_to_modify = {}
    updated_names = []

    if devices_init_data == None:
        devices_init_data = []

    for i, d in enumerate(devices_init_data):
        name = d[MarkerProperty.NAME.value]
        if name not in unique_names:
            unique_names.append(name)
        else:
            items_to_modify.update({i: d})

    # Duplicity items - append id to the name AND disable it from visualisation and SW Fusion
    for key,val in items_to_modify.items():
        devices_init_data[key][MarkerProperty.NAME.value]           = f"{key}_{devices_init_data[key][MarkerProperty.NAME.value]}"
        devices_init_data[key][MarkerProperty.ENABLED_FUSION.value] = False

        updated_names.append( devices_init_data[key][MarkerProperty.NAME.value] )

    # Update list of selected rows as well
    selected_rows = [i for i, item in enumerate(devices_init_data) if item[MarkerProperty.ENABLED_FUSION.value]]

    return devices_init_data, updated_names, selected_rows


def initialiseDevicesToColor(devices_init_data, color):
    ''' Initialise data with color. '''
    for d in devices_init_data:
        d[MarkerProperty.COLOR.value] = color.value
    return devices_init_data


def getDeviceProperty(devices_data, name, prop):
    ''' Returns given property for device specified by param name. '''
    for i,d in enumerate(devices_data):
        if d[MarkerProperty.NAME.value] == name:
            return d[prop.value]
    return None


def updateDeviceDataWithRuntimeProps(devices_data, devices_data_colors):
    ''' Update runtime props in devices_data based on devices_data_colors  - lists are the same size and everything. '''
    for i,d in enumerate(devices_data):
        # Try-Except - in case processed row is new one, those runtime properties does not exist - put a default ones
        try:
            d[MarkerProperty.COLOR         .value] = devices_data_colors[i][MarkerProperty.COLOR         .value]
        except IndexError:
            d[MarkerProperty.COLOR         .value] = MarkerColor.GREEN.value
        
        try:
            d[MarkerProperty.TIME_JAM_SINCE.value] = devices_data_colors[i][MarkerProperty.TIME_JAM_SINCE.value]
        except IndexError:
            d[MarkerProperty.TIME_JAM_SINCE.value] = 0.0
        
        try:
            d[MarkerProperty.RMS_CONFIRMED .value] = devices_data_colors[i][MarkerProperty.RMS_CONFIRMED .value]
        except IndexError:
            d[MarkerProperty.RMS_CONFIRMED .value] = False

    return devices_data


def updateDeviceDataWithoutRuntimeProps(device_data, device_data_colors):
    for i,d in enumerate(device_data_colors):
        for prop in MarkerProperty:
            if prop != MarkerProperty.COLOR or prop != MarkerProperty.TIME_JAM_SINCE or prop != MarkerProperty.RMS_CONFIRMED:
                
                try:
                    d[prop.value] = device_data[i][prop.value]
                except (IndexError, KeyError):
                    # IndexError is raised if device with given index was removed - no need to process it further
                    # KeyError is raised if property is not stored array, did not exists before this runtime - no action needed
                    continue

    return device_data_colors



def areDevicesEntriesValid(devices_data):
    ''' Checks if all entries in DataTable are valid. If not, particular entry is stringified and returned, 
        so user can be informed about it. '''

    def isEmpty(item, k):
        return item == None or item[k] == "" or item[k] == None

    def isNotImportantProperty(item):
        # Probability and Tile Index are created during runtime, no need to check it
        return item.lower() == MarkerProperty.PROBABILITY   .value.lower() or \
               item.lower() == MarkerProperty.COLOR         .value.lower() or \
               item.lower() == MarkerProperty.ENABLED_FUSION.value.lower() or \
               item.lower() == MarkerProperty.TIME_JAM_SINCE.value.lower() or \
               item.lower() == MarkerProperty.IDENTIFIER    .value.lower() or \
               item.lower() == MarkerProperty.RMS_CONFIRMED .value.lower() or \
               item.lower() == MarkerProperty.TILE_INDEX    .value.lower()

    def get_title(property, text):
        return html.Div([ html.B( item[MarkerProperty.NAME.value.lower()] ), f" - {property}{text}" ])

    not_valid_devices = []
    for item in devices_data:
        active_columns = getActiveColumns(item[MarkerProperty.TYP.value])
        for k in item:
            if k not in active_columns and not isEmpty(item, k):
                item[k] = None
                if not isNotImportantProperty(k):
                    not_valid_devices.append(get_title(MarkerProperty.get_enum_by_value(k).string, " (should be blank)"))

            if k in active_columns and isEmpty(item, k):
                if not isNotImportantProperty(k):
                    not_valid_devices.append(get_title(MarkerProperty.get_enum_by_value(k).string, " (should NOT be blank)"))

        # Check GNSS ranges < -180 ; 180 >
        def check_and_append(not_valid_devices, property_typ, item):
            wrong_range = check_range([], property_typ, item, [-180.0,180.0], unit=" [deg]", get_text_only=True)        
            if wrong_range is not None:
                not_valid_devices.append(get_title("", wrong_range))

        check_and_append(not_valid_devices, MarkerProperty.LAT, item)
        check_and_append(not_valid_devices, MarkerProperty.LON, item)

    return not_valid_devices




def processNewData(selected_rows, devices_data, devices_data_hidden, dash_markers, rms_info_table, devices_count):
    ''' Handles new data during Apply Changes - mostly makes the changes related to 'dash_markers' structure
        holding all Dash Markers in Map view in visualisation.py
        
        But firstly - check if all data are correctly set, nothing is wrong, all cells are filled etc.
    '''
    notValidDevices = areDevicesEntriesValid(devices_data)

    if len(notValidDevices) > 0:
        return [notValidDevices] + [no_update]*5

    # Due to circular imports, declared here
    from views.visualisation import createDashLeafMarker, getMarkerIndex

    device_names       = {}
    dash_device_names  = []

    devices_data = updateDeviceDataWithRuntimeProps(devices_data, devices_data_hidden)

    # Get list of Device names to Render/Pass to JSON request
    if selected_rows != None:
        for i in selected_rows:
            try:
                device_names.update({i: devices_data[i][MarkerProperty.NAME.value]})
            except IndexError as e:
                app.logger.warning(f"Active row {i} does not exists, probably the Table was cleared.")

    # Check for Devices to remove from Dash/Render list
    from views.visualisation import getMarkerByIdentifier   # Due to circular imports, declared here
    dash_to_remove = {}

    for m in dash_markers:

        m_identifier = getMarkerIndex(m)                                    # get unique identifier 
        m_index      = getMarkerByIdentifier(m_identifier, devices_data)    # get index in array of given item with that identifier
        m_index_2    = getMarkerByIdentifier(m_identifier, devices_data_hidden)    # get index in array of given item with that identifier

        def find_dash_marker_name(m_identifier, devices_data):
            m_index = getMarkerByIdentifier(m_identifier, devices_data)
            if m_index != None:
                try:
                    marker_name = devices_data[m_index][MarkerProperty.NAME.value]
                except IndexError as e:
                    # IndexError is raised if device with given index was removed - no need to process it further
                    app.logger.warning(f"devices.display_delete_rows(): isRendering.dash_markers.devices_data: m_index: {m_index} | {e}")
                    return None
                return marker_name
            else:
                return None


        marker_name = find_dash_marker_name(m_identifier, devices_data)
        if marker_name == None:
            marker_name = find_dash_marker_name(m_identifier, devices_data_hidden)

        dash_device_names.append(marker_name)
     
        if marker_name not in device_names.values():
            dash_to_remove.update({marker_name: m})


    def findOriginalIdentifier(devices, name, devices_count):
        for i,d in enumerate(devices):
            if d[MarkerProperty.NAME.value] == name:
                return i
        return devices_count

    # Check for New Devices to add to Dash/Render List
    if len(device_names) > 0:
        for i, d in device_names.items():
            if d not in dash_device_names:
                original_index = findOriginalIdentifier(devices_data, devices_data[i][MarkerProperty.NAME.value], devices_count)
                dash_markers.append( createDashLeafMarker(
                                        original_index,
                                        devices_data[i][MarkerProperty.NAME.value],
                                        devices_data[i][MarkerProperty.TYP.value],
                                        MarkerColor.GREEN.value,        # default color is always green
                                        float(devices_data[i][MarkerProperty.LAT.value]),
                                        float(devices_data[i][MarkerProperty.LON.value]) ) )

                devices_data[i][MarkerProperty.COLOR.value]          = MarkerColor.GREEN.value
                devices_data[i][MarkerProperty.ENABLED_FUSION.value] = True

                # Technically I am NOT adding new device, only Marker
                ## devices_count += 1  # global device count, for unique ID

    # Create new dash devices list without not wanted ones
    dash_devices_to_keep = []
    for m in dash_markers:
        if m not in dash_to_remove.values():
            dash_devices_to_keep.append(m)

    # "Problem" - need to remove the Device from RMs table as well if present in there
    rows_keep = []
    for i, r in enumerate(rms_info_table):
        if r['Name'] not in dash_to_remove.keys():
            rows_keep.append(r)

    # Same goes for list of devices for jamming at location - here it is easy as i just need to disable given device
    for i, d in enumerate(devices_data):
        if d[MarkerProperty.NAME.value] not in device_names.values():
            d[MarkerProperty.ENABLED_FUSION.value] = False

    # Update list of active device for graph 'jamming at location'
    from views.visualisation import getActiveDevices    # Due to circular import declared here
    devices_jamming_signal_list = getActiveDevices(devices_data)

    return None, dash_devices_to_keep, devices_data, rows_keep, devices_jamming_signal_list, devices_count



@app.callback(
    Output('devices_table',                 'data',          allow_duplicate=True),
    Output('devices_table_hidden',          'data',          allow_duplicate=True),
    Output('devices_apply_changes_button',  'disabled',      allow_duplicate=True),
    Output('devices_revert_changes_button', 'disabled',      allow_duplicate=True),
    Output('dash_markers_group',            'children',      allow_duplicate=True),
    Output('visual_rm_table',               'data',          allow_duplicate=True),
    Output('dropdown_jamming_at_location',  'options'),
    Output('stored_global_device_id',       'data'),

    Output(DEVICES_WARNING_VALID,          'is_open'),
    Output(f'devices-modal-{DEVICES_WARNING_VALID}', 'children'),
  
    Input ('devices_apply_changes_button',  'n_clicks'),
    State ('devices_table',                 'data'),
    State ('devices_table_hidden',          'data'),
    State ('devices_table',                 'selected_rows'),
    State ('dash_markers_group',            'children'),    # Dash Markers in map   / in visualisation.py
    State ("visual_rm_table",               'data'),        # RMs Info Table        / in visualisation.py
    State ('stored_global_device_id',       'data'))         # Number of devices, for unique device ID
def devicesApplyForChanges(apply, visible, hidden, selected_rows, dash_markers, rms_info_table, devices_count):

    not_valid_devices, markers_keep, device_updated, rms_rows_keep, devices_jam_list, devices_count = \
        processNewData(selected_rows, visible, hidden, dash_markers, rms_info_table, devices_count)

    if not_valid_devices != None:
        # not_valid = "\n".join(not_valid_devices)
        return [no_update]*8 + [True] + [not_valid_devices]

    app.logger.warning(f"Devices: Applying changes")

    # Must enable global, because we are updating hidden values of the datatable - without this, Apply/Revert buttons would flick and stay enabled
    global global_devices_apply_revert_active
    global_devices_apply_revert_active = True

    return device_updated, device_updated, True, True, markers_keep, rms_rows_keep, devices_jam_list, devices_count, False, None




@app.callback(
    Output('devices_table',                 'data'),
    Output('devices_table',                 'selected_rows', allow_duplicate=True),
    Output('devices_apply_changes_button',  'disabled',      allow_duplicate=True),
    Output('devices_revert_changes_button', 'disabled',      allow_duplicate=True),
    Input ('devices_revert_changes_button', 'n_clicks'),
    State ('devices_table',                 'data'),
    State ('devices_table_hidden',          'data'))
def devicesRevertChanges(revert, visible, hidden):
    global global_devices_apply_revert_active
    global_devices_apply_revert_active = True

    app.logger.warning(f"Devices: Reverting changes in DataTable")
    _, _ , selected_rows = checkForDuplicities(hidden)

    return hidden, selected_rows, True, True



def newRow(enabled,name,typ,lat,lon,hgt,sensitivity,incoming,backup,mlat_system,color=None,identifier=None,tileIndex=None,timeSince=None, probability=None, rms_confirmed=None):
    return {
        MarkerProperty.NAME          .value: name,
        MarkerProperty.TYP           .value: typ,
        MarkerProperty.COLOR         .value: color,
        MarkerProperty.LAT           .value: lat,
        MarkerProperty.LON           .value: lon,
        MarkerProperty.HGT           .value: hgt,
        MarkerProperty.SENSITIVITY   .value: sensitivity,
        MarkerProperty.INCOMING      .value: incoming,
        MarkerProperty.BACKUP        .value: backup,
        MarkerProperty.ENABLED_FUSION.value: enabled,
        MarkerProperty.IDENTIFIER    .value: identifier,
        MarkerProperty.TILE_INDEX    .value: tileIndex,
        MarkerProperty.TIME_JAM_SINCE.value: timeSince,
        MarkerProperty.PROBABILITY   .value: probability,
        MarkerProperty.RMS_CONFIRMED .value: rms_confirmed,
        MarkerProperty.MLAT_SYSTEM   .value: mlat_system,
    }


# Updates table when new Row is being added or the whole Table being deleted.
@app.callback(
    Output('devices_table',              'data',          allow_duplicate=True),
    Output('devices_table',              'selected_rows', allow_duplicate=True),
    Input ('devices_add_row_button',     'n_clicks'),
    Input ('devices_clear_table_button', 'n_clicks'),
    State ('devices_table',              'derived_virtual_selected_rows'),
    State ('devices_table',              'data'    ))
def update_table(add_clicks, clear_clicks, selected_rows, current_data):
    ctx = callback_context
    if ctx.triggered[0]['prop_id'] == 'devices_add_row_button.n_clicks':
        new_row = newRow(False, '', MarkerType.AIRPLANE.value, '', '', '', '', '', '', '', '', '', '')
        current_data.append(new_row)
    elif ctx.triggered[0]['prop_id'] == 'devices_clear_table_button.n_clicks':
        current_data.clear()
        selected_rows.clear()
    else:
        raise PreventUpdate
    return current_data, selected_rows



def getActiveColumns(selected_type: str, get_negative=False) -> list:
    ''' Based on paramater elected_type returns list of active columns for given type. '''
    default = [ MarkerProperty.IDENTIFIER.value,
                MarkerProperty.ENABLED_FUSION.value,
                MarkerProperty.NAME.value,
                MarkerProperty.TYP.value,
                MarkerProperty.COLOR.value,
                MarkerProperty.LAT.value,
                MarkerProperty.LON.value,
                MarkerProperty.SENSITIVITY.value,
                MarkerProperty.PROBABILITY.value,
                MarkerProperty.TIME_JAM_SINCE.value,
                MarkerProperty.TILE_INDEX.value,    # TILE_INDEX is not visible in table, but has to be present,
                                                    # otherwise it's zeroed and not visible in Map
                MarkerProperty.RMS_CONFIRMED.value
            ]

    # In case TYPE is None, note about that one - Dash has bug with this one, Type column should not be Clearable, but it is.
    if selected_type != None:
        selected_type = selected_type.lower()   # lower case for easier testing
    else:
        return default
    
    if selected_type == MarkerType.AIRPLANE    .value.lower() or \
       selected_type == MarkerType.VEHICLE     .value.lower():
       # Only Height
        active_columns = [MarkerProperty.HGT.value]

    elif selected_type == MarkerType.VEHICLE.value.lower():
        # Vehicle is special, as it has only default properties
        active_columns = []

    elif selected_type == MarkerType.MLAT_ANTENNA.value.lower():
        # MLAT Antena is the only one with MLAT System column defining to which MLAT System the antena belongs to
        active_columns = [MarkerProperty.HGT.value, MarkerProperty.MLAT_SYSTEM.value]

    elif selected_type == MarkerType.RADAR     .value.lower() or \
         selected_type == MarkerType.ADSB      .value.lower() or \
         selected_type == MarkerType.PROCESSING.value.lower():
        # Height, Backup
        active_columns = [MarkerProperty.HGT.value, MarkerProperty.BACKUP.value]

    elif selected_type == MarkerType.MLAT.value.lower():
        # height, Incoming, Backup
        active_columns = [MarkerProperty.HGT.value, MarkerProperty.INCOMING.value, MarkerProperty.BACKUP.value]

    else:
        active_columns = []

    default.extend(active_columns)

     # in case we want opposite of all active columns
    if get_negative:
        default_opposite = []
        
        for item in MarkerProperty:
            if item.value.lower() not in default:
                default_opposite.append(item.value.lower())

        default.clear()
        default = default_opposite

    return default



@app.callback(
    Output('devices_apply_changes_button',  'disabled', allow_duplicate=True),
    Output('devices_revert_changes_button', 'disabled', allow_duplicate=True),
    Output('devices_table',                 'dropdown', allow_duplicate=True),
    Output('devices_table',                 'data',     allow_duplicate=True),
    Input ('devices_table',                 'data'                          ),
    Input ('devices_table',                 'derived_virtual_selected_rows' ),
    State ('devices_table',                 'dropdown'                      ))
def update_columns(initial_data, selected_rows, dropdown):
    global global_devices_apply_revert_active
    if global_devices_apply_revert_active:
        global_devices_apply_revert_active = False
        return [no_update]*4

    # Update MLAT System Dropdown and based on that, update MLAT System cell for devices with type MLAT Antena, if necessary
    dropdown[MarkerProperty.MLAT_SYSTEM.value]['options'], listofMLATs = getListofMLATdevices(initial_data)

    for d in initial_data:
        if d[MarkerProperty.TYP.value] == MarkerType.MLAT_ANTENNA.value:
            if d[MarkerProperty.MLAT_SYSTEM.value] not in listofMLATs:
                d[MarkerProperty.MLAT_SYSTEM.value] = None

    return False, False, dropdown, initial_data



''' Load Devices .CSV file '''
@app.callback(
    Output('devices_table',            'data',          allow_duplicate=True),
    Output('devices_table',            'selected_rows', allow_duplicate=True),
    Output( DEVICES_WARNING_MODAL,     'is_open',       allow_duplicate=True),
    Output('dash_markers_group',       'children',      allow_duplicate=True),
    Output('stored_global_device_id',  'data',          allow_duplicate=True),
    Input('load_devices_file',         'contents'       ),
    State('devices_table',             'data'           ),
    State('devices_table',             'selected_rows'  ),
    State('dash_markers_group',        'children'       ),
    State('stored_global_device_id',   'data'           ),  # Number of devices, for unique device ID
)
def load_devices_csv_file(contents, devices_data, selected_rows, dash_markers, devices_count):
    if contents is None:
        return no_update, no_update, no_update, False

    # data, rows = loadCsvFile(contents, devices_data, selected_rows, DEVICES_COLS, f"Selected file is not correct {DEVICES_NAME} .CSV file.", newRow)
    new_data, new_rows = loadCsvFile(contents, DEVICES_COLS, f"Selected file is not correct {DEVICES_NAME} .CSV file.", newRow)

    # if data & rows are None and not 'no_update', .csv file was not correct
    if new_data == None and selected_rows == None:
        return no_update, no_update, True, no_update, no_update

    num_of_rows = len(devices_data)

    # selected new_rows indexes are related to the the new data only, not for the whole dataset
    for r in new_rows:
        selected_rows.append( num_of_rows + r )
    
    devices_data.extend(new_data)

    return devices_data, selected_rows, False, dash_markers, devices_count



''' Closes Modal Warning Window. '''
@app.callback(
    Output(DEVICES_WARNING_MODAL, "is_open", allow_duplicate=True),
    Input(f"close-{DEVICES_WARNING_MODAL}", "n_clicks"),
)
def close_modal(n):
    if n:
        return False
    return True



''' Close Modal Warning window about sensor entry/row not valid. '''
@app.callback(
    Output(DEVICES_WARNING_VALID,           "is_open", allow_duplicate=True),
    Input(f"close-{DEVICES_WARNING_VALID}", "n_clicks"),
)
def close_modal(n):
    if n:
        return False
    return True


