import sys
sys.path.append('../Visualiser')

from main_dash import app
from marker    import MarkerColor, MarkerType
from utils     import check_range, loadCsvFile, getModal_wrongCsvFile, getModal_rmsNotValid

from dash import Output, Input, State, dash_table, dcc, html, callback_context, no_update
from enum import Enum

class RmsProperty(Enum):
    ACTIVE      = ("active",      "Active"     )
    ID          = ("id",          "ID"         )
    TYPE        = ("Type",        "Type"       )
    PROBABILITY = ("Probability", "Probability")
    INCOMING    = ("Incoming",    "Incoming"   ) # Former KP1
    BACKUP      = ("Backup",      "Backup"     ) # Former KP2
    REDUNDANCY  = ("Redundancy",  "Redundancy" ) # Former KP3
    HOLDOVER    = ("Holdover",    "Holdover"   ) # Former KP4, [min]
    RM          = ("RM",          "RM"         )
    COLOR       = ("color",       "Color"      ) # not stored in csv, just for rendering purpose

    def __new__(cls, value, string_value):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.string = string_value
        return obj

    @classmethod
    def get_name(cls):
        return "RMs"

# Globals for Configuration
global global_rms_apply_revert_active   # Apply / Revert disabled button

RMS_COLS          = len(RmsProperty) - 1    # except COLOR
RMS_NAME          = 'RMs'
RMS_WARNING_MODAL = 'rms_modal_wrong_csv'
RMS_WARNING_VALID = 'rms_modal_not_valid'

buttons_style       = {'margin-left':'0.5em'}
buttons_style_right = {'margin-left':'auto'}

# Dropdown options
type_options = [{"label": member.value, "value": member.value} for member in MarkerType]


def getRmById(rms_table, name: str) -> str:
    ''' Returns RM text based on name of the RM. '''
    for r in rms_table:
        if r[RmsProperty.ID.value.lower()] == name:
            return r[RmsProperty.RM.value]
    return None



def getRmsByName(rms_table, name: str) -> list:
    ''' Returns RMs rows based on name of the RM. '''
    result = []
    for i,r in enumerate(rms_table):
        if r[RmsProperty.ID.value.lower()] == name:
            result.append(i)
    return result



def getRmsByDeviceType(rms_table, device_type: MarkerType):
    ''' Returns RMs rows based on device type. '''
    result = []
    for i,r in enumerate(rms_table):
        if r[RmsProperty.TYPE.value.lower()] == device_type:
            result.append(i)
    return result



def initialiseRmsDataToColor(data, color):
    ''' Init data from stored file/config might have color set to RED, which is not proper value.'''
    for r in data:
        r[RmsProperty.COLOR.value.lower()] = color.value
    return data



def createStyleConditions():
    ''' Creates conditional styles for DataTable, so not valid cells for given type are highlighted. '''
    
    def getTypeStyles():
        ''' Iterator for all MarkerTypes. '''
        styles_wrong = []
        styles_ok    = []
        for device_type in MarkerType:
            
            # Cells which should be empty, but are not - 'wrong' paints cells red, 'ok' paints valid cells back to white
            not_filled_wrong = [ {'if': {'filter_query': '{{{}}} = "{}" && {{{}}} != ""'   .format(RmsProperty.TYPE.value.lower(), device_type.value, item), 'column_id': item}, 'backgroundColor': '#F70D1A', 'color': '#FBFBF9'} for item in getActiveColumns(device_type.value, get_negative=True) ]
            not_filled_ok    = [ {'if': {'filter_query': '{{{}}} = "{}" && {{{}}} is blank'.format(RmsProperty.TYPE.value.lower(), device_type.value, item), 'column_id': item}, 'backgroundColor': 'white',   'color': 'black'}   for item in getActiveColumns(device_type.value, get_negative=True) ]

            # Cells which should be filled, but are empty - paint them red
            not_empty_wrong  = [ {'if': {'filter_query': '{{{}}} = "{}" && {{{}}} is blank'.format(RmsProperty.TYPE.value.lower(), device_type.value, item), 'column_id': item}, 'backgroundColor': '#F70D1A', 'color': '#FBFBF9'} for item in getActiveColumns(device_type.value, get_negative=False) ]

            styles_wrong.append(not_filled_wrong)
            styles_ok   .append(not_filled_ok)
            styles_wrong.append(not_empty_wrong)

        return styles_wrong, styles_ok

    wrong, ok = getTypeStyles()

    styles = (
        [ {'if': {'column_id': RmsProperty.RM         .value.lower()}, 'width': '40%', 'textAlign': 'left'} ] +
        [ {'if': {'column_id': RmsProperty.ID         .value.lower()}, 'width': '15%', 'textOverflow': 'ellipsis', 'overflow': 'hidden', 'maxWidth':0} ] +
        [ {'if': {'column_id': RmsProperty.TYPE       .value.lower()}, 'width': '12%'                                                                } ] +
        [ {'if': {'column_id': RmsProperty.PROBABILITY.value.lower()}, 'width':  '5%', 'textOverflow': 'ellipsis', 'overflow': 'hidden', 'maxWidth':0} ] +
        [ {'if': {'column_id': RmsProperty.INCOMING   .value.lower()}, 'width':  '5%', 'textOverflow': 'ellipsis', 'overflow': 'hidden', 'maxWidth':0} ] +
        [ {'if': {'column_id': RmsProperty.BACKUP     .value.lower()}, 'width':  '5%', 'textOverflow': 'ellipsis', 'overflow': 'hidden', 'maxWidth':0} ] +
        [ {'if': {'column_id': RmsProperty.REDUNDANCY .value.lower()}, 'width':  '5%', 'textOverflow': 'ellipsis', 'overflow': 'hidden', 'maxWidth':0} ] +
        [ {'if': {'column_id': RmsProperty.HOLDOVER   .value.lower()}, 'width':  '5%', 'textOverflow': 'ellipsis', 'overflow': 'hidden', 'maxWidth':0} ] +
        [ {'if': {'column_id': RmsProperty.COLOR      .value.lower()}, 'width':  '5%', 'textOverflow': 'ellipsis', 'overflow': 'hidden', 'maxWidth':0} ] )

    for i in wrong:
        styles.extend(i)
    
    for i in ok:
        styles.extend(i)

    return styles



def layout_rm_tab(rms_init_data):
    # Init global variables
    global global_rms_apply_revert_active
    global_rms_apply_revert_active = True

    rms_init_data = rms_init_data if rms_init_data != None else []
    
    return html.Div([
        
        # Modal Warning Window about wrong CSF file
        getModal_wrongCsvFile(RMS_WARNING_MODAL, RMS_NAME),

        # Modal Warning Window about invalid Sensor entry/row
        getModal_rmsNotValid(RMS_WARNING_VALID),

        # Load CVS file button
        html.Br(),
        html.Div([
            dcc.Upload(
                id='load_rm_file',
                children=html.Button('Load RMs. CSV File'),
                multiple=False,
                accept='text/csv'
            ),
            html.Button('Add RM',      id='rms_add_row_button',     style=buttons_style),
            html.Button('Clear Table', id='rms_clear_table_button', style=buttons_style),
            html.Div([
                html.Button('Apply Changes', id='rms_apply_changes_button',   style=buttons_style, disabled=True),
                html.Button('Revert Changes', id='rms_revert_changes_button', style=buttons_style, disabled=True),
            ], style=buttons_style_right)
        ], style={'display': 'flex', 'flexDirection': 'row'}),
        html.Br(),

        dash_table.DataTable(
            id='rms_table',
            columns=[
                {'name': RmsProperty.ID.value.upper(),                             'id': RmsProperty.ID         .value.lower()},
                {'name': RmsProperty.TYPE.value,                                   'id': RmsProperty.TYPE       .value.lower(), "presentation": "dropdown"},
                {'name': RmsProperty.PROBABILITY.value,         "type": "numeric", 'id': RmsProperty.PROBABILITY.value.lower(), 'editable': True, 'clearable': True},
                {'name': RmsProperty.INCOMING.value,            "type": "text",    'id': RmsProperty.INCOMING   .value.lower(), 'editable': True, 'clearable': True},
                {'name': RmsProperty.BACKUP.value,              "type": "numeric", 'id': RmsProperty.BACKUP     .value.lower(), 'editable': True, 'clearable': True},
                {'name': RmsProperty.REDUNDANCY.value,          "type": "numeric", 'id': RmsProperty.REDUNDANCY .value.lower(), 'editable': True, 'clearable': True},
                {'name': f"{RmsProperty.HOLDOVER.value} [min]", "type": "numeric", 'id': RmsProperty.HOLDOVER   .value.lower(), 'editable': True, 'clearable': True},
                {'name': RmsProperty.RM.value,                  "type": "text",    'id': RmsProperty.RM         .value.lower(), 'editable': True, 'clearable': True},
                {'name': RmsProperty.COLOR.value,                                  'id': RmsProperty.COLOR      .value.lower(), 'editable': True}
            ],
            hidden_columns=[RmsProperty.COLOR.value],
            data=rms_init_data,
            dropdown={
                RmsProperty.TYPE.value.lower(): {
                    'options': type_options,
                    'clearable': False,
                },

            },
            editable=True,
            row_deletable=True,
            row_selectable="multi",
            sort_action="native",
            selected_rows=[i for i, item in enumerate(rms_init_data) if item['active']],
            style_cell={'textAlign': 'center'},
            style_data={'whiteSpace': 'normal', 'height': 'auto'},
            style_data_conditional=createStyleConditions(),
            css=[{"selector": ".show-hide", "rule": "display: none"}]   # hide 'Toggle Columns' button, enabled be attribute 'hidden_columns'
        ),

        # Dash Store to store data: 1) visible editable 2) hidden original
        dcc.Store(id="rms_table_hidden", data=rms_init_data)
    ])



def newRow(active, identifier, device_type, probability, incoming, backup, redundancy, holdover, rm, color=MarkerColor.GREEN.value):
    return {
        RmsProperty.ACTIVE     .value.lower(): active,
        RmsProperty.ID         .value.lower(): identifier,
        RmsProperty.TYPE       .value.lower(): device_type,
        RmsProperty.PROBABILITY.value.lower(): probability,
        RmsProperty.INCOMING   .value.lower(): incoming,
        RmsProperty.BACKUP     .value.lower(): backup,
        RmsProperty.REDUNDANCY .value.lower(): redundancy,
        RmsProperty.HOLDOVER   .value.lower(): holdover,
        RmsProperty.RM         .value.lower(): rm,
        RmsProperty.COLOR      .value.lower(): color
    }



def getActiveColumns(selected_type: str, get_negative=False) -> list:
    ''' Based on selected_type, returns list columns which should be filled/active. '''
    default = [ RmsProperty.ACTIVE     .value.lower(),
                RmsProperty.ID         .value.lower(),
                RmsProperty.TYPE       .value.lower(),
                RmsProperty.PROBABILITY.value.lower(),
                RmsProperty.RM         .value.lower(),
                RmsProperty.COLOR      .value.lower()]

    # In case TYPE is None, note about that one - Dash has bug with this one, Type column should not be Clearable, but it is.
    if selected_type != None:
        selected_type = selected_type.lower()   # lower case for easier testing
    else:
        return default

    active_columns = []

    if selected_type == MarkerType.AIRPLANE    .value.lower() or \
       selected_type == MarkerType.VEHICLE     .value.lower() or \
       selected_type == MarkerType.GROUND      .value.lower() or \
       selected_type == MarkerType.MLAT_ANTENNA.value.lower():
        active_columns = []

    elif selected_type == MarkerType.RADAR     .value.lower() or \
         selected_type == MarkerType.ADSB      .value.lower() or \
         selected_type == MarkerType.PROCESSING.value.lower():
        active_columns = [RmsProperty.BACKUP.value.lower(), RmsProperty.HOLDOVER.value.lower()]

    elif selected_type == MarkerType.MLAT.value.lower():
        active_columns = [RmsProperty.INCOMING.value.lower(), RmsProperty.BACKUP.value.lower(), RmsProperty.REDUNDANCY.value.lower(), RmsProperty.HOLDOVER.value.lower()]

    default.extend(active_columns)

    # in case we want opposite of all active columns
    if get_negative:
        default_opposite = []
        
        for item in RmsProperty:
            if item.value.lower() not in default:
                default_opposite.append(item.value.lower())

        default.clear()
        default = default_opposite

    return default



def areEntriesValid(data):
    ''' Checks if all data are valid, if not informa user. '''
    def isEmpty(item, k):
        return item == None or item[k] == "" or item[k] == None

    def isNotImportantProperty(item):
        # Color is created during runtime, no need to check it
        return item.lower() == RmsProperty.COLOR.value.lower()

    def get_title(property, text):
        return html.Div([
                    html.B( item[RmsProperty.ID.value.lower()] ),
                    f" - {property}{text}"
                ])

    not_valid_rms = []
    for item in data:
        active_columns = getActiveColumns(item[RmsProperty.TYPE.value.lower()])        
        for k in item:
            if k not in active_columns and not isEmpty(item, k):
                item[k] = None
                if not isNotImportantProperty(k):
                    not_valid_rms.append(get_title(k.title(), " (should be blank)"))

            if k in active_columns and isEmpty(item, k):
                if not isNotImportantProperty(k):
                    not_valid_rms.append(get_title(k.title(), " (should NOT be blank)"))

        # Check GNSS ranges < -180 ; 180 >
        wrong_range = check_range([], RmsProperty.PROBABILITY, item, [0.0,1.0], unit="", get_text_only=True)
        if wrong_range != None:
            not_valid_rms.append(get_title("", wrong_range))

    return len(not_valid_rms) > 0, not_valid_rms



def processNewData(selected_rows, data):
    openModal, notValidEntries = areEntriesValid(data)
    return openModal, notValidEntries



def checkForDuplicities(init_data):
    ''' Checks for duplicity names and if any found, later name is modified with id
        and disabled from SW Fusion and Visualisation. '''
    unique_names    = []
    items_to_modify = {}
    updated_names   = []

    if init_data == None:
        init_data = []

    for i, d in enumerate(init_data):
        name = d[RmsProperty.ID.value]
        if name not in unique_names:
            unique_names.append(name)
        else:
            items_to_modify.update({i: d})

    # Duplicity items - append id to the name AND disable it from visualisation and SW Fusion
    for key,val in items_to_modify.items():
        init_data[key][RmsProperty.ID.value]     = f"{key}_{init_data[key][RmsProperty.ID.value]}"
        init_data[key][RmsProperty.ACTIVE.value] = False

        updated_names.append( init_data[key][RmsProperty.ID.value] )

    # Update list of selected rows as well
    selected_rows = [i for i, item in enumerate(init_data) if item[RmsProperty.ACTIVE.value]]

    return init_data, updated_names, selected_rows




@app.callback(
    Output('rms_table_hidden',          'data'),
    Output('rms_apply_changes_button',  'disabled',      allow_duplicate=True),
    Output('rms_revert_changes_button', 'disabled',      allow_duplicate=True),
    # Output('rms_table',                 'selected_rows', allow_duplicate=True),
    Output(RMS_WARNING_VALID,           'is_open'),
    Output(f'rms-modal-{RMS_WARNING_VALID}', 'children'),
    Input ('rms_apply_changes_button',  'n_clicks'),
    State ('rms_table',                 'data'),
    State ('rms_table_hidden',          'data'),
    State ('rms_table',                 'selected_rows'))
def rmsApplyForChanges(apply, visible, hidden, selected_rows):

    openModal, modalText = processNewData(selected_rows, visible)

    if openModal:
        return no_update, False, False, openModal, modalText

    app.logger.warning(f"rms: Applying changes")
    return visible, True, True, no_update, no_update




@app.callback(
    Output('rms_table',                 'data'),
    Output('rms_table',                 'selected_rows', allow_duplicate=True),
    Output('rms_apply_changes_button',  'disabled',      allow_duplicate=True),
    Output('rms_revert_changes_button', 'disabled',      allow_duplicate=True),
    Input ('rms_revert_changes_button', 'n_clicks'),
    State ('rms_table',                 'data'),
    State ('rms_table_hidden',          'data'))
def rmsRevertChanges(revert, visible, hidden):
    global global_rms_apply_revert_active
    global_rms_apply_revert_active = True

    app.logger.warning(f"rms: Reverting changes in DataTable")
    _, _ , selected_rows = checkForDuplicities(hidden)

    return hidden, selected_rows, True, True



''' Updates RM Table when row is being added or table being cleared. '''
@app.callback(
    Output('rms_table',              'data',          allow_duplicate=True),
    Output('rms_table',              'selected_rows', allow_duplicate=True),
    Input ('rms_add_row_button',     'n_clicks'),
    Input ('rms_clear_table_button', 'n_clicks'),
    State ('rms_table',              'derived_virtual_selected_rows'),
    State ('rms_table',              'data'))
def update_table(add_clicks, clear_clicks, selected_rows, current_data):
    ctx = callback_context
    if ctx.triggered[0]['prop_id'] == 'rms_add_row_button.n_clicks':
        new_row = newRow(False, '', '', '', '', '', '', '', '')
        current_data.append(new_row)
    elif ctx.triggered[0]['prop_id'] == 'rms_clear_table_button.n_clicks':
        current_data.clear()
        selected_rows.clear()
    return current_data, selected_rows



''' Load RMs .CSV file '''
@app.callback(
    Output('rms_table',                    'data',          allow_duplicate=True),
    Output('rms_table',                    'selected_rows', allow_duplicate=True),
    Output(RMS_WARNING_MODAL,              'is_open'),
    Output('dropdown_jamming_at_location', 'value'),    # Force update in dropdown to re-render graph
    Input('load_rm_file',                  'contents'),
    State('rms_table',                     'data'),
    State('rms_table',                     'selected_rows'),
    State('dropdown_jamming_at_location',  'value'),)
def load_rms_csv_file(contents, current_data, selected_rows, dropdown_jam):
    if contents is None:
        return no_update, no_update, False, no_update

    new_data, new_rows = loadCsvFile(contents, RMS_COLS, f"Selected file is not correct {RMS_NAME} .CSV file.", newRow)

    # if data & rows are None and not 'no_update', .csv file was not correct
    if new_data == None and new_rows == None:
        return no_update, no_update, True, no_update

    # selected new_rows indexes are related to the the new data only, not for the whole dataset
    num_of_rows = len(current_data)
    for r in new_rows:
        selected_rows.append( num_of_rows + r )

    current_data.extend(new_data)
    
    return current_data, selected_rows, False, dropdown_jam



''' Closes Modal Warning Window. '''
@app.callback(
    Output(RMS_WARNING_MODAL, "is_open", allow_duplicate=True),
    Input(f"close-{RMS_WARNING_MODAL}", "n_clicks"))
def close_modal(n):
    if n:
        return False
    return True


''' Close Modal Warning window about RM entry/row not valid. '''
@app.callback(
    Output(RMS_WARNING_VALID,           "is_open", allow_duplicate=True),
    Input(f"close-{RMS_WARNING_VALID}", "n_clicks"))
def close_modal(n):
    if n:
        return False
    return True



''' Callback to update DataTable columns based on selected RMS type.
    If eg. Airplane has only Probability parameter, other columns will not allow save any value. '''
@app.callback(
    Output('rms_table',                 'data',          allow_duplicate=True),
    Output('rms_apply_changes_button',  'disabled',      allow_duplicate=True),
    Output('rms_revert_changes_button', 'disabled',      allow_duplicate=True),
    Input ('rms_table',                 'data'),
    Input ('rms_table',                 'derived_virtual_selected_rows'),
    State ('rms_table',                 'columns'))
def update_columns(initial_data, selected_rows, original_columns):
    global global_rms_apply_revert_active
    if global_rms_apply_revert_active:
        global_rms_apply_revert_active = False
        return no_update, no_update, no_update

    initial_data, _ , _ = checkForDuplicities(initial_data)

    return initial_data, False, False


