import sys
sys.path.append('../Visualiser')

from main_dash import app
from utils     import check_range, loadCsvFile, getModal_wrongCsvFile, getModal_sensorNotValid, convertToNumeric

from dash import Output, Input, State, callback_context, dash_table, dcc, html, no_update
from enum import Enum

import dash_bootstrap_components as dbc
import json

# Globals for Sensors
global global_apply_revert_active   # Apply / Revert disabled button



class SensorType(Enum):
    YN       = "Y/N"
    PPOWER   = "Power"
    AZIMUTH  = "Azimuth"
    POSITION = "Position"

class SensorProperty(Enum):
    ACTIVE = ('active', 'Active', 'boolean')
    ID     = ('id',     'ID',     'text'   )
    TYPE   = ('type',   'Type',   'text'   )
    IP     = ('IP',     'IP',     'text'   )
    PORT   = ('port',   'Port',   'numeric')

    def __new__(cls, value, string_value, value_type):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.value_type = value_type
        obj.string = string_value
        return obj

    @classmethod
    def get_name(cls):
        return "Sensors"


SENSORS_COLS          = len(SensorProperty)
SENSORS_NAME          = 'Sensors'
SENSORS_WARNING_MODAL = 'sensors_modal_wrong_csv'
SENSORS_WARNING_VALID = 'sensors_modal_not_valid'

type_options = [{"label": member.value, "value": member.value} for member in SensorType]

buttons_style      = {'margin-left':'0.5em'}
buttons_style_right= {'margin-left':'auto'}

def layout_sensors_tab(init_data: dict):
    global global_apply_revert_active
    global_apply_revert_active = True

    if init_data == None:
        init_data = []

    init_data, _, _ = checkForDuplicities(init_data)

    return html.Div([

        # Modal Warning Window about wrong CSF file
        getModal_wrongCsvFile(SENSORS_WARNING_MODAL, SENSORS_NAME),

        # Modal Warning Window about invalid Sensor entry/row
        getModal_sensorNotValid(SENSORS_WARNING_VALID),

        # Load CVS file button
        html.Br(),
        html.Div([
            dcc.Upload(
                id='load_sensors_file',
                children=html.Button('Load Sensors .CSV File'),
                multiple=False,
                accept='text/csv'
            ),
            html.Button('Add Sensor',    id='sensors_add_row_button',       style=buttons_style),
            html.Button('Clear Table',   id='sensors_clear_table_button',   style=buttons_style),
            html.Div([
                html.Button('Apply Changes',  id='sensors_apply_changes_button',  style=buttons_style, disabled=True),
                html.Button('Revert Changes', id='sensors_revert_changes_button', style=buttons_style, disabled=True),
            ], style=buttons_style_right)
        ], style={'display': 'flex', 'flexDirection': 'row'}),
        html.Br(),

        dash_table.DataTable(
            id='sensors_table',
            columns=[
                {'name': SensorProperty.ID.value.upper(),   'id': SensorProperty.ID  .value.lower(), 'type': SensorProperty.ID  .value_type},
                {'name': SensorProperty.TYPE.value.title(), 'id': SensorProperty.TYPE.value.lower(), 'type': SensorProperty.TYPE.value_type, 'presentation': 'dropdown'},   # type has to stay this way, otherwise it cannot be changed
                {'name': SensorProperty.IP.value,           'id': SensorProperty.IP  .value.lower(), 'type': SensorProperty.IP  .value_type},
                {'name': SensorProperty.PORT.value.title(), 'id': SensorProperty.PORT.value.lower(), 'type': SensorProperty.PORT.value_type},
            ],
            data=init_data,
            editable=True,
            row_deletable=True,
            row_selectable="multi",
            sort_action="native",
            selected_rows=[i for i, item in enumerate(init_data) if item[SensorProperty.ACTIVE.value]],
            dropdown={
                'type': {
                    'clearable': False,
                    'options': type_options
                }
            },
            style_cell={'textAlign': 'center', 'width': '25%', 'textOverflow': 'ellipsis'},
            css=[{'selector': 'table', 'rule': 'table-layout: fixed'}]
        ),

        # Hidden divs to store JSON representation of the table data and deleted row indices
        html.Div(id='table-data',   style={'display': 'none'}),
        html.Div(id='deleted-rows', style={'display': 'none'}),

        # Alert for checking content of cells
        dbc.Alert(
            "",
            id="sensors_alert",
            dismissable=True,
            is_open=False,
            color="danger",
        ),

        # Dash Store to store data: 1) visible editable 2) hidden original
        dcc.Store(id="sensors_table_hidden", data=init_data)
    ])



def getSensorsConfig(sensors_data: dict, includeNonActive: bool) -> list:
    ''' Lists all 'active' sensors to the structure forwarded to SW Fusion / Redis '''
    def getSensorProperty(data):
        result = None
        try:
            result = { prop.value.lower(): convertToNumeric(data[prop.value.lower()]) if prop.value_type == 'numeric' else data[prop.value.lower()] for prop in SensorProperty }
        except AttributeError:
            app.logger.warning(f'Sensors: Following Sensor row is not valid and not passed to Redis: {data}')
            pass
        return result


    sensors = []
    for data in sensors_data:

        # Only 'active' sensors are added to configuration
        if includeNonActive or data[SensorProperty.ACTIVE.value.lower()]:
            sensor = None
            if includeNonActive:
                sensor = getSensorProperty(data) #{prop.value.lower(): data[prop.value.lower()] for prop in SensorProperty}
            else:
                if data[SensorProperty.ACTIVE.value.lower()]:
                    sensor = getSensorProperty(data) #{prop.value.lower(): data[prop.value.lower()] for prop in SensorProperty}

            if sensor != None:
                sensors.append(sensor)

    return sensors


@app.callback(
    Output('sensors_table_hidden',          'data'),
    Output('sensors_apply_changes_button',  'disabled',      allow_duplicate=True),
    Output('sensors_revert_changes_button', 'disabled',      allow_duplicate=True),
    Output('sensors_table',                 'selected_rows', allow_duplicate=True),
    Output(SENSORS_WARNING_VALID,           'is_open'),
    Output(f'sensor-modal-{SENSORS_WARNING_VALID}', 'children'),
    Input ('sensors_apply_changes_button',  'n_clicks'),
    State ('sensors_table',                 'data'),
    State ('sensors_table_hidden',          'data'),
    State ('sensors_table',                 'selected_rows')
)
def sensorApplyForChanges(apply, visible, hidden, selected_rows):

    # Before updating, check all rows if they have valid entries. If not, and given entry is active - disable it.
    shouldReturnSelectedRowsIndexes = True
    visible, _, _, openModal, modelText, rows_keep_select = processUnselectedRows(selected_rows, visible, shouldReturnSelectedRowsIndexes)

    if openModal:
        return no_update, False, False, rows_keep_select, openModal, modelText

    app.logger.warning(f"Sensors: Applying changes")
    return visible, True, True, rows_keep_select, no_update, no_update



@app.callback(
    Output('sensors_table',                 'data'),
    Output('sensors_table',                 'selected_rows', allow_duplicate=True),
    Output('sensors_apply_changes_button',  'disabled',      allow_duplicate=True),
    Output('sensors_revert_changes_button', 'disabled',      allow_duplicate=True),
    Input ('sensors_revert_changes_button', 'n_clicks'),
    State ('sensors_table',                 'data'),
    State ('sensors_table_hidden',          'data'),
)
def sensorRevertChanges(revert, visible, hidden):
    global global_apply_revert_active
    global_apply_revert_active = True

    app.logger.warning(f"Sensors: Reverting changes in DataTable")
    _, _ , selected_rows = checkForDuplicities(hidden)

    return hidden, selected_rows, True, True



def isSensorEntryValid(sensor):
    ''' Checks for validity of values inside the entry, if all of the are correct. '''
    def get_title_empty(prop):
        return f"{prop.string} is empty"

    sensor_type = sensor[SensorProperty.TYPE.value.lower()] != None and sensor[SensorProperty.TYPE.value.lower()] != ""
    identifier  = sensor[SensorProperty.ID  .value.lower()] != None and sensor[SensorProperty.ID  .value.lower()] != ""
    ip_addr     = sensor[SensorProperty.IP  .value.lower()] != None and sensor[SensorProperty.IP  .value.lower()] != ""
    port        = sensor[SensorProperty.PORT.value.lower()] != None and sensor[SensorProperty.PORT.value.lower()] != ""

    result = sensor_type and identifier and ip_addr and port

    indexes = []
    if not sensor_type:
        indexes.append(get_title_empty(SensorProperty.TYPE))
    if not identifier:
        indexes.append(get_title_empty(SensorProperty.ID))
    if not ip_addr:
        indexes.append(get_title_empty(SensorProperty.IP))
    if not port:
        indexes.append(get_title_empty(SensorProperty.PORT))

    # Check Ports range
    wrong_range = check_range([], SensorProperty.PORT, sensor, [1,65535], unit="", get_text_only=True)
    range_ok = wrong_range == None
    if not range_ok:
        indexes.append(wrong_range)

    return result and range_ok, indexes


def stringifySensor(sensor, invalid_indexes):
    properties = ', '.join(invalid_indexes)
    return html.Div([html.B("ID: "), f"{sensor[SensorProperty.ID.value.lower()]} - ", html.B("Invalid properties: "), f"{properties}"])


def processUnselectedRows(selected_rows, sensor_data, shouldReturnSelectedRowsIndexes):
    ''' Process unselecting of rows, which have to propagate to SensorProperty.ACTIVE 
        In case rows is selected, content is tested if it has all properties correct. If not,
        modal is raised and that selected row is unselected.
    '''
    openModal = False
    not_valid_sensors = []
    rows_keep_select = []

    for i,sensor in enumerate(sensor_data):
        sensor_data[i][SensorProperty.ACTIVE.value.lower()] = i in selected_rows

        sensor_valid, invalid_indexes = isSensorEntryValid(sensor)
        if i in selected_rows and not sensor_valid:
            # Open Modal to inform user
            openModal = True
            not_valid_sensors.append(stringifySensor(sensor, invalid_indexes))

        # Keep selected ones
        if i in selected_rows:
            rows_keep_select.append(i)

    if shouldReturnSelectedRowsIndexes:
        return sensor_data, True, True, openModal, not_valid_sensors, rows_keep_select
    else:
        return sensor_data, True, True, openModal, not_valid_sensors



# @app.callback(
#     Output('sensors_table',                 'data',          allow_duplicate=True),
#     Output('sensors_apply_changes_button',  'disabled',      allow_duplicate=True),
#     Output('sensors_revert_changes_button', 'disabled',      allow_duplicate=True),
#     Output(SENSORS_WARNING_VALID,           'is_open'),
#     Output(f'sensor-modal-{SENSORS_WARNING_VALID}', 'children'),
#     Input( 'sensors_table',                 'derived_virtual_selected_rows'),
#     State( 'sensors_table',                 'data')
# )
# def processUnSelectedRow(selected_rows, sensor_data):
#     shouldReturnSelectedRowsIndexes = False
#     return processUnselectedRows(selected_rows, sensor_data, shouldReturnSelectedRowsIndexes)



def checkForDuplicities(sensors_init_data):
    ''' Checks for duplicity names and if any found, later name is modified with id
        and disabled from SW Fusion and Visualisation. '''
    unique_ids = []
    items_to_modify = {}
    updated_ids = []

    if sensors_init_data == None:
        sensors_init_data = []

    for i, d in enumerate(sensors_init_data):
        identifier = d[SensorProperty.ID.value]
        if identifier not in unique_ids:
            unique_ids.append(identifier)
        else:
            items_to_modify.update({i: d})

    # Duplicity items - append id to the name AND disable it from visualisation and SW Fusion
    for key,val in items_to_modify.items():
        sensors_init_data[key][SensorProperty.ID.value]     = f"{key}_{sensors_init_data[key][SensorProperty.ID.value]}"
        sensors_init_data[key][SensorProperty.ACTIVE.value] = False

        updated_ids.append( sensors_init_data[key][SensorProperty.ID.value] )

    # Update list of selected rows as well
    selected_rows = [i for i, item in enumerate(sensors_init_data) if item[SensorProperty.ACTIVE.value]]

    return sensors_init_data, updated_ids, selected_rows



@app.callback(
    Output('sensors_table', 'data',          allow_duplicate=True),
    Output('sensors_apply_changes_button',  'disabled', allow_duplicate=True),
    Output('sensors_revert_changes_button', 'disabled', allow_duplicate=True),
    Input( 'sensors_table',                 'data'),
    Input( 'sensors_table',                 'derived_virtual_selected_rows'))
def update_table(data, selected_rows):
    ''' Updates table when data being changed. Mostly for checking for duplicities.
        Triggers -ONLY- if tha data changed - in case something is rewritten into same content, this will not tigger.
    '''
    # In case DataTable is reverted, the table is updated, but we do not want enable apply/revert buttons
    global global_apply_revert_active
    if global_apply_revert_active:
        global_apply_revert_active = False
        return no_update, no_update, no_update

    data, _ , selected_rows = checkForDuplicities(data)
    return data, False, False



def newRow(active, identifier, typ, ip, port):
    ''' Return new sensor row. '''
    return {
        SensorProperty.ACTIVE.value.lower(): active,
        SensorProperty.ID    .value.lower(): str(identifier),
        SensorProperty.TYPE  .value.lower(): str(typ),
        SensorProperty.IP    .value.lower(): str(ip),
        SensorProperty.PORT  .value.lower(): port
    }



''' Updates table when new row is being added. '''
@app.callback(
    Output('sensors_table',              'data',          allow_duplicate=True),
    Output('sensors_table',              'selected_rows', allow_duplicate=True),
    Input ('sensors_add_row_button',     'n_clicks'),
    Input ('sensors_clear_table_button', 'n_clicks'),
    State ('sensors_table',              'derived_virtual_selected_rows'),
    State ('sensors_table',              'data'))
def update_table_add_clear(add_clicks, clear_clicks, selected_rows, current_data):
    ctx = callback_context
    if ctx.triggered[0]['prop_id'] == 'sensors_add_row_button.n_clicks':
        new_row = newRow(False, '','','','')
        current_data.append(new_row)
    elif ctx.triggered[0]['prop_id'] == 'sensors_clear_table_button.n_clicks':
        current_data.clear()
        selected_rows.clear()
    return current_data, selected_rows



''' Load Sensors .CSV file '''
@app.callback(
    Output('sensors_table', 'data',          allow_duplicate=True),
    Output('sensors_table', 'selected_rows', allow_duplicate=True),
    Output(SENSORS_WARNING_MODAL, 'is_open'),
    Input('load_sensors_file', 'contents'),
    State('sensors_table', 'data'),
    State('sensors_table', 'selected_rows'),
)
def load_sensors_csv_file(contents, current_data, selected_rows):
    if contents is None:
        return no_update, no_update, False

    new_data, new_rows = loadCsvFile(contents, SENSORS_COLS, f"Selected file is not correct {SENSORS_NAME} .CSV file.", newRow)
    
    # if data & rows are None and not 'no_update', .csv file was not correct
    if new_data == None and new_rows == None:
        return no_update, no_update, True

    if new_data != no_update and new_rows != None:
        new_data, _, rows = checkForDuplicities(new_data)
    
    # selected new_rows indexes are related to the the new data only, not for the whole dataset
    num_of_rows = len(current_data)
    for r in new_rows:
        selected_rows.append( num_of_rows + r )

    current_data.extend(new_data)

    return current_data, selected_rows, False



''' Close Modal Warning window about file being not ocrrect. '''
@app.callback(
    Output(SENSORS_WARNING_MODAL,           "is_open", allow_duplicate=True),
    Input(f"close-{SENSORS_WARNING_MODAL}", "n_clicks"),
)
def close_modal(n):
    if n:
        return False
    return True



''' Close Modal Warning window about sensor entry/row not valid. '''
@app.callback(
    Output(SENSORS_WARNING_VALID,           "is_open", allow_duplicate=True),
    Input(f"close-{SENSORS_WARNING_VALID}", "n_clicks"),
)
def close_modal(n):
    if n:
        return False
    return True
