import sys
sys.path.append('../Visualiser')

from main_dash import app

from dash                       import Output, Input, State, Patch, dash_table, dcc, html, no_update, ALL, MATCH
from dash.exceptions            import PreventUpdate
from datetime                   import datetime

from appEnums                   import AppTabs, ConfigEnum, ConfigStrings, FlipEnum, LimitsEnum, RedisEnum, RequestEnums, TabsEnum
from marker                     import Marker, MarkerIcon, MarkerColor, MarkerType, MarkerProperty
from png_generator              import ColorIntensity
from units                      import PowerUnits, HeightUnits, GnssUnits
from utils                      import (convertValueDeg2Rad,
                                        create_geojson_polygons,
                                        getFlipImageVars,
                                        getJammingIndex,
                                        getJammingIndexes,
                                        getPolygonCoords,
                                        getSliderStep_byRads,
                                        fillWithZeros,
                                        fillWithZeros3D,
                                        isPointInsideTile,
                                        prepare_geodata,
                                        prepareVars,
                                        split_interval,
                                        updateMarkersWithTileIndex,
                                        updateOneMarkerWithTileIndex,
                                        attempt_create_int,
                                        is_integer )
from views.configuration        import getRedisConfig, getProcessingConfig, getLimitsConfig, getGnssLimits
from views.rms                  import RmsProperty, getRmsByDeviceType
from views.sensors              import getSensorsConfig
from views.view                 import getFigureLayouWithAnnotation, getFigureLayouWithoutAnnotation, getFigureLayouWithoutAnnotationWithAxis, getIntegrationLimits

import globals

import dash_bootstrap_components as dbc
import dash_daq                  as daq
import dash_leaflet              as dl
import datetime
import json
import math
import numpy                     as np
import os
import plotly.graph_objects      as go
import time


# Global variables
queue_request = None
queue_data    = None    # Queue for multiprocessing
flipGpsImage  = {}      # Based on GPS position, image with jamming data might flip along some axis
shouldRender  = True    # Rendering play or pause
geodata       = None    # Geodata with tiles, for easier obtaining jamming intensity at location
jam_map_limits= None    # Need to store dimension limits of Jamming Data for Map - when it will change 
                        #   during runtime, geodata will have to be updated as well.
redis_updated = False   # For User notification - REDIS updated, but no new data from Fusion - 'Recalculating ...'

# Method 'updateMarkersWithTileIndex()' called in 'tickLogic()'' takes more time than 'interal_logic'
#   dcc.Interval itself  - this will cause computed data are not Outputed via Dash Callback - new data are 
#   just thrown away. Thus, global variable has to be used to store the new data.
global_tiles_index  = None      # list of tile indexes
global_devices_data = None      # dict of devices
global_fusion_time  = None      # previous processed fusion time
global_jam_data_previous = None     # Previous jam specific data

APP_CONFIG_FILE = '_app_config.json'


# Define the OpenStreetMap tile source
url         = "https://a.tile.openstreetmap.org/{z}/{x}/{y}.png"
attribution = '&copy; <a href="https://openstreetmap.org/">OSM</a> '


def createColorbar(min_val, max_val):
    colorscale     = createColorbarScale(min_val, max_val)
    colorbar_style = dict(weight=0, opacity=1, color='black', dashArray='3', fillOpacity=0.3)
    return dl.Colorbar(id='map_colorbar', colorscale=colorscale, width=20, height=150, min=0, max=0.0001, position="bottomright", style=colorbar_style)

def createColorbarScale(min_val, max_val):
    colormap = ColorIntensity(min_val, max_val)
    return [colormap.intensity_to_hex_color(x) for x in np.linspace(min_val, max_val, 100)]


def createGeodata(limits):
    global flipGpsImage, jam_map_limits

    jam_map_limits = limits     # Store for next update
    lon_min, lon_max, lon_steps, lat_min, lat_max, lat_steps = limits
    data_points = prepare_geodata(lon_min, lon_max, lon_steps, lat_min, lat_max, lat_steps)

    # Will need to keep the geodata as it is simpler to check which Device/Marker is inside a specific tile
    geodata      = create_geojson_polygons(data_points, lon_steps, lat_steps)
    flipGpsImage = getFlipImageVars(lon_min, lon_max, lat_min, lat_max)
    image_bounds = [[lat_min, lon_min], [lat_max, lon_max]]
    return geodata, image_bounds


def getMarkerDataAsJson(geodata, marker_data):
    ''' Returns list of Markers in json format - by default Marker instance is not dictionary,
        so all Markers has to be tranformed to dictionary to be able to store them in dcc.Store '''
    marker_data, tiles_w_device = updateMarkersWithTileIndex(geodata, marker_data)
    return Marker.marker_to_dict(marker_data), tiles_w_device


def createSlider_marks(n_lim, p_lim, slider_step, steps, can_be_int):
    ''' Creates marks for slider based on min/max values of the slider.
        IMPORTANT NOTE - marks HAVE to be the same multplier of the slider_step, otherwise it will bring problems
        for Limits with small steps - eg. for Lat steps 10 and setup with slider to display only 2 tiles, it might show 3 tiles
        because of rounding error in marks position and slider step value.
    '''

    if any(value is None for value in (n_lim, p_lim)):
        return {}

    num_marks = 5.0

    # In case # of steps is smaller then 5, dipslay only # of steps (5 steps is good viewing trade-off)
    if (steps / num_marks) <= 1.0:
        marks_count = steps
        stepping = True            # Way to tell if we should not make marks one by one, as we do not have enough steps to draw as marks
    else:
        marks_count = round(steps/num_marks)
        stepping = False

    n_lim = attempt_create_int(n_lim)
    p_lim = attempt_create_int(p_lim)

    mid_lim     = [0] * (int(num_marks)+2 if not stepping else steps+1)     # +2 to accomodate first and last value, +1 to accomodate only first, as we are 'stepping'
    mid_lim[0]  = n_lim
    mid_lim[-1] = p_lim

    for i in range(len(mid_lim)-2):

        if stepping:
            marks_count = 1

        multplier = (i+1)*marks_count
        mid_lim[i+1] = attempt_create_int(n_lim + multplier*slider_step)

    is_int = (is_integer(n_lim) and is_integer(p_lim)) if can_be_int else False

    mids = { mid_i: f"{float(mid_i):.3f}" if not is_int else f"{round(mid_i)}" for mid_i in mid_lim }

    return mids



def createSlider(identifier, index, n_lim, p_lim, min_value, max_value, steps, width_slidr, can_be_int):

    # In case of GNSS Limits, step has to be calculated from original (and SW Fusion used) Radians, not from Degrees.
    if index == LimitsEnum.LATITUDE.value or index == LimitsEnum.LONGITUDE.value:
        slider_step = getSliderStep_byRads(n_lim, p_lim, steps)
    else:
        slider_step = math.ceil( ((p_lim - n_lim) / steps)*1e5 ) / 1e5

    return html.Div([
        dcc.RangeSlider(id=identifier, #{"type": "integrate_view_slider", "index": index},
            min=n_lim,
            max=p_lim,
            value=[float(min_value), float(max_value)],
            marks=createSlider_marks(n_lim, p_lim, slider_step, steps, can_be_int),
            # marks={n_lim: f"{float(n_lim):.3f}", mid_lim: f"{float(mid_lim):.3f}", p_lim: f"{float(p_lim):.3f}"},
            # marks={n_lim: f"{float(n_lim):.3f}", p_lim: f"{float(p_lim):.3f}"},
            tooltip={"placement": "bottom", "always_visible": False},
            step=slider_step,
            pushable=slider_step/1.5)   # Ensures sliders cannot overlap
    ])



def createRow_SliderValue(value):
    ''' Formats value based on float or int type. '''
    return f"{float(value):.3f}" if not is_integer(value) else f"{round(value)}"



def createLayout_Sliders(initial_data, limits_init_data):
    ''' Returns layout with sliders for limits. '''

    def createToggleSwitch(isMin, index):
        return daq.ToggleSwitch(id={"type": f"integrate_view_toggle_{'min'if isMin else 'max'}", "index": index}, value=False, color="green", style={'justify-content': 'center'})

    def createInput(name, index, value, width_input):
        return dcc.Input(type="number", id={"type": f"integrate_view_input_{name}", "index": index}, value=value, style={'width':width_input, 'text-align':'center'})

    def createDropdown(unitEnum, initial_value, index, width_ddown):
        unit_list = [x.value for x in unitEnum]
        return dcc.Dropdown(unit_list, initial_value if initial_value != None else unit_list[0],
            id={"type": "integrate_units", "index": index},
            clearable=False,
            style={'width':width_ddown}),


    def createRow(name, index, unitEnum, init_unit, min_lim, max_lim, steps, values, can_be_int):
        width_name  = '150px'
        width_ddown = '5vw'
        width_slidr = '35vw'
        width_input = '7vw'
        width_value = '75px'

        # unit      = init_unit[index] if init_unit[index] != None else None
        min_limit  = min_lim[index]   if min_lim != None and min_lim[index]   != None else -1e30
        max_limit  = max_lim[index]   if max_lim != None and max_lim[index]   != None else -1e30
        min_value  = values[index][0] if values  != None and values[index][0] != None else min_limit
        max_value  = values[index][1] if values  != None and values[index][1] != None else max_limit
        step_value = steps[index]     if steps   != None                              else 100 

        return html.Tr([
            html.Td(f"{name}",                                                                                                                                                                           style={'width':width_name,  'text-align': 'right', 'padding':'0em 0em 2em 0em'}),
            html.Td(createRow_SliderValue(min_value), id={"type": "integrate_view_value_min", "index": index},                                                                                           style={'width':width_value, 'text-align': 'right', 'padding':'0em 0em 2em 0em'}),
            html.Td(         createSlider(               {"type": "integrate_view_slider",    "index": index}, index,  min_limit, max_limit, min_value, max_value, step_value, width_slidr, can_be_int), style={'width':width_slidr, 'text-align': 'center'}),
            html.Td(createRow_SliderValue(max_value), id={"type": "integrate_view_value_max", "index": index},                                                                                           style={'width':width_value, 'text-align': 'left', 'padding':'0em 0em 2em 0em'})
            # html.Td(createDropdown(unitEnum, unit, index, width_ddown),                                        style={'width':width_ddown, 'text-align': 'center'}),
        ])

    # -------------------------------------------------------------------------
    init_unit     = None
    min_lim       = None
    max_lim       = None
    slider_values = None
    steps         = None
    
    if initial_data != None:
        # init_unit     = initial_data['units']
        min_lim       = initial_data['limit_mins']
        max_lim       = initial_data['limit_maxs']
        slider_values = initial_data['slider_values']

    if limits_init_data != None:
        steps = {
            LimitsEnum.LATITUDE.value:  limits_init_data[ConfigStrings.LAT.value]  [ConfigStrings.STEPS.value],
            LimitsEnum.LONGITUDE.value: limits_init_data[ConfigStrings.LON.value]  [ConfigStrings.STEPS.value],
            LimitsEnum.HEIGHT.value:    limits_init_data[ConfigStrings.HGT.value]  [ConfigStrings.STEPS.value],
            LimitsEnum.POWER.value:     limits_init_data[ConfigStrings.POWER.value][ConfigStrings.STEPS.value]
        }

    return dbc.Row([
        html.Table([
            # html.Thead([
            #     # html.Th([]),
            #     # html.Th('Units',    style={'justify-content':'center', 'text-align':'center'}),
            #     # html.Th(['Min'],    style={'justify-content':'center', 'text-align':'center'}),
            #     html.Th(['Limits'], colSpan=2, style={'justify-content':'center', 'text-align':'center'}),
            #     # html.Th(['Max'],    style={'justify-content':'center', 'text-align':'center'}),
            # ]),
            html.Tbody([
                createRow(  f"{LimitsEnum.LATITUDE.string} [{LimitsEnum.LATITUDE.unit}]: ", LimitsEnum.LATITUDE.value,  GnssUnits,   init_unit, min_lim, max_lim, steps, slider_values, LimitsEnum.LATITUDE.can_be_int ),
                createRow(f"{LimitsEnum.LONGITUDE.string} [{LimitsEnum.LONGITUDE.unit}]: ", LimitsEnum.LONGITUDE.value, GnssUnits,   init_unit, min_lim, max_lim, steps, slider_values, LimitsEnum.LONGITUDE.can_be_int),
                createRow(      f"{LimitsEnum.HEIGHT.string} [{LimitsEnum.HEIGHT.unit}]: ", LimitsEnum.HEIGHT.value,    HeightUnits, init_unit, min_lim, max_lim, steps, slider_values, LimitsEnum.HEIGHT.can_be_int   ),
                createRow(        f"{LimitsEnum.POWER.string} [{LimitsEnum.POWER.unit}]: ", LimitsEnum.POWER.value,     PowerUnits,  init_unit, min_lim, max_lim, steps, slider_values, LimitsEnum.POWER.can_be_int    ),
            ])
        ], style={'margin': '1em 0em 0em -2em'})
    ])

def createLimitsDropdown(initial_item):
    ''' Returns simple dropdown with all limits.'''
    limits = []
    for el in LimitsEnum:
        if el != LimitsEnum.GRAPH_PROBAB_DENSITY:
            limits.append(el.string)

    # Default limit is Power (at least by documentation for SW Fusion)
    return dcc.Dropdown(limits, initial_item.string, id="dropdown_2d_probability_density", clearable=False),

def createOneSliderRow(initial_data, limits_init_data):
    ''' For Power Probability Density Graph.'''
    def createRow(slider_index, initial_item, min_lim, max_lim, steps, values, can_be_int):
        # width_slidr = '41vw'
        # width_drop  = '5vw'
        width_slidr = '40vw'
        width_drop  = '110px'

        index = initial_item.value

        min_limit  = min_lim[index]   if min_lim != None and min_lim[index]   != None else -1e30
        max_limit  = max_lim[index]   if max_lim != None and max_lim[index]   != None else -1e30
        min_value  = values[index][0] if values  != None and values[index][0] != None else min_limit
        max_value  = values[index][1] if values  != None and values[index][1] != None else max_limit
        step_value = steps[index]     if steps   != None                              else 100 

        return html.Tr([
            html.Td(  createSlider({"type": "integrate_view_slider",    "index": slider_index}, slider_index,  min_limit, max_limit, min_value, max_value, step_value, width_slidr, can_be_int), style={'width':width_slidr, 'text-align': 'center'}),
            html.Td(  createLimitsDropdown(initial_item),    style={'width':width_drop, 'text-align': 'center'}),
        ])
    # -------------------------------------------------------------------------
    min_lim       = None
    max_lim       = None
    slider_values = None
    steps         = None

    if initial_data != None:
        min_lim       = initial_data['limit_mins']
        max_lim       = initial_data['limit_maxs']
        slider_values = initial_data['slider_values']

    if limits_init_data != None:
        steps = {
            LimitsEnum.LATITUDE.value:  limits_init_data[ConfigStrings.LAT.value] [ConfigStrings.STEPS.value],
            LimitsEnum.LONGITUDE.value: limits_init_data[ConfigStrings.LON.value] [ConfigStrings.STEPS.value],
            LimitsEnum.HEIGHT.value:    limits_init_data[ConfigStrings.HGT.value] [ConfigStrings.STEPS.value],
            LimitsEnum.POWER.value:     limits_init_data[ConfigStrings.POWER.value][ConfigStrings.STEPS.value]
        }

    return html.Div([
        html.Table([
            html.Tbody([
                # Initial item is Power
                createRow(LimitsEnum.GRAPH_PROBAB_DENSITY.value, LimitsEnum.POWER, min_lim, max_lim, steps, slider_values, can_be_int=True) # if Power is initial, it can_be_int
            ])
        ], style={'margin': '1em 0em 0em 2em'})
    ])


def createRMsWindow():
    return html.Div([
        dcc.Store(id="visual_rm_table_original", data=[]),
        # html.H4("RMs", style={"margin-top":"1em"}),
        dash_table.DataTable(
            id='visual_rm_table',
            columns=[
                {'name': 'Name',  'id': 'Name'},
                {'name': 'RM ID', 'id': 'RM_ID'},
                {'name': 'RMs',   'id': 'RMs'},
                {'name': 'Color', 'id': 'Color'}
            ],
            hidden_columns=['Color'],
            data=[],
            editable=False,
            row_deletable=True,
            style_cell={'cursor': 'pointer'},  # Add cursor pointer to indicate clickability
            style_data={'whiteSpace': 'normal', 'height': 'auto', 'textAlign':'center'},
            style_data_conditional=[{'if': {'column_id': 'RMs'}, 'width': '58vh','textAlign':'left'}],
            style_header={'textAlign':'left'},
            page_size=10,  #   Number of rows visible at once
            css=[{"selector": ".show-hide", "rule": "display: none"}]   # hide 'Toggle Columns' button, enabled be attribute 'hidden_columns'
        ),
    ], style={'margin-top':'3em'})


def getPopupRmsTableRow(name, prob_actual, prob_limit, time_current, time_jammed_since, time_limit):
    ''' time_* - seconds '''
    def getColorText(text, color):
        return html.Span(text, style={'color':color})

    def checkNumber(value, mult=1):
        ''' Checks for number - if already float, nothing happens. If textual number, returns float representation. Otherwise, returns empty string.'''
        if type(value) == float:
            return value * mult
        if type(value) == str:
            if value.replace(".","",1).isdigit():
                return float(value) * mult
        else:
            return None

    # In case probabilitites and times were for some reason as strings
    prob_actual       =       float(prob_actual) if prob_actual != None else 0.0
    prob_limit        = checkNumber(prob_limit)
    time_jammed_since =       float(time_jammed_since) if time_jammed_since != None else None
    time_limit        = checkNumber(time_limit, mult=60)    # Holdover limits are in minutes!!! Thus *60
    time_current      =       float(time_current)
    
    str_prob_actual = getColorText(f"{prob_actual:.3f}", ('green' if prob_actual < prob_limit else 'red') )
    str_prob_limit  = getColorText(f" / {prob_limit:.3f}".rstrip('0'), 'black')
    str_time_actual, str_time_limit = '', ' / '

    if time_jammed_since != None and time_limit != None:
        str_time_actual = getColorText(f"{formatTime(time_current-time_jammed_since)}", ('green' if (time_current-time_jammed_since) < time_limit else 'red') )
    elif time_jammed_since != None:
        str_time_actual = getColorText(f"{formatTime(time_current-time_jammed_since)}", 'black')
    
    if time_limit != None:
        str_time_limit  = getColorText(f" / {formatTime(time_limit)}", 'black')
    if time_jammed_since == None:
        str_time_actual = "-:--:--"
    if time_limit == None:
        str_time_limit = " / -:--:--"

    return getPopupRmsTableRow_htmlRow(name, [str_prob_actual, str_prob_limit], [str_time_actual, str_time_limit])


def getPopupRmsTableRow_htmlRow(name, prob, time):
    style = {'padding-left': '0.5em', 'text-align':'center'}
    return html.Tr([
        html.Td(name),
        html.Td(prob, style=style),
        html.Td(time, style=style)
    ])


def getMarkerPopupContent(name, device_type, intensity, sensitivity=None, probability=None, probability_limits=None, time_jammed=None, time_current=None, time_limits=None):
    ''' Creates String with Name and Intensity of a device. '''
    str_intensity   = f"Intensity:   {intensity:.5f}"           if intensity   != None else "No intensity"
    str_sensitivity = f"Sensitivity: {sensitivity:.5f}"         if sensitivity != None else "No sensitivity"
    str_probability = f"Probability: {probability:.5f}"         if probability != None else "No probability"

    index = 0
    rows, time_jam = [], []
    style={'text-align':'center'}
    appended_at_least_one = False
    
    if time_jammed != None:
        time_jam.extend([html.Br(), html.Br(), html.B("Time Jammed")])
        rows = [html.Tr([html.Th("RMs",style=style), html.Th("P",style=style), html.Th("Time",style=style)])]

        for rms_id,time_jammed_since in time_jammed.items():
            try:
                if len(probability_limits) > index and len(time_limits) > index:
                    rows.append( getPopupRmsTableRow(rms_id, probability, probability_limits[index], time_current, time_jammed_since, time_limits[index]) )
                    appended_at_least_one = True
            except IndexError as e:
                app.logger.warning(f"visualisation.py::getMarkerPopupContent()")
                app.logger.warning(f"    Error: {e}")
                app.logger.warning(f"      index:              {index}")
                app.logger.warning(f"      probability_limits: {probability_limits}")
                app.logger.warning(f"      time_limits:        {time_limits}")
            index += 1

    # In case no actual Time Jamed info we will get empty 'table' which might confuse user - add 'empty' row with '-'
    time_jammed_html = []
    if appended_at_least_one:
        table = html.Table(rows, style={'font-size':'inherit'})
        time_jammed_html = time_jam + [table]

    return [html.B(f"{name} / {device_type}"), html.Br(), 
            # f"{str_intensity}",                html.Br(),
            f"{str_sensitivity}",              html.Br(),
            f"{str_probability}"] + time_jammed_html



def createDashLeafMarker(marker_id, name, typ, color, lat, lon):
    return dl.Marker(
                position=[lat, lon],
                icon=MarkerIcon(MarkerType.from_string(typ), MarkerColor.from_string(color)).getProperties(),
                children=[
                    dl.Popup(
                        children=[
                            html.Div(getMarkerPopupContent(name, typ, None), id={"type": "marker_intensity", "index": marker_id})
                        ]),
                    html.Div(id=f"marker_{marker_id}", style={"display":"None"})
                ]
            )


def fillDashMarkers(geodata, devices_init_data, updateMarkersTileIndex=False) -> list:
    ''' Takes devices data and makes Dash Leaflet Markers from them. '''
    if devices_init_data == None:
        devices_init_data = []

    # Initial marker_data has to be udpated with tileIndex !
    if updateMarkersTileIndex:
        devices_init_data, tiles_w_device = updateMarkersWithTileIndex(geodata, devices_init_data)

    dash_markers = []
    for marker_id, marker in enumerate(devices_init_data):

        # Add only markers/devices which are enabled to fusion
        if marker[MarkerProperty.ENABLED_FUSION.value]:
            dash_markers.append( createDashLeafMarker(
                                    marker_id,
                                    marker[MarkerProperty.NAME.value],
                                    marker[MarkerProperty.TYP.value],
                                    marker[MarkerProperty.COLOR.value],
                                    marker[MarkerProperty.LAT.value],
                                    marker[MarkerProperty.LON.value] ) )

    return dash_markers


def getActiveDevices(devices_init_data) -> list:
    ''' Returns list of names of active devices. '''
    if devices_init_data == None:
        devices_init_data = []

    result = []
    for d in devices_init_data:
        if d[MarkerProperty.ENABLED_FUSION.value]:
            result.append( d[MarkerProperty.NAME.value] )
    return result


def getPlayPauseHtml(isPlay: bool):
    icon_name = "play.svg" if isPlay else "pause.svg"
    return html.Img(src=os.path.join("assets", "icons", icon_name),  width="15px", height="15px", style={'margin': '0px 10px 3px 10px'})
    

def getConnectionOnOffHtml(isOn: bool):
    icon_name  = "connection.svg" if isOn else "connection_no.svg"
    icon_alt   = "Connection to Redis is active" if isOn else "Connection to Redis is not active"
    icon_title = icon_alt
    return html.Img(src=os.path.join("assets", "icons", icon_name), alt=icon_alt, title=icon_title, width="20px", height="20px", style={'margin': '0px 10px 3px 10px'})
        


def get_probab_jam_power_layout(limit, x_range):
    return go.Layout(
            margin=go.layout.Margin(l=50, r=130, b=50, t=10),
            xaxis_title=f"{limit.string} [{limit.unit}]",
            yaxis_title="pmf [1]",
            xaxis_range=x_range
        )

def getWhiteMarker():
    ''' For Probability in Legend, the line and color are distracting, thus making them White.'''
    return dict(color='White', size=10, line=dict(color='White', width=1))


def init(init_queue_request, init_queue_data):
    ''' Initialisation, mostly for multiprocessing.Manager to pass jamming data
        from jamming_data_process to Dash app '''
    global queue_request, queue_data
    queue_request = init_queue_request
    queue_data    = init_queue_data
    

def layout_visualisation_tab(devices_init_data, initial_app_data, queue_request, queue_data, limits, limits_init_data):
    ''' Main Layout for Visualisation Tab. '''
    global jam_map_limits
    jam_map_limits = limits

    # Init all global variables
    init(queue_request, queue_data)

    geodata, image_bounds = createGeodata(limits)

    initial_dash_markers = fillDashMarkers(geodata, devices_init_data)

    devices_jamming_signal_list = getActiveDevices(devices_init_data)

    # Temporary solution - return updated markers with tileIndex and Layout
    #   tileIndex is needed in Device Tab, which is initiated after this Layout Tab (see app.py createLayout())
    marker_data, tiles_w_device = getMarkerDataAsJson(geodata, devices_init_data)


    return marker_data, dbc.Row([

        # Store to access marker_data
        dcc.Store(id='stored_marker_data', data=marker_data),

        # Store to access list of tiles with device in it
        dcc.Store(id='stored_tiles_w_device', data=tiles_w_device),

        # Store jamming signal at location
        dcc.Store(id='stored_jamming_at_location'),

        # Store data for probability density graph
        dcc.Store(id='stored_2d_probability_density'),

        # Store number of devices globally - so the id of device is unique no matter how many devices
        #   was deleted & updated & deleted and so on
        dcc.Store(id="stored_global_device_id", data=len(devices_init_data)),

        # Store Geodata tiles for easier obtaining of jamming intensity at location
        dcc.Store(id='store_geodata', data=geodata),

        # Store min and max values of Map Colorscale/Colorbar
        dcc.Store(id='store_colorbar_min_max', data=[0,0.0001]),

        # Left Column with map only
        dbc.Col([
         
            # Map and lat vertical
            dbc.Row(
                [
                    dbc.Col(
                        html.Div([
                            # Setup a map with the edit control.
                            dl.Map(
                                # center=[49.74375, 15.338639],  # Stred CR
                                # center=[50.0844,  14.4231],    # Praha
                                center=[50.1037708353729, 14.278326732305842], # LKPR
                                zoom=8,
                                children=[
                                    dl.TileLayer(url=url, maxZoom=20, attribution=attribution),
                                    createColorbar(0,0.00001),    # default min max values are 0,0 - updated below
                                    dl.ImageOverlay(opacity=0.6, url=os.path.join('assets', 'icons', 'empty_map_jam.png'), bounds=image_bounds, id="jamming_image"),
                                    dl.FeatureGroup(children=initial_dash_markers, id="dash_markers_group"),
                                ],
                                bounds=image_bounds,
                                id="map",
                                style={'width': '100%', 'height': '60vh'}
                            ),
                        ]),
                        width=12,
                    )
                ]
            ),     

            # Sliders
            createLayout_Sliders(initial_app_data, limits_init_data),
        ], width=6),

        # Right Column with 2d graphs
        dbc.Col([

            # html.H4("Devices", style={"text-align":"center", "margin-top":"0.5em"}),

            # Dropdown with list of active devices, so the jamming signal density at given location can be displayed
            dcc.Dropdown(devices_jamming_signal_list, devices_jamming_signal_list[0] if len(devices_jamming_signal_list) > 0 else "", id="dropdown_jamming_at_location", clearable=False),

            # 2d Graph of "Jamming Signal Probability Distribution at device location"
            #   "Hustota pravdepodobnosti rusiciho signalu v miste zarizeni"
            dcc.Graph(
                id='graph_2d_probab_jam_signal',
                config={
                    'displayModeBar': False,
                    'displaylogo': False
                },
                figure=go.Figure(data=[
                        go.Bar(                        name='Probability', showlegend=True),    # Power Probability Graph/Bar
                        go.Scatter(                    name='Sensitivity', showlegend=True),    # Device Sensitivity
                        go.Scatter(x=[None], y=[None], name='P = 0.000',   showlegend=True, marker=getWhiteMarker())     # Calculated Probability - more in getDataJammingAtLocation()
                    ],
                    layout=get_probab_jam_power_layout(LimitsEnum.POWER, [ initial_app_data['limit_mins'][LimitsEnum.POWER.value], initial_app_data['limit_maxs'][LimitsEnum.POWER.value] ])),
                style={'height': '20vh'} 
            ),

            # 2d Graph of "Jammer Power Probability Distribution"
            #   "Hustota pravdepodobnosti vykonu rusicky"
            dcc.Graph(
                id='graph_2d_probab_jam_power',
                config={
                    'displayModeBar': False,
                    'displaylogo': False
                },
                figure=go.Figure(data=[
                        go.Bar(                        name='Probability', showlegend=True),
                        go.Scatter(                    name="Min",         showlegend=True),
                        go.Scatter(                    name="Max",         showlegend=True),
                        go.Scatter(x=[None], y=[None], name='P = 0.000',   showlegend=True, marker=getWhiteMarker())
                    ],
                    layout=get_probab_jam_power_layout(LimitsEnum.POWER, [ initial_app_data['limit_mins'][LimitsEnum.POWER.value], initial_app_data['limit_maxs'][LimitsEnum.POWER.value] ])),
                style={'height': '20vh'} 
            ),

            createOneSliderRow(initial_app_data, limits_init_data),

            # RMs Table
            createRMsWindow(),

            # Dummy Div
            html.Div(id="child_map_idv")
        ], width=6),
    ])



def layout_timebar():
    # Timestamp at the bottom
    return dbc.Row([
        html.Div([
                html.Button([getPlayPauseHtml(isPlay=False)], id="play_pause_sim"),
                html.P("Time: ", id="time_interval", style={'display': 'inline-block', 'margin': '10px 10px 10px 10px'})
            ],
            id="time_interval_div",
            style={
                "position": "fixed",
                "bottom": "0",                      # Stick it to the bottom
                "width": "100%",                    # Full width of the container
                "background-color": "lightgray",    # Just for visualization
                "font-size": "1.2em",               # Bigger font
            },
        ),
    ])


def visualisation_modal():
    return dbc.Row([
            dbc.Col(
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle("Header"), close_button=True),
                        dbc.ModalBody([
                            dbc.Container([
                                dbc.Row([
                                    dbc.Col([
                                        html.Label("Name"),
                                        dcc.Input(id="marker-name", placeholder="Name")
                                    ])
                                ]),
                                dbc.Row([
                                    dbc.Col([
                                        html.Label("Latitude"),
                                        dcc.Input(id="marker-lat", placeholder="Latitude")
                                    ]),
                                    dbc.Col([
                                        html.Label("Longitude"),
                                        dcc.Input(id="marker-lon", placeholder="Longitude")
                                    ])
                                ]),
                                dbc.Row([
                                    dbc.Col([
                                        html.Label("Type"),
                                        dcc.Input(id="marker-type", placeholder="Type")
                                    ]),
                                    dbc.Col([
                                        html.Label("Capacity"),
                                        dcc.Input(id="marker-capacity", placeholder="Capacity")
                                    ])
                                ]),
                                dbc.Row([
                                    dbc.Col([
                                        html.Label("Color"),
                                        dcc.Input(id="marker-color", placeholder="Color"),
                                        html.Div(id="marker_id_hidden", style={"display": "none"}),
                                    ])
                                ])
                            ])
                        ]),
                        dbc.ModalFooter([
                            dbc.Button("Save Changes", id="save_changes", color="primary"),
                            dbc.Button("Close", id="modal-close", color="secondary")
                        ])
                    ],
                    id="modal-centered",
                    centered=True,
                    is_open=False,
                ), width=6
                ), 
            ])


def getMarkerIndex(m) -> int:
    ''' m - json structure '''
    # return m['props']    ['children'][0]['props']['children'][0]['props']['children'][0]['props']['id']['index']
    # #     | Dash.Marker |        0. Popup       |      Intensity Div    |       dbc.Button      |
    try:
        result = m['props']    ['children'][0]['props']['children'][0]['props']['id']['index']
        #       | Dash.Marker |        0. Popup       |      Intensity Div    |      id      |
    except KeyError as e:
        result = None
        app.logger.warning(f"visualisation.getMarkerIndex(): {e}")

    return result



def getMarkerByIdentifier(m_identifier, inner_markers):
    ''' Returns index in marker array of item with attrbitu 'indetifier' equal to param m_identifier.'''
    for i,marker in enumerate(inner_markers):
        if marker[MarkerProperty.IDENTIFIER.value] == m_identifier:
            return i
    return None


def getMarkersCoordsByName(dash_markers, inner_markers, rm_name):
    ''' Returns marker coordinates lat, lon based on it's unique name. '''
    for m in dash_markers:
        marker = inner_markers[ getMarkerIndex(m) ]
        if rm_name == marker["name"]:
            return marker['lat'], marker['lon']

    return None, None


def getDataJammingAtLocation(jamming_data, device_sensitivity, power_limits, device_probability):
    ''' Prepares data structure for Graph visualising jamming signal at location.
        There are two graphs:
            - first is simple interference line of given device
            - second is jamming signal itself obtained from SW Fusion
    power_limits - tuple [min,max,step]
    '''
    if power_limits[0] != None and  power_limits[1] != None and  power_limits[2] != None:
        power_min, power_max, power_step = power_limits[0], power_limits[1], power_limits[2]

        original_step = (power_max - power_min) / power_step

        power_x = np.linspace(power_min+original_step/2, power_max-original_step/2, len(jamming_data))
        power_y = jamming_data

        sensitivity_x = [device_sensitivity,device_sensitivity]
        sensitivity_y = [0, np.array(jamming_data).max()]

        probab_x    = [None]
        probab_y    = [None]
        probab_name = f'P = {device_probability:.3f}'

        return power_x, power_y, sensitivity_x, sensitivity_y, probab_x, probab_y, probab_name

    else:
        return '-', '-', '-', '-', '-', '-', 'P = ...'




def getInterferences(rms_table: dict, device_name: str) -> list:
    ''' Returns list of probabilities of device in RMs table, which is Active. '''
    result = []
    for r in rms_table:
        if device_name == r[RmsProperty.ID.value.lower()] and r[RmsProperty.ACTIVE.value.lower()]:
            result.append(r[RmsProperty.PROBABILITY.value.lower()])
    return result


def getDeviceProbability(devices_data: dict, device_name: str) -> float:
    ''' Returns probability value of device by given name. '''
    for d in devices_data:
        if device_name == d[MarkerProperty.NAME.value]:
            if d[MarkerProperty.PROBABILITY.value] != None:
                return float(d[MarkerProperty.PROBABILITY.value])
            else:
                return None
    return None


def getDeviceSensitivity(devices_data: dict, device_name:str) -> float:
    ''' Returns sensitivity value of device by given name. '''
    for d in devices_data:
        if device_name == d[MarkerProperty.NAME.value]:
            return float(d[MarkerProperty.SENSITIVITY.value])
    return None


@app.callback(
    Output('graph_2d_probab_jam_signal',  'figure'),
    Input('dropdown_jamming_at_location', 'value'),
    Input('dropdown_jamming_at_location', 'options'),
    State('stored_jamming_at_location',   'data'),
    State('devices_table_hidden',         'data'),
    State('rms_table',                    'data'),
    State({'type': 'hidden_config_mins', 'index': ALL}, 'data'),
    State({'type': 'hidden_config_maxs', 'index': ALL}, 'data'),
    State({'type': 'hidden_config_step', 'index': ALL}, 'data') )
def updateJammingAtLocationGraph(selected_device, opts, jamming_data, devices_data, rms_table, config_mins, config_maxs, config_steps):
    ''' Renders jamming signal density for given device at given location.
        If no data available (for some unnown reason), graph will be empty with
        info about this state. '''
    device_sensitivity = getDeviceSensitivity(devices_data, selected_device)
    device_probability = getDeviceProbability(devices_data, selected_device)
    power_limits = [ config_mins[LimitsEnum.POWER.value], config_maxs[LimitsEnum.POWER.value], config_steps[LimitsEnum.POWER.value] ]
    patched_data = Patch()

    try:
        index = opts.index(selected_device)
        if jamming_data != None:
            power_x, power_y, sensitivity_x, senstivity_y, probab_x, probab_y, probab_name = getDataJammingAtLocation( jamming_data[ConfigEnum.DENSITY.value][index], device_sensitivity, power_limits, device_probability)
            patched_data['data'][0]['x'], patched_data['data'][0]['y'] = power_x, power_y
            patched_data['data'][1]['x'], patched_data['data'][1]['y'] = sensitivity_x, senstivity_y
            patched_data['data'][2]['x'], patched_data['data'][2]['y'] = probab_x, probab_y
            patched_data['data'][2]['name'] = probab_name
    except (ValueError, IndexError):
        # ValueError - no index found, thus device not in list and active - no need to render
        # IndexError - jamming_data has no data at given index (should not happen as dat amissing
        #              from SW Fusion - maybe request was wrong/not enough requested devices)
        patched_data['data'][0]['x'], patched_data['data'][0]['y'] = [None], [None]
        patched_data['data'][1]['x'], patched_data['data'][1]['y'] = [None], [None]
        patched_data['data'][2]['x'], patched_data['data'][2]['y'] = [None], [None]
        patched_data['data'][2]['name'] = ''

    return patched_data


@app.callback(
    Output({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "min",      allow_duplicate=True),
    Output({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "max",      allow_duplicate=True),
    Output({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "value",    allow_duplicate=True),
    Output({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "marks",    allow_duplicate=True),
    Output({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "step",     allow_duplicate=True),
    Output({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "pushable", allow_duplicate=True),
    Output('graph_2d_probab_jam_power', 'figure'),
    Input('dropdown_2d_probability_density', 'value'),
    Input('dropdown_2d_probability_density', 'options'),
    State('stored_2d_probability_density',   'data'),
    State({'type': 'hidden_config_mins', 'index': ALL}, 'data'),
    State({'type': 'hidden_config_maxs', 'index': ALL}, 'data'),
    State({'type': 'hidden_config_step', 'index': ALL}, 'data'),
    State({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "min"),
    State({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "max"),
    State({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "value"),
    State({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "marks"), 
    State({"type": "integrate_view_slider", "index": LimitsEnum.GRAPH_PROBAB_DENSITY.value}, "step") )
def update_graph_probability_density(selected_limit, opts, probability_data, config_mins, config_maxs, config_steps, slider_min, slider_max, slider_values, slider_marks, slider_step):
    ''' Updates slider for 2d probability density graph with correct min max values.'''
    patched_data = no_update

    # We need to propagate the value (and integrate toggle!) to the final config for SW FUsion... Hm, how
    limit = (LimitsEnum.get_enum_by_string(selected_limit))

    if probability_data != None and config_mins[limit.value] != None and config_maxs[limit.value] != None:

        original_step = (config_maxs[limit.value] - config_mins[limit.value]) / config_steps[limit.value]

        # Change x-axis label on 2d graph as well
        patched_data = Patch()
        patched_data['layout'] = get_probab_jam_power_layout(limit, [ config_mins[limit.value],config_maxs[limit.value] ])
        patched_data['data'][0]['x'] = np.linspace(config_mins[limit.value] + original_step/2, config_maxs[limit.value] - original_step/2, config_steps[limit.value])
        patched_data['data'][0]['y'] = probability_data

        # Limits between which the probability is calculated
        y_max = np.array(probability_data).max()
        patched_data['data'][1]['x'] = [slider_values[0],slider_values[0]]
        patched_data['data'][1]['y'] = [0,y_max]
        patched_data['data'][2]['x'] = [slider_values[1],slider_values[1]]
        patched_data['data'][2]['y'] = [0,y_max]

        # Calculated Probability value between limits
        calculated_probablity = calculateProbabilityBasedOnSensitivity(probability_data, config_mins[limit.value], config_maxs[limit.value], slider_values[0], slider_values[1], debugPrint=False)
        patched_data['data'][3]['name'] = f'P = {calculated_probablity:.3f}'

        # Update values only if min/max changed - otherwise it will be impossible to use the slider
        if slider_min != config_mins[limit.value] and slider_max != config_maxs[limit.value]:
            slider_values = [ config_mins[limit.value], config_maxs[limit.value] ]
        else:
            slider_values = no_update

        # Update marks and steps in slider
        mid_lim      = (config_mins[limit.value] + config_maxs[limit.value]) / 2.0
        slider_step  = (config_maxs[limit.value] - config_mins[limit.value]) / config_steps[limit.value]
        slider_marks = createSlider_marks(config_mins[limit.value], config_maxs[limit.value], slider_step, config_steps[limit.value], limit.can_be_int)
        

    return config_mins[limit.value], config_maxs[limit.value], slider_values, slider_marks, slider_step, slider_step/1.5, patched_data



''' Play / Pause Feature, so the data on screen are not updating.'''
@app.callback(
    Output("play_pause_sim", "children"),
    Input("play_pause_sim", "n_clicks") )
def pauseSimulation(clicks):
    global shouldRender
    shouldRender = not shouldRender
    pause = [ getPlayPauseHtml(isPlay=False) ]
    play  = [ getPlayPauseHtml(isPlay=True)  ]
    return pause if shouldRender else play



@app.callback(
    Output('map',                      'center'),
    Output('visual_rm_table_original', 'data', allow_duplicate=True),
    Input('visual_rm_table',    'active_cell'),
    State('visual_rm_table',    'data'),
    State('dash_markers_group', 'children'),
    State('devices_table',      'data') )
def highlightMarkerOnMapBasedOnRow(active_cell, data, dash_markers, inner_markers):
    ''' Need to store content of DataTable to the dcc.Store as well, so when row is deleted from DataTable
        we can find out what row was deleted and update marker on map accordingly. '''
    new_coords = no_update

    if active_cell:
        row = data[active_cell['row']]

        center_lat, center_lon = getMarkersCoordsByName(dash_markers, inner_markers, row['Name'])

        if center_lat is not None and center_lon is not None:
            new_coords = [center_lat, center_lon]
        else:
            app.logger.warn(f'There is no marker with name: {row["Name"]}')
    
    return new_coords, data



def find_matching_row(row, rms_table) -> int:
    ''' Find index of RM row in the RMs Table. '''
    for i, r in enumerate(rms_table):
        if row['Name'] == r[RmsProperty.ID.value.lower()] and \
            row['RMs'] == r[RmsProperty.RM.value.lower()]:
            return i
    return None


# -----------------------------------------------------------------------------
# If Row is deleted in RMs Info Table, meaning User is aware of specific RM for given Device.
#   Flag in MarkerProperty.RMS_CONFIRMED can be raised for given RM. If all RM's are confirmed,
#   color can be changed to Orange.
@app.callback(
    Output('visual_rm_table_original', 'data',     allow_duplicate=True),
    Output('dash_markers_group',       'children', allow_duplicate=True),
    Output('devices_table',            'data',     allow_duplicate=True),
    Output('devices_table_hidden',     'data',     allow_duplicate=True),
    # Output('rms_table',                'data',     allow_duplicate=True),
    Input('visual_rm_table',           'derived_virtual_data'),
    State('visual_rm_table_original',  'data'),
    State('dash_markers_group',        'children'),
    State('devices_table_hidden',      'data'),
    State('rms_table',                 'data'))
def update_table_after_delete(rms_info_table, original_data, dash_markers, devices_data, rms_table):

    if rms_info_table is None:
        raise PreventUpdate

    deleted_rows = [row for row in original_data if row not in rms_info_table]

    if deleted_rows:
        for row in deleted_rows:
            for i, m in enumerate(dash_markers):
                m_index = getMarkerIndex(m)

                try:
                    if devices_data[m_index][MarkerProperty.NAME.value.lower()] == row['Name']:
                        
                        # Flag the RMS Confirmed                            
                        # RMS_CONFIRMED and TIME_SINCE_JAM has to be JSON.loads & JSON.dumps processed to be able to work with them
                        json_rms_confirm = json.loads(devices_data[m_index][MarkerProperty.RMS_CONFIRMED.value])
                        json_rms_confirm[ row['RM_ID'] ] = True

                        # RMS_Cofirmed == False represents RM is active but not confirmed by user - cannot be switched to Orange
                        # Thus Device can be switched to Orange only if none of the RMS_Cofirmed is 'False'
                        if False in json_rms_confirm.values():
                            pass
                        else:
                            new_color = MarkerColor.ORANGE
                            icon_type = MarkerType.from_string(devices_data[m_index][MarkerProperty.TYP.value.lower()])
                            new_icon  = MarkerIcon(icon_type, new_color).getProperties()
                            
                            # Update inner structure and in dash markers
                            devices_data[m_index][MarkerProperty.COLOR.value.lower()] = new_color.value
                            m['props']['icon'] = new_icon

                        # Store the new RMS_CONFIRMED values into DataTable cell as stringified json
                        devices_data[m_index][MarkerProperty.RMS_CONFIRMED.value] = json.dumps(json_rms_confirm)

                except IndexError as e:
                    print(f"{__name__}::{update_table_after_delete.__name__} - {e}")

        return original_data, dash_markers, devices_data, devices_data#, rms_table

    raise PreventUpdate


def updateIcon(dash_m, inner_m, newColor, icon_type, time_tick=None, rms_id=None):
    ''' Updates Dash Leaflet Marker Icon with new properties. '''
    newIcon = MarkerIcon(icon_type, newColor).getProperties()
    inner_m[MarkerProperty.COLOR.value] = newColor.value  # Update in inner structure
    
    # Update in dash markers structure
    dash_m['props']['icon'] = newIcon

    return dash_m, inner_m



def loadAppConfigurationJson():
    ''' Loads an app config if any config.json available in root folder. '''
    if os.path.exists(APP_CONFIG_FILE):
        with open(APP_CONFIG_FILE, 'r') as json_file:
            return json.load(json_file)
    else:
        app.logger.warn(f'No app configuration file \'{APP_CONFIG_FILE}\' exists, initilizing new session.')
        return None


def saveAppConfigurationJson(sensors_data, devices_data, rms_data, slider_values, limit_mins, limit_maxs, units, processings, config_mins, config_maxs, config_steps, redis, limits_integrate_map_sliders, spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle):
    ''' Saves all values present in the app, to be used in next session. '''
    includeNonActive = True
    convertUnits     = False
    full_config = getFullConfiguration(includeNonActive, convertUnits, processings, config_mins, config_maxs, config_steps, sensors_data, devices_data, limits_integrate_map_sliders, spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle, None, True)
    devices     = devices_data
    rms         = rms_data

    # Update outgoing Devices, unnecessary parameters created during runtime are not stored in JSON
    for d in devices_data:
        d[MarkerProperty.TIME_JAM_SINCE.value] = None
        d[MarkerProperty.RMS_CONFIRMED .value] = None

    full_config.update({ConfigEnum.REDIS.value: getRedisConfig(redis)})

    limits_main_vis = {
        'app_data': {
            "slider_values": slider_values,
            "limit_mins":    limit_mins,
            "limit_maxs":    limit_maxs,
            "units":         units,
        },
        'config':  full_config,
        'devices': devices,
        'rms':     rms
        # sensors are already present in 'config'
    }
    with open(APP_CONFIG_FILE, 'w') as json_file:
        json.dump(limits_main_vis, json_file)
    pass


def getDefaultLimitsForMap(limits_integrate_map_sliders, config_mins, config_maxs, config_steps, convertUnits):
    ''' Returns default Integration Limits for Map Visualisation - only Density, as probability
        is not used in Map Visualisation. '''
    mins = [limit[0] for limit in limits_integrate_map_sliders]     # limits_integrate_map_sliders is list of all
    maxs = [limit[1] for limit in limits_integrate_map_sliders]     #   sliders, each slider has tuple of values (min,max)
    p_mins, p_maxs = None, None
    toggles = [False,False,True,True]
    density, probability = getIntegrationLimits(mins, maxs, p_mins, p_maxs, toggles, config_mins, config_maxs, config_steps, convertUnits)
    return density

def getDefaultLimitsForPowerProbabilityDensity(limits_integrate_map_sliders, dropdown_2d_probability_density_value, config_mins, config_maxs, config_steps, convertUnits):
    ''' Returns default Integration Limits for Power Probability Density.
        NOTE - this has always one and only integration toggle set on False! '''
    if dropdown_2d_probability_density_value == None:
        return {}

    limit = LimitsEnum.get_enum_by_string(dropdown_2d_probability_density_value)
    infinity = 1e30
    
    # Old before 2024-02-15, as this was not specified nowhere in documentation/request
    # mins = [-1*infinity, -1*infinity, -1*infinity, -1*infinity]
    # maxs = [   infinity,    infinity,    infinity,    infinity]

    # New after 2024-02-15, as this was new request
    mins = [limit[0] for limit in limits_integrate_map_sliders]     # limits_integrate_map_sliders is list of all
    maxs = [limit[1] for limit in limits_integrate_map_sliders]     #   sliders, each slider has tuple of values (min,max)
    
    # NOTE - For this Probability Density, sliders are used to calculate Probability, not for integration limits!!!
    # Actual value for sliders is stored in list with index GRAPH_PROBAB_DENSITY
    # mins[limit.value] = limits_integrate_map_sliders[LimitsEnum.GRAPH_PROBAB_DENSITY.value][0]
    # maxs[limit.value] = limits_integrate_map_sliders[LimitsEnum.GRAPH_PROBAB_DENSITY.value][1]
    
    p_mins, p_maxs = None, None

    toggles = [True,True,True,True]
    toggles[limit.value] = False
    
    density, probability = getIntegrationLimits(mins, maxs, p_mins, p_maxs, toggles, config_mins, config_maxs, config_steps, convertUnits)
    
    return density    


def isProbabilityActive(mins, maxs, p_mins, p_maxs, toggles, config_mins, config_maxs, config_steps, shouldConvertUnits):
    ''' Checks if Probability part of Density is valid and active or not. '''
    _, probability = getIntegrationLimits(mins, maxs, p_mins, p_maxs, toggles, config_mins, config_maxs, config_steps, shouldConvertUnits)
    return len(probability) > 0


def getLimitsForSpecificVis(spec_view_active_toggle, mins, maxs, p_mins, p_maxs, toggles, config_mins, config_maxs, config_steps, convertUnits):
    ''' Gets configuration for Specific Visualisation, only if active toggle is On '''
    if spec_view_active_toggle:
        return getIntegrationLimits(mins, maxs, p_mins, p_maxs, toggles, config_mins, config_maxs, config_steps, convertUnits)
    else:
        return {}, {}


def isDeviceInArea(config_mins, config_maxs, device):
    ''' Determine if given device with coords is inside the lat lon area specified by configuration limits. '''
    rectangle = [getPolygonCoords(config_mins, config_maxs)]
    point     = [device[MarkerProperty.LON.value], device[MarkerProperty.LAT.value]]

    return isPointInsideTile(rectangle, point)


def getJammingSignalDevices(devices_data: dict, includeNonActive: bool, config_mins: dict, config_maxs:dict, convertUnits: bool) -> list:
    ''' Return list of 'devices' to pass to the SW Fusion. 
        WARNING: devices whic are NOT IN the latitude-longitude are are discarded and not passed to the Fusion
        (SW Fusion cannot handle this kind of devices and raises an error). '''
    def deviceJam(lat, lon, hgt):
        result = {
            ConfigStrings.LAT.value: convertValueDeg2Rad(convertUnits, lat),
            ConfigStrings.LON.value: convertValueDeg2Rad(convertUnits, lon),
            ConfigStrings.HGT.value: 433 # Average Cech elevation (for precise HGT at given location - Terrain profile needed)
        }
        if hgt is not None:
            result[ConfigStrings.HGT.value] = hgt
        return result

    result = []
    for d in devices_data:
        if includeNonActive or d[MarkerProperty.ENABLED_FUSION.value]:

            # rectangle = [getPolygonCoords(config_mins, config_maxs)]
            # point     = [d[MarkerProperty.LON.value],d[MarkerProperty.LAT.value]]

            # if isPointInsideTile(rectangle, point):
            if isDeviceInArea(config_mins, config_maxs, d):
                result.append( deviceJam(d[MarkerProperty.LAT.value],d[MarkerProperty.LON.value],d[MarkerProperty.HGT.value]) )
            else:
                # Print warning to terminal only if this condition - opposite is for saving app config, which is ok.
                if not includeNonActive:
                    app.logger.warning(f"Warning - Device '{d[MarkerProperty.NAME.value]}' is not inside the specified area. Therefore not passed to SW Fusion.")
    return result


def getFullConfiguration(includeNonActive, convertUnits, processings, config_mins, config_maxs, config_steps, sensors_data, devices_data, limits_integrate_map_sliders, spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle, dropdown_2d_probability_density_value, spec_view_active_toggle):
    ''' Obtains full configuration JSON structure needed for SW Fusion - that means WITHOUT non-active '''

    # Limits for Specific Visualisation
    spec_vis_density, spec_vis_probab = getLimitsForSpecificVis(spec_view_active_toggle, spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle, config_mins, config_maxs, config_steps, convertUnits=convertUnits)
    
    probabilities = []
    densities = [getDefaultLimitsForMap (limits_integrate_map_sliders, config_mins, config_maxs, config_steps, convertUnits),
                 getDefaultLimitsForPowerProbabilityDensity(limits_integrate_map_sliders, dropdown_2d_probability_density_value, config_mins, config_maxs, config_steps, convertUnits) ]    

    # In case Specific Visualisation has no Probability - Density request goes to Integration-Jammer-Density
    #   However, if Specific Visualisation HAS Probability, the whohle Density & Probabbility reqeust goes to
    #   Integration-Jammer-Probability where Density & Probability are specified.
    if len(spec_vis_probab) > 0:
        probabilities = [{  ConfigEnum.DENSITY    .value: spec_vis_density,
                            ConfigEnum.PROBABILITY.value: spec_vis_probab    }]
    else:
        # Append spec vis density only if any (non-active spec vis might return empty {} dictionary)
        if len(spec_vis_density) > 0:
            densities.append(spec_vis_density)

    return {
        ConfigEnum.PROCESSING .value: getProcessingConfig(processings),
        ConfigEnum.LIMITS     .value: getLimitsConfig(config_mins, config_maxs, config_steps, convertUnits),
        ConfigEnum.SENSORS    .value: getSensorsConfig(sensors_data, includeNonActive),
        ConfigEnum.INTEGRATION.value: {
            ConfigEnum.JAMMER.value: { 
                ConfigEnum.DENSITY    .value: densities,
                ConfigEnum.PROBABILITY.value: probabilities
            },
            ConfigEnum.JAMMING_SIGNAL.value: {
                ConfigEnum.DENSITY.value:     getJammingSignalDevices(devices_data, includeNonActive, config_mins, config_maxs, convertUnits),
                ConfigEnum.PROBABILITY.value: []
            }
        }
    }



@app.callback(
    Output("time_interval", "children", allow_duplicate=True),
    Input("interal_save_session", "n_intervals"),

    # Sensor, Devices and RMs Data
    State('sensors_table_hidden', 'data'),
    State('devices_table_hidden', 'data'),
    State('rms_table',            'data'),

    # For saving app config file for next session
    State({"type": "integrate_view_slider", "index": ALL}, 'value'),
    State({"type": "integrate_view_slider", "index": ALL}, 'min'),
    State({"type": "integrate_view_slider", "index": ALL}, 'max'),
    State({"type": "integrate_units",       "index": ALL}, 'value'),

    # Configuration values
    State({'type': 'hidden_processings', 'index': ALL}, 'data'),
    State({'type': 'hidden_config_mins', 'index': ALL}, 'data'),
    State({'type': 'hidden_config_maxs', 'index': ALL}, 'data'),
    State({'type': 'hidden_config_step', 'index': ALL}, 'data'),
    State({'type': 'hidden_redis',       'index': ALL}, 'data'),
    
    # Specific Visualisation values
    State({'type': 'integrate_view_slider',     'index': ALL}, 'value'),
    State({'type': 'spec_vis_slider',           'index': ALL}, 'value'),
    State({'type': 'spec_vis_probab_mins',      'index': ALL}, 'value'),
    State({'type': 'spec_vis_probab_maxs',      'index': ALL}, 'value'),
    State({'type': 'spec_vis_integrate_toggle', 'index': ALL}, 'value') )
def save_session_tick(n, sensors_data, devices_data, rms_data, slider_values, limit_mins, limit_maxs, units, processings, config_mins, config_maxs, config_steps, redis, limits_integrate_map_sliders, spec_vis_slider, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle):
    app.logger.info(f"Internal State saved to file: \'{APP_CONFIG_FILE}\'")

    # Get only mins/maxs from Sliders as separate variables
    spec_vis_mins = [subarray[0] for subarray in spec_vis_slider]
    spec_vis_maxs = [subarray[1] for subarray in spec_vis_slider]

    saveAppConfigurationJson(sensors_data, devices_data, rms_data, slider_values, limit_mins, limit_maxs, units, processings, config_mins, config_maxs, config_steps, redis, limits_integrate_map_sliders, spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle)

    # return no_update as we do not need any Dash update and Dash itself does not allow callback without Output
    return no_update


def update_device_color(m, data, icon_type, new_color):
    ''' Updates color of Dash Marker, Device Marker and if necessary, in RMS Table. '''
    dash_marker, device_marker = updateIcon(m, data, new_color, icon_type)#, time_tick, rms_id)
    return dash_marker, device_marker, new_color.value


def formatTime(seconds: float) -> str:
    ''' Returns time text format with secnds as input. '''
    return str(datetime.timedelta(seconds=seconds)).split(".")[0] if seconds != None else "-"


def calculateProbabilityBasedOnSensitivity(jam_data, power_min, power_max, device_sensitivity, probab_max, debugPrint=False):
        ''' Calculates integral for dataset above device_sensitivity.'''

        x_values  = np.linspace(power_min, power_max, len(jam_data))

        start_idx_where = np.where(x_values >= device_sensitivity)
        end_idx_where   = np.where(x_values <= probab_max)

        if start_idx_where[0].size == 0 or end_idx_where[0].size == 0:
            return 0

        start_idx = start_idx_where[0][0]
        end_idx   =   end_idx_where[0][-1]

        subset_jam_data = jam_data[start_idx:end_idx+1]

        if debugPrint:
            print(f"x_values:        {x_values}")
            print(f"start_idx:       {start_idx}")
            print(f"end_idx:         {end_idx}")
            print(f"jam_data:        {jam_data}")
            print(f"subset_jam_data: {subset_jam_data}")

        return np.sum(subset_jam_data)
        # Note - using np.trapz look ok, but since we have probablity density already, simple sum is sufficient


def sortLimits(limits):
    ''' Shape of incoming data is based ordered limits list - probably, but empirically tested
        1) Power 2) Height 3) Latitude 4) Longitude
        Reshaping has to be done with steps from these. So if request is Latitude & Power, we must not use
        np.reshape( lat.steps, power.steps ), but according to the list, so mp.reshape( power.steps, lat.steps ),
        otherwise the resulted graph will not make any sense.
    '''
    return sorted(limits, key=lambda x: x.priority, reverse=True)



def process_specific_vis(jam_specific_data, config_mins, config_maxs, config_steps, spec_vis_integrate_toggle, spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, full_config):
    ''' Process Specific Visualisation - based on Toggles setup, active graph Tab is selected and all patched data
        are updated with correct datasets. '''
    patched_1d_figure = None        # As 1d is just number/text, no need for Figure Patch()
    patched_2d_figure = Patch()
    patched_3d_figure = Patch()
    patched_4d_figure = Patch()

    active_tab = 'tab_2d'
    count      = 0
    dimensions = []
    limits     = []     # Limits for correct data
    limits_dim = []     # Complementary Limits, used in getJammingIndexes() to update indexes of 3d graph
    dims_compl = []

    div_block = {'display': 'block'}
    show_divs = [{'display': 'none'} for _ in range(len(TabsEnum))]

    for el in LimitsEnum:
        try:
            # False Toggles are used in reshaping
            if not spec_vis_integrate_toggle[el.value]:
                count += 1
                dimensions.append( config_steps[el.value] )
                limits    .append( el )
            else:
                limits_dim.append( el )
                dims_compl.append( config_steps[el.value] )
        except IndexError:
            # LimitsEnum has more items than spec_vis_integrate_toggle
            pass

    # Limits has to be sorted by Limits Priority to get correct shaping
    limits = sortLimits(limits)

    probabilityActive = isProbabilityActive(spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle, config_mins, config_maxs, config_steps, False)

    # Set correct Spec Viz tab
    if count == 0 or probabilityActive:
        show_divs[TabsEnum.NUMBER.value]  = div_block
    
    elif count == 1:
        show_divs[TabsEnum.GRAPH.value]   = div_block
        patched_2d_figure['layout']['xaxis']['title']['text'] = f"{limits[0].string} [{limits[0].unit}]"
    
    elif count == 2:
        show_divs[TabsEnum.HEAT.value]    = div_block
        show_divs[TabsEnum.SPACE.value]   = div_block

        # Note - x axis has the 'less significant' values, thus limits[1] - limits array is sorted by priority
        patched_3d_figure['layout']['xaxis']['title']['text'] = f"{limits[1].string} [{limits[1].unit}]"
        patched_3d_figure['layout']['yaxis']['title']['text'] = f"{limits[0].string} [{limits[0].unit}]"
        patched_3d_figure['layout']['annotations'] = []
    
    elif count >= 3 and count <= 4:
        show_divs[TabsEnum.UNKNOWN.value] = div_block


    # In case no data available for Specific Visualisation, update only active Tab
    if jam_specific_data == None:
        return show_divs, no_update, no_update, no_update, no_update


    if count == 0 or probabilityActive:
        if len(jam_specific_data) > 1:
            patched_1d_figure = html.Span('Recalculating...')
        else:
            patched_1d_figure= html.Span(' '.join(str(e) for e in jam_specific_data))
        
    elif count == 1:
        try:
            patched_2d_figure['data'][0]['x'] = [x for x in np.linspace(config_mins[limits[0].value], config_maxs[limits[0].value], len(jam_specific_data)+1)]
            patched_2d_figure['data'][0]['y'] = jam_specific_data
        except (IndexError, ValueError) as e:
            patched_2d_figure['layout'] = getFigureLayouWithAnnotation("Recalculating...")


    elif count == 2:
        # Apply limits for correct reshaping - Density index is [2] for non-probability data.
        # For 3d Volume graph is not necesary to fill the data with zeros, as the graph will handle it by itself
        #   (Not the same as for map jamming overlay, which is using proprietary png overlay)
        dim = []

        for i in range(2):
            i_min, i_max, local_min, local_max = getJammingIndex(full_config, limits[i], 2)

            if i_min != None and i_max != None:
                dim.append(i_max - i_min + 1)

        try:
            graph_3d_data = np.array(jam_specific_data)
            graph_3d_data = graph_3d_data.reshape(dim[0], dim[1])

            # Note - x axis has the 'less significant' values, thus limits[1] - limits array is sorted by priority
            patched_3d_figure['data'][0]['x'] = np.array(split_interval(config_mins[limits[1].value], config_maxs[limits[1].value], config_steps[limits[1].value]))
            patched_3d_figure['data'][0]['y'] = np.array(split_interval(config_mins[limits[0].value], config_maxs[limits[0].value], config_steps[limits[0].value]))
            patched_3d_figure['data'][0]['z'] = graph_3d_data
            
            patched_3d_figure['data'][0]['origin'] = 'lower' if flipGpsImage[FlipEnum.VERTICAL.value] else 'upper'

            # In case units are the same for both of the axis, 'turn on' feature 'axis equal'
            if limits[0].unit == limits[1].unit:
                patched_3d_figure['layout']['xaxis']['scaleanchor'] = 'y';
            else:
                patched_3d_figure['layout']['xaxis']['scaleanchor'] = False;

        except (IndexError, ValueError) as e:
            app.logger.warning(f"Specific Visualisation Graphs - Error: {e}")
            patched_3d_figure['layout'] = getFigureLayouWithAnnotation("Recalculating...")

        # Note - cannot use functions as overwriting the whole 'layout' will change the camera angle - and reading it is way to difficult
        # patched_4d_figure['layout']['annotations'] = []

        try:
            # Note - x axis has the 'less significant' values, thus limits[1] - limits array is sorted by priority
            patched_4d_figure['data'][0]['x'] = np.array(split_interval(config_mins[limits[1].value], config_maxs[limits[1].value], config_steps[limits[1].value]))
            patched_4d_figure['data'][0]['y'] = np.array(split_interval(config_mins[limits[0].value], config_maxs[limits[0].value], config_steps[limits[0].value]))
            patched_4d_figure['data'][0]['z'] = graph_3d_data

            patched_4d_figure['layout']['scene']['xaxis']['title']['text'] = f"{limits[1].string} [{limits[1].unit}]"
            patched_4d_figure['layout']['scene']['yaxis']['title']['text'] = f"{limits[0].string} [{limits[0].unit}]"
            patched_4d_figure['layout']['scene']['zaxis']['title']['text'] = "pmf [1]"

        except ValueError as e:
            app.logger.warning(f"Specific Visualisation Graphs 4D - Error: {e}")
            patched_4d_figure['layout'] = getFigureLayouWithAnnotation("Recalculating...")

    elif count >= 3 and count <= 4:
        pass

    return show_divs, patched_1d_figure, patched_2d_figure, patched_3d_figure, patched_4d_figure


def getActualTimeInSeconds():
    ''' Returns actual time in seconds since midnight. '''
    current_time_seconds = time.time()
    current_time_struct  = time.localtime(current_time_seconds)
    return (current_time_struct.tm_hour * 3600) + (current_time_struct.tm_min * 60) + current_time_struct.tm_sec





def updateColorbarScale(data, store_colorbar_min_max):
    ''' Updates Map Colorbar Scale if necessary. Based on stored min/max in dcc.Store and current value. '''
    try:
        map_data_min = np.array(data).min()
        map_data_max = np.array(data).max()
    except ValueError as e:
        # In case data has no valaid data (empty for example), return no_update
        app.logger.warning(f"updateColorbarScale: {e}")
        return no_update, store_colorbar_min_max

    colorscale = no_update
    update_colorbar = False
    if store_colorbar_min_max[0] == None or store_colorbar_min_max[0] > map_data_min:
        store_colorbar_min_max[0] = map_data_min
        update_colorbar = True
    
    # if store_colorbar_min_max[1] == None or store_colorbar_min_max[1] < map_data_max:
    
    # Update max everytime
    store_colorbar_min_max[1] = map_data_max
    update_colorbar = True
    
    if update_colorbar:
        colorscale = createColorbarScale(store_colorbar_min_max[0], store_colorbar_min_max[1])

    return colorscale, store_colorbar_min_max



def clear_redis_data(data):
    ''' Clears data from redis, which are in json format. Keys 'jammer', 'jammer_density' and nested ones
        has to be cleared. '''
    def short_for(data):
        for i,jd in enumerate(data):
            data[i] = [0.0, 0.0, 0.0]
        return data

    data[ConfigEnum.JAMMER.value]        [ConfigEnum.DENSITY.value]     = short_for(data[ConfigEnum.JAMMER.value]        [ConfigEnum.DENSITY.value])
    data[ConfigEnum.JAMMER.value]        [ConfigEnum.PROBABILITY.value] = short_for(data[ConfigEnum.JAMMER.value]        [ConfigEnum.PROBABILITY.value])
    data[ConfigEnum.JAMMING_SIGNAL.value][ConfigEnum.DENSITY.value]     = short_for(data[ConfigEnum.JAMMING_SIGNAL.value][ConfigEnum.DENSITY.value])
    data[ConfigEnum.JAMMING_SIGNAL.value][ConfigEnum.PROBABILITY.value] = short_for(data[ConfigEnum.JAMMING_SIGNAL.value][ConfigEnum.PROBABILITY.value])

    return data



@app.callback(
    Output({'type': 'integrate_view_value_min', 'index': MATCH}, 'children'),
    Output({'type': 'integrate_view_value_max', 'index': MATCH}, 'children'),
    Input ({'type': 'integrate_view_slider',    'index': MATCH}, 'value'))
def update_sliders_min_max_label_value(values):
    ''' Updates label values in front of / next to the sliders below the map. '''
    # Note: sliders have min/max values, thus @param values is 2-item list
    return createRow_SliderValue(values[0]), createRow_SliderValue(values[1])



# Due to circular imports, defined here
from views.devices import initialiseDevicesToColor

@app.callback(
    [
        Output('time_interval',                            'children'),                         # Display time
        Output('dash_markers_group',                       'children', allow_duplicate=True),                         # Markers in Dash Map
        Output('devices_table_hidden',                     'data'),                             # Markers Original form
        Output('visual_rm_table',                          'data'),                             # RMs Info Table
        Output('visual_rm_table_original',                 'data',     allow_duplicate=True),   # RMs Table for keeping originals (used when deleting rows)
        Output('jamming_image',                            'url'),                              # URL to jamming ImageOverlay in Map
        Output('jamming_image',                            'bounds',   allow_duplicate=True),   # Image bounds for ImageOverlay, in case limits are changed
        Output('stored_tiles_w_device',                    'data',     allow_duplicate=True),   # List of jamming data tile indexes with device in it
        Output('stored_jamming_at_location',               'data'),                             # Data for Jamming at Location Graph
        Output('stored_2d_probability_density',            'data'),                             # Data for Probability Density Graph
        Output('spec_view_graph_1d',                       'children', allow_duplicate=True),   # Specific Visualisation, 1d graph
        Output('spec_view_graph_2d',                       'figure',   allow_duplicate=True),   # Specific Visualisation, 2d graph
        Output('spec_view_graph_3d',                       'figure',   allow_duplicate=True),   # Specific Visualisation, 3d graph
        Output('spec_view_graph_4d',                       'figure',   allow_duplicate=True),   # Specific Visualisation, 4d graph
        Output('dropdown_jamming_at_location',             'value',    allow_duplicate=True),   # Force update in dropdown to re-render 2d graph
        Output('dropdown_2d_probability_density',          'value',    allow_duplicate=True),   # Force update in dropdown to re-render 2d graph
        Output('redis_connected',                          'children'),                         # Info note about Redis connecteded or not
        Output({'type': 'marker_intensity', 'index': ALL}, 'children'),
        Output({'type': 'spec_view_tab', 'index': ALL},    'style'),
        Output('map_colorbar',                             'colorscale'),                       # Scale of Colorbar in map
        Output('store_colorbar_min_max',                   'data'),                             # Stored Colorbar min,max values
        Output('map_colorbar',                             'max'),                              # Max value in Colorbar map
    ],
    Input('interal_logic',                                     'n_intervals'),
    State('jamming_image',                                     'bounds'),
    State('dash_markers_group',                                'children'),
    State('devices_table_hidden',                              'data'),
    State('visual_rm_table',                                   'data'),
    State('rms_table_hidden',                                  'data'),             # RMs Table on RMs Tab
    State('stored_tiles_w_device',                             'data'),
    State('sensors_table_hidden',                              'data'),
    State('dropdown_jamming_at_location',                      'value'),
    State({'type': 'hidden_redis',              'index': ALL}, 'data'),
    State({'type': 'hidden_config_mins',        'index': ALL}, 'data'),
    State({'type': 'hidden_config_maxs',        'index': ALL}, 'data'),
    State({'type': 'hidden_config_step',        'index': ALL}, 'data'),
    State({'type': 'hidden_processings',        'index': ALL}, 'data'),
    State({'type': 'integrate_view_slider',     'index': ALL}, 'value'),
    State({'type': 'spec_vis_slider',           'index': ALL}, 'value'),
    State({'type': 'spec_vis_probab_mins',      'index': ALL}, 'value'),
    State({'type': 'spec_vis_probab_maxs',      'index': ALL}, 'value'),
    State({'type': 'spec_vis_integrate_toggle', 'index': ALL}, 'value'),
    State({'type': 'marker_intensity',          'index': ALL}, 'children'),     # Jamming Intensity at device
    State({'type': 'marker_intensity',          'index': ALL}, 'id'),           # location in Device Tooltip/Pop-up
    State('tabs_main',                                         'value'),
    State('dropdown_2d_probability_density',                   'value'),
    State('spec_view_active_toggle',                           'value'),        # Toggle if have spec vis active or not
    State('store_colorbar_min_max',                            'data'),         # Min/Max values stored in Map Colorscale/Colorbar
    blocking=True)  # for queue
def tickLogic(time_tick, incoming_bounds, dash_markers, devices_data, rm_info_table_data, rms_table, tiles_indexes, sensors_data, dropdown_jam, redis, config_mins, config_maxs, config_steps, processings, limits_integrate_map_sliders, spec_vis_slider, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle, marker_popup_intensity, marker_popup_id, main_tabs, dropdown_2d_probability_density_value, spec_view_active_toggle, store_colorbar_min_max):

    # Get only mins/maxs from Sliders as separate variables
    spec_vis_mins = [subarray[0] for subarray in spec_vis_slider]
    spec_vis_maxs = [subarray[1] for subarray in spec_vis_slider]

    # Method 'updateMarkersWithTileIndex()' takes more time than 'interal_logic' dcc.Interval itself
    #   - this will cause computed data are not Outputed via Dash Callback - new data are just thrown away.
    #   Thus, global variable has to be used to store the new data.
    global global_tiles_index, global_devices_data, global_fusion_time, global_jam_data_previous
    global queue_request, queue_data, flipGpsImage, shouldRender, jam_map_limits, redis_updated

    # Save global Dash Markers, as changing them and updating them from callback did not work
    if globals.global_dash_markers != None:
        dash_markers = globals.global_dash_markers
        globals.global_dash_markers = None

    # active_spec_vis_tab = [for _no_update,no_update,no_update,no_update,no_update] # Uff - active spec vis tab has three childrens - how to specifiy dynamically?
    active_spec_vis_tab = [no_update for _ in range(len(TabsEnum))]

    # In case Tab other than Main, Specific Visualisation, Sensors or Config is visible, prevent update
    # (otherwise it will be impossbile to change values in every DataTable)
    if main_tabs  != AppTabs.MAIN.value    and \
        main_tabs != AppTabs.SPEC.value    and \
        main_tabs != AppTabs.SENSORS.value and \
        main_tabs != AppTabs.CONFIG.value:
        return [no_update] + [dash_markers] + [no_update] * 15 + [marker_popup_intensity] + [active_spec_vis_tab] + [no_update] * 3
        # raise PreventUpdate
    
    no_up               = no_update
    picture_path        = no_update
    colorscale          = no_update
    intensities         = None
    thread_time         = None
    fusion_time         = None
    jamming_signal      = None
    redis_connected     = None

    
    if not shouldRender or config_steps[0] == None or config_steps[1] == None:
        raise PreventUpdate

    gnssLimits, geo_bounds = getGnssLimits(config_mins, config_maxs, config_steps)

    # In case new GNSS limits - change bounds, update tileIndexes and reset RMs - reset everything,
    #   as SW Fusion will reset as well.
    if jam_map_limits != gnssLimits:
        
        # Pause Render, as Dash Callback are non-blocking and 'updateMarkersWithTileIndex()' takes more than
        #   one tick to process - SW Fusion is restarted regardles, so ne need to render anything
        shouldRender = False
        geodata, geo_bounds = createGeodata(gnssLimits)
        devices_data, new_tiles_indexes = updateMarkersWithTileIndex(geodata, devices_data)

        # Set colors of Devices/Markers to Green and Clear RMs Info Table
        devices_data = initialiseDevicesToColor(devices_data, MarkerColor.GREEN)
        rm_info_table_data.clear()

        global_tiles_index  = new_tiles_indexes
        global_devices_data = devices_data

        shouldRender = True

        # Send values to bounds, as nothing else is actually needed in rendering - SW Fusion will restart anyway
        return  no_up,                  \
                no_up,                  \
                devices_data,           \
                rm_info_table_data,     \
                rm_info_table_data,     \
                no_up,                  \
                geo_bounds,             \
                new_tiles_indexes,      \
                no_up,                  \
                no_up,                  \
                no_up,                  \
                no_up,                  \
                no_up,                  \
                no_up,                  \
                no_up,                  \
                no_up,                  \
                no_up,                  \
                marker_popup_intensity, \
                active_spec_vis_tab,    \
                no_up,                  \
                no_up,                  \
                no_up

    # Save new tiles and devices_data, as obtaining them from callback did not work
    if global_tiles_index != None:
        tiles_indexes = global_tiles_index
        global_tiles_index = None

    if global_devices_data != None:
        devices_data = global_devices_data
        global_devices_data = None

    lat_dim, lon_dim = config_steps[LimitsEnum.LATITUDE.value], config_steps[LimitsEnum.LONGITUDE.value]
    full_config = getFullConfiguration(False, True, processings, config_mins, config_maxs, config_steps, sensors_data, devices_data, limits_integrate_map_sliders, spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle, dropdown_2d_probability_density_value, spec_view_active_toggle)

    # Send request
    request = {
        RequestEnums.INDEXES   .value: tiles_indexes,
        RequestEnums.FLIP      .value: flipGpsImage,
        RequestEnums.REDIS     .value: getRedisConfig(redis),
        RequestEnums.CONFIG    .value: full_config,
        RequestEnums.DIMENSIONS.value: [ lat_dim, lon_dim ]
    }
    queue_request.put(request)

    # Check for data from PNG Generator
    all_data            = None
    probability_density = None

    if not queue_data.empty():
        requests = []
        while not queue_data.empty():
            requests.append(queue_data.get())
        processed_data = requests[-1]

        picture_path    = processed_data[ConfigEnum.IMAGE.value]
        intensities     = processed_data[ConfigEnum.DATA.value]
        thread_time     = processed_data[ConfigEnum.TIME.value]
        redis_connected = processed_data[ConfigEnum.REDIS_ACTIVE.value]
        all_data        = processed_data[ConfigEnum.ALL_DATA.value]
        jamming_signal  = all_data[ConfigEnum.JAMMING_SIGNAL.value]
        redis_up_date   = processed_data[ConfigEnum.REDIS_UPDATED.value]

        fusion_time    = float( all_data["time"] )

    actual_time = getActualTimeInSeconds()

    # Timer / Recalculating Bar at the bottom
    if fusion_time != None:
        time_str = f'Fusion Time: {formatTime(fusion_time)} | Current Time: {formatTime(actual_time)}'

        # If Fusion time changed - reset Redis update (LAST step)
        if fusion_time != global_fusion_time:
            redis_updated = False

        # If REDIS updated recently, start printing info label
        if redis_up_date:
            redis_updated = True

        if redis_updated:
            redis_updated_recalculating = f" | Recalculating ..."
            time_str += redis_updated_recalculating

            # As requested, when 'recalculating' is active, clear all data from redis, so no graph has any data
            all_data = clear_redis_data(all_data)
            picture_path = os.path.join('assets', 'icons', 'empty_map_jam.png')
    else:
        time_str = no_update

    redis_connected_str = getConnectionOnOffHtml(redis_connected)
    

    # Update Marker's / Device's Property - Integer of Probability
    if jamming_signal != None:
        active_index = 0
        power_min, power_max = config_mins[LimitsEnum.POWER.value], config_maxs[LimitsEnum.POWER.value]

        if power_min != None and power_max != None:
            for d in devices_data:
                try:
                    if isDeviceInArea(config_mins, config_maxs, d) and d[MarkerProperty.ENABLED_FUSION.value]:
                        d[MarkerProperty.PROBABILITY.value] = calculateProbabilityBasedOnSensitivity(jamming_signal[ConfigEnum.DENSITY.value][active_index], float(power_min), float(power_max), float(d[MarkerProperty.SENSITIVITY.value]), float(power_max))
                        active_index += 1
                except IndexError as e:
                    print(f"visualisation.py::tick() {e} | Len(densitiy): {len(jamming_signal[ConfigEnum.DENSITY.value])} vs active_index: {active_index}")

    # No data from Fusion, pdate probability variable
    if all_data != None:
        probability_density = all_data[ConfigEnum.JAMMER.value][ConfigEnum.DENSITY.value][1]     # [0] is Map Jam Data, [1] Probability Density, [2] is Specific

    # Update processes and properties dependent on data - ONLY if fusion time changed, not different time
    patched_1d_figure      = no_update
    patched_2d_figure      = no_update
    patched_3d_figure      = no_update
    patched_4d_figure      = no_update
    colorscale             = no_update
    store_colorbar_maximum = no_update

    # Update Map Colorbar range if necessary
    if not redis_updated and all_data != None:
        colorscale, store_colorbar_min_max = updateColorbarScale(all_data[ConfigEnum.JAMMER.value][ConfigEnum.DENSITY.value][0], store_colorbar_min_max)
        store_colorbar_maximum = store_colorbar_min_max[1]
        
    # In case Specific Visualisation Probability request is active, process incoming data differently
    passSpecVis = False
    if all_data != None:
        if isProbabilityActive(spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, spec_vis_integrate_toggle, config_mins, config_maxs, config_steps, False):
            jam_specific_data = all_data[ConfigEnum.JAMMER.value][ConfigEnum.PROBABILITY.value]
        elif len(all_data[ConfigEnum.JAMMER.value][ConfigEnum.DENSITY.value]) >= 3:
            jam_specific_data = all_data[ConfigEnum.JAMMER.value][ConfigEnum.DENSITY.value][2]     # [0] is Map Jam Data, [1] Probability Density, [2] is Specific
        else:
            jam_specific_data = None
    else:
        jam_specific_data = None
    # else:
        # passSpecVis = True

    # Fill the patched_data
    if not passSpecVis:
        # Update only if new data are different than the old ones

        overwriteCondition = False
        if fusion_time != global_fusion_time:
            app.logger.info(f"Fusion Time changed {global_fusion_time} -> {fusion_time}. Updating dependent properties.")
            global_fusion_time = fusion_time
            overwriteCondition = True
        
        # if True:
        newData = global_jam_data_previous == None or \
                    len(global_jam_data_previous) != len(jam_specific_data) or \
                    not np.allclose(np.array(global_jam_data_previous), np.array(jam_specific_data), atol=1e-15)

        ### NOTE 2024 02 07 - this part was commented to be able to be updated immendiately based on Integrate toggles
        #       However not tested true performance with 'live' data, as updating every 'tick' might have effecct on performace.
        #       That is why this 'if' was here in the first place.

        # if overwriteCondition or newData:
        #     global_jam_data_previous = jam_specific_data

        #     # Update jamming densities in Specific Visualisation Graph(s)
        active_spec_vis_tab, patched_1d_figure, patched_2d_figure, patched_3d_figure, patched_4d_figure = process_specific_vis(jam_specific_data, config_mins, config_maxs, config_steps, spec_vis_integrate_toggle, spec_vis_mins, spec_vis_maxs, spec_vis_probab_mins, spec_vis_probab_maxs, full_config)
        #     app.logger.info("Specific visualisation - incoming data CHANGED")
        # else:
        #     app.logger.info("Specific visualisation - incoming data not changed")
    else:
        app.logger.info("Specific visualisation - nothing to do.")



    ### Iterate over RMs Info Table to remove rows, which does not exist in RMs Table (RM entry was deleted/deactivated)
    rm_info_rows_to_keep = []
    for rm_info_entry in rm_info_table_data:

        rm_entry_exist = False
        for rm_entry in rms_table:
            if rm_info_entry['RM_ID'] == rm_entry[RmsProperty.ID.value]:
                rm_entry_exist = True
                break

        if rm_entry_exist:
            rm_info_rows_to_keep.append(rm_info_entry)

    rm_info_table_data.clear()
    rm_info_table_data = rm_info_rows_to_keep



    ## For all Active Devices (displayed on map as Dash Markers) - in case we have data for the map (intensities)
    if intensities is not None:
        for i, m in enumerate(dash_markers):

            m_identifier = getMarkerIndex(m)                                    # get unique identifier 
            m_index      = getMarkerByIdentifier(m_identifier, devices_data)    # get index in array of given item with that identifier

            if m_index == None:
                continue

            tile_index         =       devices_data[m_index][MarkerProperty.TILE_INDEX.value]
            device_name        =       devices_data[m_index][MarkerProperty.NAME.value]
            device_sensitivity = float(devices_data[m_index][MarkerProperty.SENSITIVITY.value])
            device_probability =       devices_data[m_index][MarkerProperty.PROBABILITY.value]
            device_type        =       devices_data[m_index][MarkerProperty.TYP.value]

            ### Update Position if any changed
            lat_inner = float(devices_data[m_index][MarkerProperty.LAT.value])
            lon_inner = float(devices_data[m_index][MarkerProperty.LON.value])

            if lat_inner != m['props']['position'][0] or lon_inner != m['props']['position'][1]:
                m['props']['position'] = [ lat_inner, lon_inner]


            ### Update Icon if any changed
            device_icon_type = MarkerType .from_string(devices_data[m_index][MarkerProperty.TYP.value])
            device_color     = MarkerColor.from_string(devices_data[m_index][MarkerProperty.COLOR.value])

            if MarkerIcon.getType(m['props']['icon']) != device_icon_type:
                m['props']['icon'] = MarkerIcon(device_icon_type, device_color).getProperties()


            def getNumberOfJammedMLATantenas(devices_data):
                ''' MLAT Antena Jammed - has color RED or ORANGE.'''
                count = 0
                for d in devices_data:
                    if d[MarkerProperty.TYP.value] == MarkerType.MLAT_ANTENNA.value:
                        color = d[MarkerProperty.COLOR.value]
                        if color == MarkerColor.RED.value or color == MarkerColor.ORANGE.value:
                            count += 1
                return count

            def compare_probability(rms_row, device_data, rms_id):
                if device_data[MarkerProperty.PROBABILITY.value.lower()] == None:
                    return False
                try:
                    device = float(device_data[MarkerProperty.PROBABILITY.value.lower()])
                    rms    = float(rms_row    [RmsProperty.   PROBABILITY.value.lower()])
                except ValueError:
                    app.logger.warning(f"Probability for RM {rms_id} is not a number!")
                    return False
                return device >= rms

            def compare_incoming(rms_row, device_data, rms_id):
                device = str(device_data[MarkerProperty.INCOMING.value.lower()])
                rms    = str(rms_row    [RmsProperty.   INCOMING.value.lower()])
                return device == rms


            def compare_backup(rms_row, device_data, rms_id):
                device = str(device_data[MarkerProperty.BACKUP.value.lower()])
                rms    = str(rms_row    [RmsProperty.   BACKUP.value.lower()])
                return device == rms

            def compare_redundancy(rms_row, device_data, rms_id):
                # Redundancy is obtained from number of jammed MLAT Antenas - jammed MLAT Anetna is defined as state, 
                #   when at least one RM for MLAT Antena is active
                if device_data[MarkerProperty.TYP.value] == MarkerType.MLAT.value:
                    try:
                        rms = int( rms_row[RmsProperty.REDUNDANCY.value.lower()] )
                        cnt = getNumberOfJammedMLATantenas(devices_data)
                        return cnt >= rms
                    except ValueError:
                        # ValueError is raised when RM Redundancy is not a number
                        app.logger.warning(f"Redundancy for RM {rms_id} is not a number!")
                return False

            def compare_holdover(rms_row, time_jam_since, actual_time, rms_id):
                # Device does not have holdover, time is taken from TIME_JAM_SINCE
                rms_seconds = rms_row[RmsProperty.HOLDOVER.value.lower()]

                if time_jam_since == None or rms_seconds == None:
                    return False
                
                try:
                    time_jammed = int(time_jam_since)
                    rms         = int(rms_seconds) * 60      # RM Holdover is in minutes!
                except ValueError:
                    # Not all RMs have Holdver attribute
                    return False
                return (actual_time-time_jammed) >= rms


            # Get list of RMs rows and sort the given RMs by it's Probability ascending order
            rms_rows_indices         = getRmsByDeviceType(rms_table, device_type)
            rms_rows                 = [ (row_index, rms_table[row_index]) for row_index in rms_rows_indices ]
            rms_sorted_probabilities = sorted(rms_rows, key=lambda x: x[1][RmsProperty.PROBABILITY.value.lower()])
            rms_sorted_indexes       = [index for index, _ in rms_sorted_probabilities]

            rms_probabilities = [ value[RmsProperty.PROBABILITY.value.lower()] for index,value in rms_sorted_probabilities ]
            time_limits       = [ value[RmsProperty.HOLDOVER   .value.lower()] for index,value in rms_sorted_probabilities ]
            # Holdover is in minutes!


            # Booleans for check after following for-cycle, to check if this tested device:
            deviceCouldBeOrange  = None
            deviceCannotBeOrange = False
            deviceShouldBeRed    = None

            # Due to fact DataTable cannot store Dictionary in cell, TIME_JAM_SICE and RMS_CONFIRMED are stored in 'devices_table' as stringified json
            json_time_jammed = json.loads(devices_data[m_index][MarkerProperty.TIME_JAM_SINCE.value])
            json_rms_confirm = json.loads(devices_data[m_index][MarkerProperty.RMS_CONFIRMED .value])

            # First - filter RM ids which are not in lists (eg. device type changed)
            def filterUnusedRMS(rms_sorted_indexes, rms_table, json_dict):
                filtered = {}
                if json_dict != None:
                    for rms_row_index in rms_sorted_indexes:
                        rms_row = rms_table[rms_row_index]
                        rms_id  = rms_row[RmsProperty.ID.value]
                        for key,value in json_dict.items():
                            if key == rms_id:
                                filtered.update({key:value})
                return filtered

            json_time_jammed = filterUnusedRMS(rms_sorted_indexes, rms_table, json_time_jammed)
            json_rms_confirm = filterUnusedRMS(rms_sorted_indexes, rms_table, json_rms_confirm)


            # Iterato over all RM for given device type
            for rms_row_index in rms_sorted_indexes:

                canBeAddedToRmsInfoTable = False

                rms_row = rms_table[rms_row_index]
                rms_id  = rms_row[RmsProperty.ID.value]

                # Check if given RMs_id exists in both lists (TIME_JAM_SINCE, RMS_CONFIRMED) - device could change type in runtime etc.
                json_time_jammed.setdefault(rms_id, None)
                json_rms_confirm.setdefault(rms_id, None)

                is_probability = compare_probability(rms_row, devices_data[m_index], rms_id)
                is_incoming    = compare_incoming(   rms_row, devices_data[m_index], rms_id)
                is_backup      = compare_backup(     rms_row, devices_data[m_index], rms_id)
                is_redundancy  = compare_redundancy( rms_row, devices_data[m_index], rms_id)
                is_holdover    = compare_holdover(   rms_row, json_time_jammed[rms_id], fusion_time, rms_id)


                # Save or reset timers for given Device and RM
                if is_probability and json_time_jammed[rms_id] == None:
                    # json_time_jammed[rms_id] = time_tick      # In case to use app/host pc/Dash tick time
                    json_time_jammed[rms_id] = fusion_time      # In case to use time from SW Fusin


                if not is_probability:
                    json_time_jammed[rms_id] = None                

                # Check only for Probability
                if device_type == MarkerType.AIRPLANE.value     or \
                   device_type == MarkerType.VEHICLE.value      or \
                   device_type == MarkerType.GROUND.value       or \
                   device_type == MarkerType.MLAT_ANTENNA.value:
                    if is_probability:
                        canBeAddedToRmsInfoTable = True
                    pass
                
                # Check for Probability, Backup and Holdover
                elif device_type == MarkerType.RADAR.value      or \
                     device_type == MarkerType.ADSB.value       or \
                     device_type == MarkerType.PROCESSING.value:
                    if is_probability and is_backup and is_holdover:
                        canBeAddedToRmsInfoTable = True
                    pass

                # Check for Probability, Backup, Incoming, Holdover and Redundancy
                elif device_type == MarkerType.MLAT.value:
                    if is_probability and is_backup and is_holdover and is_incoming and is_redundancy:
                        canBeAddedToRmsInfoTable = True
                    pass

                # No more of Device Types
                else:
                    app.logger.warning(f"There is such no Device type to handle: {device_type}")

                ### Logic behid adding or removing Entry from RMs Info Table
                rm_item = {'Name': device_name, 'RMs': rms_table[rms_row_index][RmsProperty.RM.value.lower()], 'RM_ID': rms_id}

                # RM Entry can be added to RMs Info Table - BUT - should it?
                if canBeAddedToRmsInfoTable:

                    if not json_rms_confirm[rms_id]:
                        deviceCouldBeOrange      = False
                        deviceCannotBeOrange     = True
                        deviceShouldBeRed        = True
                        json_rms_confirm[rms_id] = False

                        if rm_item not in rm_info_table_data:
                            rm_info_table_data.insert(0, rm_item)

                    else:
                        deviceCouldBeOrange = True

                else:
                    # Remove row from RMs Info Table if present
                    if rm_item in rm_info_table_data:
                        rm_info_table_data.remove(rm_item)

                    # RM not active - RMs_Confirmed has to be set to None (False means RM is active, but not confirmed by User)
                    json_rms_confirm[rms_id] = None


            ### Update color and other properties only if there is any RM to process
            if rms_rows_indices:

                new_color = None
                if deviceCouldBeOrange and not deviceCannotBeOrange:
                    new_color = MarkerColor.ORANGE
                elif deviceShouldBeRed:
                    new_color = MarkerColor.RED
                else:
                    new_color = MarkerColor.GREEN

                # Save the TIME_JAM_SINCE and TIME_JAM_SINCE into 'devices_table' as stringified json
                devices_data[m_index][MarkerProperty.TIME_JAM_SINCE.value] = json.dumps(json_time_jammed)
                devices_data[m_index][MarkerProperty.RMS_CONFIRMED .value] = json.dumps(json_rms_confirm)

            else:
                # In case there are NO RMs for given device type, I can safely change color to green (as no RM is affecting give device type)
                new_color = MarkerColor.GREEN

            # Update color only if color changed
            if device_color != new_color:
                m, devices_data[m_index], _ = update_device_color(m, devices_data[m_index], device_icon_type, new_color)

            ### Update Marker Popup content
            # Update Intensity in Marker's Pop-up
            marker_index   = next((index for index, item in enumerate(marker_popup_id) if item.get('index') == m_index), None)
            time_jam_since = json.loads(devices_data[m_index][MarkerProperty.TIME_JAM_SINCE.value])

            if tile_index != None:
                tile_intensity = intensities.get(tile_index)

                if marker_index != None:
                    marker_popup_intensity[marker_index] = getMarkerPopupContent(device_name, device_type, tile_intensity, device_sensitivity, device_probability, rms_probabilities, time_jam_since, fusion_time, time_limits)

            else:
                marker_popup_intensity[marker_index] = getMarkerPopupContent(device_name, device_type, intensity=None, probability=device_probability, sensitivity=device_sensitivity)
                continue




    return time_str,                           \
        dash_markers,                          \
        devices_data,                          \
        rm_info_table_data,                    \
        rm_info_table_data,                    \
        picture_path,                          \
        geo_bounds,                            \
        tiles_indexes,                         \
        jamming_signal,                        \
        probability_density,                   \
        patched_1d_figure,                     \
        patched_2d_figure,                     \
        patched_3d_figure,                     \
        patched_4d_figure,                     \
        dropdown_jam,                          \
        dropdown_2d_probability_density_value, \
        redis_connected_str,                   \
        marker_popup_intensity,                \
        active_spec_vis_tab,                   \
        colorscale,                            \
        store_colorbar_min_max,                \
        store_colorbar_maximum



# -----------------------------------------------------------------------------
if __name__ == '__main__':
    pass

